//----------------------------------------------------------------------------
//	SOURCE:	App.cpp
//	AUTHOR:	Ian Taylor
//	NOTES:	Curling game application class
//----------------------------------------------------------------------------

#include "Curling.h"
#include "App.h"

TEA_NAMESPACE_USE

void Application::fatalError(const Error& error)
{
	MONITOR(Application::fatalError)

	// Report a fatal error
	error.report(hwnd(), "Curling - Fatal Error");
	return;
}

bool Application::initialise(void)
{
	MONITOR(Application::initialise)

	// Initialise the application object
	m_config.player[0].set("Human", true, false, REAL(1.0), REAL(0.5));
	m_config.player[1].set("Computer", false, false, REAL(1.0), REAL(0.5));
	m_config.game.stones = 8;
	m_config.game.ends = 10;
	m_config.game.guards = 3;
	m_config.game.random_ice = false;
	m_config.quick = false;
	m_config.sound = m_sound.initialise();
	m_renderer = NULL;
	m_random.reset();
	VALIDATE(*this);
	return true;
}

void Application::silence(void)
{
	MONITOR(Application::silence)

	// Silence any sound
	if (m_sound.isActive())
		m_sound.shutdown();
	return;
}

bool Application::enableSound(bool enable)
{
	MONITOR(Application::enableSound)

	// Enable/disable sound
	m_config.sound = enable;
	if (enable != m_sound.isActive())
		if (!enable)
			m_sound.shutdown();
		else if (!m_sound.initialise())
		{
			// Failed to initialise the sound system
			m_config.sound = false;
			return false;
		}
	return true;
}

bool Application::enableQuick(bool enable)
{
	MONITOR(Application::enableQuick)

	// Enable/disable quick camera
	m_config.quick = enable;
	return true;
}

void Application::newRenderer(Renderer::Type type)
{
	MONITOR(Application::newRenderer)

	// Create a new renderer
	if (m_renderer != NULL)
		m_renderer->destroy();
	m_renderer = Renderer::create(type, hcanvas());
	ASSERT(m_renderer != NULL);
	return;
}

void Application::shutdown(void)
{
	MONITOR(Application::shutdown)

	// Shutdown the application object
	if (m_renderer != NULL)
		m_renderer->destroy();
	m_renderer = NULL;
	m_sound.shutdown();
	return;
}

void Application::onIdle(DialogMain& main)
{
	MONITOR(Application::onIdle)

	// The application is asynchronously idle
	if ((m_renderer != NULL) && m_renderer->windowReady() && m_game.onIdle())
	{
		// Play any sounds
		m_game.sound.updateSound(m_sound);

		// Check for the end of the throw
		if (!m_game.view.motion)
			m_game.onMotionComplete(main, m_sound);

		// Refresh the canvas
		m_renderer->sceneUpdate(m_game.view);
		m_renderer->windowPaint(hcanvas());
	}
	return;
}

void Application::onWakeup(DialogMain& main)
{
	MONITOR(Application::onWakeup)

	// The application is ready to continue
	VALIDATE(*this);
	m_game.nextStone(main);
	ASSERT(m_renderer != NULL);
	return;
}

void Application::onCanvasClick(DialogMain& main, int mx, int my)
{
	MONITOR(Application::onCanvasClick)

	// Mouse click in the canvas window
	int sx, sy;
	canvasSize(sx, sy);
	ASSERT(sx > 0);
	ASSERT(sy > 0);
	real scale = REAL(1.0) / sy;
	real x = (mx * 2 - sx) * scale;
	real y = (my * 2 - sy) * scale;
	m_game.view.camera.toWorld(x, y);
	m_game.onClick(main, x, y);
	return;
}

void Application::onCanvasThrow(DialogMain& main)
{
	MONITOR(Application::onCanvasThrow)

	// The player has launched the stone
	m_game.onThrow(main);
	return;
}

void Application::onCanvasZoom(real zoom)
{
	MONITOR(Application::onCanvasZoom)

	// The user has moved the zoom slider
	real ymin = CURLING_ZOOM_MIN - (CURLING_ZOOM_MAX + CURLING_ZOOM_MIN) * zoom;
	real pz = GameCamera::lookDownZ(ymin, CURLING_ZOOM_MAX);
	if (!Math::isNear(m_game.view.camera.pz, pz))
	{
		// Change the camera height
		m_game.view.camera.lookDown(ymin, CURLING_ZOOM_MAX);
		ASSERT(m_renderer != NULL);
		m_renderer->sceneUpdate(m_game.view);
		m_renderer->windowPaint(hcanvas());
	}
	return;
}
