//----------------------------------------------------------------------------
//	SOURCE:	Game.cpp
//	AUTHOR:	Ian Taylor
//	NOTES:	Curling computer intelligence
//----------------------------------------------------------------------------

#include "Curling.h"
#include "Game.h"
#include "Config.h"
#include "Dynamic.h"

TEA_NAMESPACE_USE

// The grid of payoff values
struct GameGrid
{
	// The payoff function
	struct Payoff
	{
		// The state of the world after all event are processed
		struct FinalState : public GameWorld
		{
			GameStone::Colour me;			// The colour of the computer
			int score;						// Score relative to me
			real closest[2];				// The closest stones for each player (squared)
			real scorers[GameMaxStones];	// The scoring stones (squared)

			// Methods
			bool fromWorld(const GameWorld& world, real spin, real power, real aim);
			bool isNaughty(const GameWorld& world, bool free_guards) const;
		};

		// Payoff evaluation for non-crap states
		virtual real payoff(const FinalState& state) const = 0;

		// The value assigned to crap states
		static const real crap;
	};

	// Constants
	enum ComputerConstants
	{
		SpinsMin = 4,
		SpinsMax = 30,
		PowersMin = 11,
		PowersMax = 30,
		AimsMin = 7,
		AimsMax = 30
	};

	// Data members
	const GameWorld* before;
	int thrown;

	// The simulation grid
	int spins;
	int powers;
	int aims;
	real spa[SpinsMax][PowersMax][AimsMax];

	// Properties
	real toSpin(int index) const
	{
		// Compute spin from a grid index
		MONITOR(GameGrid::toSpin)
		VALIDATE(*this);
		ASSERT((index >= 0) && (index < spins));
		return REAL(CURLING_SPIN_MAX - CURLING_SPIN_MIN) * index / (spins - 1) + REAL(CURLING_SPIN_MIN);
	}
	real toPower(int index) const
	{
		// Compute power from a grid index
		MONITOR(GameGrid::toPower)
		VALIDATE(*this);
		ASSERT((index >= 0) && (index < powers));
		real p = Math::cube(index / (real)(powers - 1) - REAL(0.5)) * 4 + REAL(0.5);
		ASSERT((p >= 0) && (p <= 1));
		return REAL(CURLING_POWER_MAX - CURLING_POWER_MIN) * p + REAL(CURLING_POWER_MIN);
	}
	real toAim(int index) const
	{
		// Compute aim from a grid index
		MONITOR(GameGrid::toAim)
		VALIDATE(*this);
		ASSERT((index >= 0) && (index < aims));
		return REAL(CURLING_AIM_MAX - CURLING_AIM_MIN) * index / (aims - 1) + REAL(CURLING_AIM_MIN);
	}

	// Methods
	void reset(const GameWorld& world, int already_thrown, real tactics);
	void fill(const GameComputer& me, const Payoff& payoff, Random& random);
	bool strategy(const GameComputer& me, const Game& game);
	void merge(void);
	void highest(int& spin, int& power, int& aim) const;

	// Implementation
	static real lowest(real a, real b, real c)
	{
		real x = (a < b) ? ((a < c) ? a : c) : ((b < c) ? b : c);
		ASSERT(x <= a);
		ASSERT(x <= b);
		ASSERT(x <= c);
		return x;
	}

	// Debugging
	bool isValidSPA(int spin, int power, int aim) const
	{
		return (spin >= 0) && (spin < spins) &&
			   (power >= 0) && (power < powers) &&
			   (aim >= 0) && (aim < aims);
	}
	bool isValid(void) const
	{
		return (spins >= SpinsMin) && (spins <= SpinsMax) &&
			   (powers >= PowersMin) && (powers <= PowersMax) &&
			   (aims >= AimsMin) && (aims <= AimsMax) &&
			   (before != NULL) && before->isValid();
	}
};

// Constants
const real GameGrid::Payoff::crap = -9999;

bool GameGrid::Payoff::FinalState::fromWorld(const GameWorld& world, real spin, real power, real aim)
{
	MONITOR(GameGrid::Payoff::FinalState::fromWorld)

	// Determine the final state of a fully-evolved world
	VALIDATE(world);
	ASSERT(world.stones > 0);
	int knocks = 0;
	DynamicWorld dynamic[2];
	DynamicWorld* now = &dynamic[0];
	DynamicWorld* next = &dynamic[1];
	now->reset(world, world.stones - 1, spin, power, aim);
	now->nextEvent(*next);
	while (next->event != DynamicWorld::NoEvent)
	{
		switch (next->event)
		{
			case DynamicWorld::HitNothing:
				// A stone stopped moving
				break;

			case DynamicWorld::HitLimit:
			case DynamicWorld::HitBoard:
				// If no stones have been knocked, we must have run straight out from release
				if (knocks == 0)
					return false;
				break;

			case DynamicWorld::HitStone:
				// Two stones hit each other
				knocks++;
				break;

			case DynamicWorld::HitStraighten:
				// Nothing to do
				ASSERT(knocks == 0);
				break;

			default:
				UNREACHABLE();
		}
		Memory::swap(now, next);
		now->nextEvent(*next);
	}

	// Copy the state information from the dynamic view (ignore rotation)
	stones = world.stones;
	ASSERT(stones == now->stones);
	closest[0] = closest[1] = CURLING_HOUSE_R * CURLING_HOUSE_R;
	int i;
	for (i = 0; i < stones; i++)
	{
		// Find the closest of each colour while we're at it
		stone[i].colour = world.stone[i].colour;
		stone[i].x = (real)(now->stone[i].x0);
		stone[i].y = (real)(now->stone[i].y0);
		stone[i].rot = 0;
		scorers[i] = Math::sqr(stone[i].x) + Math::sqr(stone[i].y);
		if (scorers[i] < closest[stone[i].colour])
			closest[stone[i].colour] = scorers[i];
	}
	VALIDATE(*this);

	// Determine if we've got a positive or negative score
	score = 0;
	if (closest[me] < closest[1 - me])
	{
		// We score
		for (i = 0; i < stones; i++)
			if ((stone[i].colour == me) && (scorers[i] < closest[1 - me]))
				score++;
			else
				scorers[i] = 0;
	}
	else if (closest[me] > closest[1 - me])
	{
		// Our opponents score
		for (i = 0; i < stones; i++)
			if ((stone[i].colour != me) && (scorers[i] < closest[me]))
				score--;
			else
				scorers[i] = 0;
	}
	return true;
}

bool GameGrid::Payoff::FinalState::isNaughty(const GameWorld& before, bool free_guards) const
{
	MONITOR(GameGrid::Payoff::FinalState::isNaughty)

	// Determine if the fianl state is naughty (breaking rules)
	VALIDATE(*this);
	VALIDATE(before);
	ASSERT(stones == before.stones);
	for (int i = 0; i < stones; i++)
	{
		// Free guard violations or overlapping stones
		if (free_guards && before.stone[i].isInFreeGuardZone() && !stone[i].isInFreeGuardZone())
			return true;
		for (int j = 0; j < i; j++)
		{
			real d = Math::sqr(stone[i].x - stone[j].x) + Math::sqr(stone[i].y - stone[j].y);
			if (d < REAL(CURLING_STONE_DIAMETER * CURLING_STONE_DIAMETER))
				return true;
		}
	}
	return false;
}

void GameGrid::reset(const GameWorld& world, int already_thrown, real tactics)
{
	MONITOR(GameGrid::reset)

	// Reset the game grid
	before = &world;
	thrown = already_thrown;
	spins = Math::round((SpinsMax - SpinsMin) * tactics) + SpinsMin;
	powers = Math::round((PowersMax - PowersMin) * tactics) + PowersMin;
	aims = Math::round((AimsMax - AimsMin) * tactics) + AimsMin;
	VALIDATE(*this);
	return;
}

void GameGrid::fill(const GameComputer& me, const Payoff& payoff, Random& random)
{
	MONITOR(GameGrid::reset)

	// Fill the game grid with values from a payoff
	VALIDATE(*this);
	VALIDATE(me);
	ASSERT(me.config != NULL);
	bool free_guards = (thrown < me.config->game.guards);
	Payoff::FinalState after;
	after.me = me.me;
	for (int spin_index = 0; spin_index < spins; spin_index++)
	{
		real spin = toSpin(spin_index);
		for (int aim_index = 0; aim_index < aims; aim_index++)
		{
			real aim = toAim(aim_index);
			for (int power_index = 0; power_index < powers; power_index++)
			{
				real power = toPower(power_index);
				if (after.fromWorld(*before, spin, power, aim))
				{
					// We reached a non-crap state, but check for naughtiness
					if (after.isNaughty(*before, free_guards))
						spa[spin_index][power_index][aim_index] = payoff.crap;
					else
						spa[spin_index][power_index][aim_index] = payoff.payoff(after) + random.uniform() * REAL(0.01);
				}
				else
				{
					// This and all states with greater power are crap
					do
						spa[spin_index][power_index][aim_index] = payoff.crap;
					while (++power_index < powers);
					break;
				}
			}
		}
	}
	return;
}

void GameGrid::merge(void)
{
	MONITOR(GameGrid::highest)

	// Merge cells in the grid by finding the minimum of the cells and it's eight neighbours
	VALIDATE(*this);
	int s, p, a;
	real merged[PowersMax][AimsMax];
	for (s = 0; s < spins; s++)
	{
		// Top row
		for (a = 0; a < aims; a++)
			merged[0][a] = Math::min(spa[s][0][a], spa[s][1][a]);

		// Middle rows
		for (p = 1; p < (powers - 1); p++)
			for (a = 0; a < aims; a++)
				merged[p][a] = lowest(spa[s][p-1][a], spa[s][p][a], spa[s][p+1][a]);

		// Bottom row
		for (a = 0; a < aims; a++)
			merged[powers-1][a] = Math::min(spa[s][powers-2][a], spa[s][powers-1][a]);

		// Now merge the rows
		for (p = 0; p < powers; p++)
		{
			real x = Math::min(merged[p][0], merged[p][1]);
			ASSERT(x <= spa[s][p][0]);
			spa[s][p][0] = Math::min(merged[p][0], merged[p][1]);
			for (a = 1; a < (aims - 1); a++)
			{
				x = lowest(merged[p][a-1], merged[p][a], merged[p][a+1]);
				ASSERT(x <= spa[s][p][a]);
				spa[s][p][a] = x;
			}
			x = Math::min(merged[p][aims-2], merged[p][aims-1]);
			ASSERT(x <= spa[s][p][aims-1]);
			spa[s][p][aims-1] = x;
		}
	}
	return;

}

void GameGrid::highest(int& spin, int& power, int& aim) const
{
	MONITOR(GameGrid::highest)

	// Find the highest single value in the grid
	VALIDATE(*this);
	spin = spins / 2;
	power = powers / 2;
	aim = aims / 2;
	real best = spa[spin][power][aim];
	for (int s = 0; s < spins; s++)
		for (int p = 0; p < powers; p++)
			for (int a = 0; a < aims; a++)
			{
				real next = spa[s][p][a];
				if (next > best)
				{
					spin = s;
					power = p;
					aim = a;
					best = next;
				}
			}
	ASSERT(isValidSPA(spin, power, aim));
	return;
};

void GameComputer::choose(const Game& game, real& spin, real& power, real& aim) const
{
	MONITOR(GameComputer::choose)

	// Get the computer to choose the next throw parameters
	VALIDATE(*this);
	VALIDATE(game);

	// Add the stone to be released in the simulation
	GameWorld before = game.view.world;
	GameStone& new_stone = before.stone[before.stones++];
	new_stone.colour = me;
	new_stone.x = 0;
	new_stone.y = REAL(CURLING_RINK_RELEASE);
	new_stone.rot = 0;
	VALIDATE(before);

	// Fill in the grid
	GameGrid grid;
	grid.reset(before, game.thrown, tactics);
	if (!grid.strategy(*this, game))
	{
		// No strategy
		struct PayoffSimple : public GameGrid::Payoff
		{
			virtual real payoff(const FinalState& state) const
			{
				if (state.score == 0)
					return 0;
				real total = 0;
				for (int i = 0; i < state.stones; i++)
					total += state.scorers[i];
				return state.score - total / (state.score * CURLING_HOUSE_R * CURLING_HOUSE_R);
			}
		}
		payoff_simple;
		grid.fill(*this, payoff_simple, *game.random);
	}

	// Merge values in the grid if there's possibility of error
	if (config->game.random_ice || (tactics < (game.random->uniform() * REAL(0.9))))
		grid.merge();

	// Determine the final spin/power/aim of the next shot
	int spin_index, power_index, aim_index;
	grid.highest(spin_index, power_index, aim_index);
	ASSERT(grid.isValidSPA(spin_index, power_index, aim_index));
	spin = grid.toSpin(spin_index);
	power = grid.toPower(power_index);
	aim = grid.toAim(aim_index);
	return;
}

bool GameGrid::strategy(const GameComputer& me, const Game& game)
{
	MONITOR(GameGrid::strategy)

	// Possibly use strategy to fill in the game grid
	VALIDATE(*this);
	VALIDATE(me);
	VALIDATE(game);
	ASSERT(game.config != NULL);
	const Configuration::Game& cfg = game.config->game;
	if (me.strategy <= game.random->uniform())
		return false;
	if (game.thrown == (cfg.stones * 2 - 1))
	{
		// This is the very last stone of the end, just go for points or blank
		struct PayoffLast : public GameGrid::Payoff
		{
			int blank_up_to;
			virtual real payoff(const FinalState& state) const
			{
				int score = state.score;
				real mean = 0;
				if (score != 0)
				{
					for (int i = 0; i < state.stones; i++)
						mean += state.scorers[i] / (CURLING_HOUSE_R * CURLING_HOUSE_R);
					mean /= score;
				}
				if ((score >= 0) && (score < blank_up_to))
					score = score ? 0 : 1;
				return score - mean;
			}
		}
		payoff_last;
		ASSERT(game.first != me.me);
		int points_diff = game.total[me.me] - game.total[1 - me.me];
		if (game.end >= (cfg.ends - 1))
		{
			// This is the last end of the match, blank if we cannot win
			payoff_last.blank_up_to = Math::max(-points_diff, 0);
		}
		else if ((game.end % 2) == 0)
		{
			// This is an odd end, blank if we're not losing badly
			payoff_last.blank_up_to = (points_diff >= -1) ? 1 : 0;
		}
		else
		{
			// Don't go for blanks
			payoff_last.blank_up_to = 0;
		}
		ASSERT(payoff_last.blank_up_to >= 0);
		fill(me, payoff_last, *game.random);
	}
	else if (game.first == me.me)
	{
		// I have first stone, so favour guards
		if (game.thrown < (cfg.guards - 1))
		{
			// Favour guards in front of the house
			struct PayoffGuard : public GameGrid::Payoff
			{
				virtual real payoff(const FinalState& state) const
				{
					real score = (real)state.score;
					for (int i = 0; i < state.stones; i++)
						if ((state.stone[i].colour == state.me) &&
							(Math::abs(state.stone[i].x) < REAL(CURLING_RINK_WIDTH / 4)) &&
							(state.stone[i].y > CURLING_HOG_Y) && (state.stone[i].y < -CURLING_INNER_Y))
							score += 3 + Math::hypot(state.stone[i].x, state.stone[i].y) / CURLING_HOG_Y;
					return score;
				}
			}
			payoff_guard;
			fill(me, payoff_guard, *game.random);
		}
		else
		{
			// Favour covered stones towards the front of the house
			struct PayoffFront : public GameGrid::Payoff
			{
				static real coverage(const FinalState& state, int i)
				{
					// Find the composite coverage for a single stone
					real best = 0;
					for (int j = 0; j < state.stones; j++)
						if ((i != j) && ((state.stone[i].y - state.stone[j].y) < REAL(CURLING_STONE_DIAMETER * 1.5)))
						{
							real overlap = 1 - Math::abs(state.stone[i].x - state.stone[j].x) / REAL(CURLING_STONE_DIAMETER);
							if (overlap > best)
								best = overlap;
						}
					ASSERT((best >= 0) && (best <= 1));
					return best;
				}
				virtual real payoff(const FinalState& state) const
				{
					real score = 0;
					if (state.score > 0)
					{
						// Maximise the number of covered scorers
						for (int i = 0; i < state.stones; i++)
							if (state.scorers[i] > 0)
								score += coverage(state, i);
					}
					else if (state.score < 0)
					{
						// Just place things at the front of the house
						for (int i = 0; i < state.stones; i++)
							if ((state.stone[i].colour == state.me) && state.stone[i].isInHouse())
								score += CURLING_HOUSE_R * CURLING_HOUSE_R - state.stone[i].y;
						score /= CURLING_HOUSE_R * CURLING_HOUSE_R * (state.stones + 2);
					}
					return state.score + score;
				}
			}
			payoff_front;
			fill(me, payoff_front, *game.random);
		}
	}
	else
	{
		// I have last stone, so keep the house clear of our opponent's stones
		struct PayoffTakeout : public GameGrid::Payoff
		{
			virtual real payoff(const FinalState& state) const
			{
				int taken_out = 0;
				for (int i = 0; i < state.stones; i++)
					if ((state.stone[i].colour != state.me) && !state.stone[i].isInPlay())
						taken_out++;
				if (taken_out == 0)
				{
					real q = (state.closest[1 - state.me] - state.closest[state.me]) / (CURLING_HOUSE_R * CURLING_HOUSE_R * 4);
					ASSERT((q >= -0.5) && (q <= 0.5));
					return state.score + q;
				}
				return state.score + taken_out * REAL(0.9);
			}
		}
		payoff_takeout;
		fill(me, payoff_takeout, *game.random);
	}
	return true;
}

