//----------------------------------------------------------------------------
//	SOURCE:	Dialog.cpp
//	AUTHOR:	Ian Taylor
//	NOTES:	Curling game dialog classes
//----------------------------------------------------------------------------

#include "Curling.h"
#include <windows.h>
#include <commctrl.h>
#include "Dialog.h"
#include "App.h"
#include "Resource.h"

TEA_NAMESPACE_USE

// Unique id for the wakeup timer
#define TIMER_WAKEUP 1969

void Dialog::Slider::initialise(HWND hwnd, bool central, int range)
{
	MONITOR(Dialog::Slider::initialise)

	// Set slider attributes
	ASSERT(hwnd != NULL);
	m_window = hwnd;
	ASSERT(m_window != NULL);
	if (range <= 0)
	{
		// Work out the extremes of the slider's position
		RECT left, right;
		::SendMessage(m_window, TBM_SETPOS, TRUE, MAXLONG);
		::SendMessage(m_window, TBM_GETTHUMBRECT, 0, reinterpret_cast<LPARAM>(&right));
		::SendMessage(m_window, TBM_SETPOS, TRUE, MINLONG);
		::SendMessage(m_window, TBM_GETTHUMBRECT, 0, reinterpret_cast<LPARAM>(&left));
		if (range == 0)
			range = right.left - left.left;
		else
			range = (left.left - right.left) / range;
	}
	ASSERT(range > 0);
	::SendMessage(m_window, TBM_SETRANGE, FALSE, MAKELPARAM(0, range));
	::SendMessage(m_window, TBM_SETPAGESIZE, 0, (range + 5) / 10);
	if (central)
		::SendMessage(m_window, TBM_SETTIC, 0, (range + 1) / 2);
	m_dirty = true;
	VALIDATE(*this);
	return;
}

void Dialog::Slider::set(real value)
{
	MONITOR(Dialog::Slider::set)

	// Set the aim slider control
	VALIDATE(*this);
	int range = ::SendMessage(m_window, TBM_GETRANGEMAX, 0, 0);
	int pos = Math::round(value * range);
	::SendMessage(m_window, TBM_SETPOS, TRUE, pos);
	m_value = value;
	m_dirty = false;
	return;
}

real Dialog::Slider::get(void) const
{
	MONITOR(Dialog::Slider::get)

	// Set the aim slider control
	VALIDATE(*this);
	if (!m_dirty)
		return m_value;
	int range = ::SendMessage(m_window, TBM_GETRANGEMAX, 0, 0);
	int pos = ::SendMessage(m_window, TBM_GETPOS, 0, 0);
	return real(pos) / range;
}

void Dialog::initialiseSlider(int id)
{
	MONITOR(Dialog::initialiseSlider)

	// Initialise a dumb slider
	VALIDATE(*this);
	HWND hwnd = ::GetDlgItem(m_window, id);
	ASSERT(hwnd != NULL);

	// Work out the extremes of the slider's position
	RECT left, right;
	::SendMessage(hwnd, TBM_SETPOS, TRUE, MAXLONG);
	::SendMessage(hwnd, TBM_GETTHUMBRECT, 0, reinterpret_cast<LPARAM>(&right));
	::SendMessage(hwnd, TBM_SETPOS, TRUE, MINLONG);
	::SendMessage(hwnd, TBM_GETTHUMBRECT, 0, reinterpret_cast<LPARAM>(&left));
	int range = right.left - left.left;
	ASSERT(range > 0);
	::SendMessage(hwnd, TBM_SETRANGE, FALSE, MAKELPARAM(0, range));
	::SendMessage(hwnd, TBM_SETPAGESIZE, 0, (range + 5) / 10);
	return;
}

void Dialog::setFocus(int id)
{
	MONITOR(Dialog::setFocus)

	// Set the current focus to a control
	VALIDATE(*this);
	::SetFocus(::GetDlgItem(m_window, id));
	return;
}

void Dialog::setVisible(int id, bool visible)
{
	MONITOR(Dialog::setVisible)

	// Show/hide a control
	VALIDATE(*this);
	HWND hctrl = ::GetDlgItem(m_window, id);
	ASSERT(hctrl != NULL);
	::ShowWindow(hctrl, visible ? SW_SHOWNA : SW_HIDE);
	return;
}

void Dialog::setEnable(int id, bool enable)
{
	MONITOR(Dialog::setEnable)

	// Enable a control
	VALIDATE(*this);
	HWND hctrl = ::GetDlgItem(m_window, id);
	ASSERT(hctrl != NULL);
	::EnableWindow(hctrl, enable ? TRUE : FALSE);
	return;
}

void Dialog::setMenuCheck(int id, bool value)
{
	MONITOR(Dialog::setMenuCheck)

	// Set the menu item checkbox
	VALIDATE(*this);
	HMENU hmenu = ::GetMenu(m_window);
	ASSERT(hmenu != NULL);
	hmenu = ::GetSubMenu(hmenu, (id / 100) - 1);
	ASSERT(hmenu != NULL);
	MENUITEMINFO mii = { sizeof(mii) };
	mii.fMask = MIIM_STATE;
	mii.fState = MFS_ENABLED|(value ? MFS_CHECKED : MFS_UNCHECKED);
	VERIFY(::SetMenuItemInfo(hmenu, id, FALSE, &mii));
	return;
}

void Dialog::setText(int id, const char* value)
{
	MONITOR(Dialog::setText)

	// Set the text of a control
	VALIDATE(*this);
	VERIFY(::SetDlgItemText(m_window, id, value));
	return;
}

void Dialog::setInteger(int id, int value, const char* when_zero)
{
	MONITOR(Dialog::setInteger)

	// Set the integer value of a control
	VALIDATE(*this);
	if ((value == 0) && (when_zero != NULL))
		VERIFY(::SetDlgItemText(m_window, id, when_zero));
	else
		VERIFY(::SetDlgItemInt(m_window, id, value, TRUE));
	return;
}

void Dialog::setCheck(int id, bool value)
{
	MONITOR(Dialog::setCheck)

	// Set the state of a check/radio control
	VALIDATE(*this);
	VERIFY(::CheckDlgButton(m_window, id, value ? BST_CHECKED : BST_UNCHECKED));
	return;
}

void Dialog::setCombo(int id, int value)
{
	MONITOR(Dialog::setCombo)

	// Set the position of a combo control
	VALIDATE(*this);
	HWND hctrl = ::GetDlgItem(m_window, id);
	ASSERT(hctrl != NULL);
	int index = -1;
	LRESULT lresult;
	do
	{
		lresult = ::SendMessage(hctrl, CB_GETITEMDATA, ++index, 0);
		ASSERT(lresult != CB_ERR);
	}
	while (lresult != value);
	::SendMessage(hctrl, CB_SETCURSEL, index, 0);
	return;
}

void Dialog::setSlider(int id, real value)
{
	MONITOR(Dialog::setSlider)

	// Set the position of a slider control
	VALIDATE(*this);
	HWND hctrl = ::GetDlgItem(m_window, id);
	ASSERT(hctrl != NULL);
	int range = ::SendMessage(hctrl, TBM_GETRANGEMAX, 0, 0);
	int pos = Math::round(value * range);
	::SendMessage(hctrl, TBM_SETPOS, TRUE, pos);
	return;
}

bool Dialog::getMenuCheck(int id) const
{
	MONITOR(Dialog::getMenuCheck)

	// Get the menu item checkbox state
	VALIDATE(*this);
	HMENU hmenu = ::GetMenu(m_window);
	ASSERT(hmenu != NULL);
	hmenu = ::GetSubMenu(hmenu, (id / 100) - 1);
	ASSERT(hmenu != NULL);
	return (::GetMenuState(hmenu, id, MF_BYCOMMAND) & MF_CHECKED) != 0;
}

int Dialog::getText(int id, char* buffer, size_t size) const
{
	MONITOR(Dialog::getText)

	// Fetch text from a control
	VALIDATE(*this);
	return ::GetDlgItemText(m_window, id, buffer, size);
}

bool Dialog::getCheck(int id) const
{
	MONITOR(Dialog::getCheck)

	// Fetch the check state of a control
	VALIDATE(*this);
	return ::IsDlgButtonChecked(m_window, id) == BST_CHECKED;
}

int Dialog::getCombo(int id) const
{
	MONITOR(Dialog::getCombo)

	// Fetch the selection of a combo control
	VALIDATE(*this);
	HWND hctrl = ::GetDlgItem(m_window, id);
	ASSERT(hctrl != NULL);
	LRESULT lresult = ::SendMessage(hctrl, CB_GETCURSEL, 0, 0);
	ASSERT(lresult != CB_ERR);
	return ::SendMessage(hctrl, CB_GETITEMDATA, lresult, 0);
}

real Dialog::getSlider(int id) const
{
	MONITOR(Dialog::getSlider)

	// Fetch the position of a slider control
	VALIDATE(*this);
	HWND hctrl = ::GetDlgItem(m_window, id);
	ASSERT(hctrl != NULL);
	int range = ::SendMessage(hctrl, TBM_GETRANGEMAX, 0, 0);
	int pos = ::SendMessage(hctrl, TBM_GETPOS, 0, 0);
	return real(pos) / range;
}

void Dialog::wakeup(int milliseconds)
{
	MONITOR(Dialog::wakeup)

	// Wakeup this dialog in the future
	VALIDATE(*this);
	::KillTimer(m_window, TIMER_WAKEUP);
	VERIFY(::SetTimer(m_window, TIMER_WAKEUP, milliseconds, NULL));
	return;
}

void Dialog::endDialog(int retcode)
{
	MONITOR(Dialog::endDialog)

	// Terminate this dialog session
	VALIDATE(*this);
	VERIFY(::EndDialog(m_window, retcode));
	return;
}

void DialogMain::initialise(HWND hwnd, Application* app)
{
	MONITOR(DialogMain::initialise)

	// Initialise the main dialog object
	m_window = hwnd;
	m_app = app;
	if (m_window == NULL)
		return;

	// Determine the minimum size this dialog can be (the initial size)
	RECT rect;
	VERIFY(::GetWindowRect(hwnd, &rect));
	m_minWidth = rect.right - rect.left;
	m_minHeight = rect.bottom - rect.top;

	// Set the icon (large and small)
	HINSTANCE hinstance = reinterpret_cast<HINSTANCE>(::GetWindowLong(hwnd, GWL_HINSTANCE));
	ASSERT(hinstance != NULL);
	HANDLE hicon;
	hicon = ::LoadIcon(hinstance, MAKEINTRESOURCE(IDI_ICON));
	ASSERT(hicon != NULL);
	::SendMessage(hwnd, WM_SETICON, ICON_BIG, reinterpret_cast<LPARAM>(hicon));
	hicon = ::LoadIcon(hinstance, MAKEINTRESOURCE(IDI_YELLOW));
	ASSERT(hicon != NULL);
	::SendMessage(hwnd, WM_SETICON, ICON_SMALL, reinterpret_cast<LPARAM>(hicon));

	// Configure the canvas DC
	HWND hctrl = ::GetDlgItem(hwnd, IDC_CANVAS);
	ASSERT(hctrl != NULL);
	HDC hdc = ::GetDC(hctrl);
	::SetBkMode(hdc, TRANSPARENT);

	// Set slider ranges
	m_zoom.initialise(::GetDlgItem(hwnd, IDC_ZOOM), false, -5);
	m_spin.initialise(::GetDlgItem(hwnd, IDC_SPIN), true);
	m_power.initialise(::GetDlgItem(hwnd, IDC_POWER), false);
	m_aim.initialise(::GetDlgItem(hwnd, IDC_AIM), true);
	resetSliders();
	return;
}

BOOL DialogMain::onMessage(UINT umsg, WPARAM wparam, LPARAM lparam)
{
	MONITOR(DialogMain::onMessage)

	// Message handler for main dialog box
	switch (umsg)
	{
		case WM_ERASEBKGND:
			// Exclude the canvas from the erase region
			{
				ASSERT(m_window != NULL);
				HWND hctrl = ::GetDlgItem(m_window, IDC_CANVAS);
				ASSERT(hctrl != NULL);
				RECT rect;
				VERIFY(::GetWindowRect(hctrl, &rect));
				VERIFY(::MapWindowPoints(HWND_DESKTOP, m_window, reinterpret_cast<POINT*>(&rect), 2));
				HDC hdc = reinterpret_cast<HDC>(wparam);
				ASSERT(hdc != NULL);
				VERIFY_(::ExcludeClipRect(hdc, rect.left, rect.top, rect.right, rect.bottom),!= ERROR);
			}
			break;

		case WM_GETMINMAXINFO:
			// Adjust screen-to-client coords
			{
				MINMAXINFO* mmi = reinterpret_cast<MINMAXINFO*>(lparam);
				ASSERT(mmi != NULL);
				mmi->ptMinTrackSize.x = m_minWidth;
				mmi->ptMinTrackSize.y = m_minHeight;
			}
			return TRUE;

		case WM_MOVE:
			// Main window has moved, so notify the canvas
			onCanvasMove();
			break;

		case WM_SIZE:
			// The user has resized the window (always chain this message)
			onSize(static_cast<short>(LOWORD(lparam)), static_cast<short>(HIWORD(lparam)));
			break;

		case WM_DISPLAYCHANGE:
			// The display mode has changed
			onCanvasMode();
			break;

		case WM_COMMAND:
			// Surely a button or menu item...
			return onCommand(HIWORD(wparam), LOWORD(wparam));

		case WM_HSCROLL:
			// A slider has changed
			return onCommand(SB_THUMBPOSITION, ::GetDlgCtrlID(reinterpret_cast<HWND>(lparam)));

		case WM_TIMER:
			// Probably the wakeup timer
			if (wparam == TIMER_WAKEUP)
			{
				// Shouldn't have a TimerProc
				ASSERT(lparam == NULL);
				VERIFY(::KillTimer(m_window, wparam));
				if (m_app != NULL)
					m_app->onWakeup(*this);
				return TRUE;
			}
			break;

		case WM_CLOSE:
			// The user has selected "Close" in the system menu
			return onCommand(BN_CLICKED, IDCLOSE);
	}
	return FALSE;
}

BOOL DialogMain::onCommand(int code, int id)
{
	MONITOR(DialogMain::onCommand)

	// Command handler for main dialog box
	switch (code)
	{
		case BN_CLICKED:
			// Button/menu click
			switch (id)
			{
				case IDOK:
					// Return (release stone)
					if (m_app != NULL)
						onCanvasThrow();
					return TRUE;

				case IDCANCEL:
					// Reset the inputs
					resetSliders();
					m_app->onCanvasZoom(m_zoom.get());
					return TRUE;

				case IDCLOSE:
				case IDM_GAME_EXIT:
					// Quit the application
					::PostQuitMessage(0);
					return TRUE;

				case IDM_GAME_NEW:
					// Go back to requesting a new game
					::PostQuitMessage(1);
					return TRUE;

				case IDM_OPTIONS_SOUND:
					// Toggle the sound status
					if (m_app != NULL)
					{
						bool sound = !getMenuCheck(IDM_OPTIONS_SOUND);
						sound = sound & m_app->enableSound(sound);
						setMenuCheck(IDM_OPTIONS_SOUND, sound);
					}
					return TRUE;

				case IDM_OPTIONS_QUICK:
					// Toggle the quick camera status
					if (m_app != NULL)
					{
						bool quick = !getMenuCheck(IDM_OPTIONS_QUICK);
						if (m_app->enableQuick(quick))
							setMenuCheck(IDM_OPTIONS_QUICK, quick);
					}
					return TRUE;

				case IDM_HELP_ABOUT:
					// Bring on the about box
					if ((m_app != NULL) && !m_app->onDialogAbout())
						::MessageBeep(MB_ICONEXCLAMATION);
					return TRUE;
			}
			break;

		case SB_THUMBPOSITION:
			// Trackbar changes
			switch (id)
			{
				case IDC_ZOOM:
					// Aim change
					ASSERT(m_app != NULL);
					m_zoom.makeDirty();
					m_app->onCanvasZoom(m_zoom.get());
					break;

				case IDC_SPIN:
					// Spin change
					m_spin.makeDirty();
					break;

				case IDC_POWER:
					// Power change
					m_power.makeDirty();
					break;

				case IDC_AIM:
					// Aim change
					m_aim.makeDirty();
					break;
			}
			break;

		case STN_DBLCLK:
			// The red or yellow stone icon has been double-clicked
			ASSERT((id == IDC_ICON1) || (id == IDC_ICON2));
			if (::IsWindowEnabled(::GetDlgItem(m_window, IDC_POWER)))
				onCanvasThrow();
			break;
	}
	return FALSE;
}

void DialogMain::onSize(int width, int height)
{
	MONITOR(DialogMain::onSize)

	// If necessary, resize the canvas (always full height)
	ASSERT(m_window != NULL);
	if (::IsIconic(m_window))
		return;
	HWND hctrl = ::GetDlgItem(m_window, IDC_CANVAS);
	ASSERT(hctrl != NULL);
	POINT topleft = { 0, 0 };
	VERIFY(::MapWindowPoints(hctrl, m_window, &topleft, 1));
	ASSERT(topleft.y == 0);
	VERIFY(::SetWindowPos(hctrl, NULL, 0, 0, width - topleft.x, height,
						  SWP_SHOWWINDOW|SWP_NOACTIVATE|SWP_NOZORDER|SWP_NOMOVE));
	return;
}

void DialogMain::onCanvasCreate(HWND hwnd)
{
	MONITOR(DialogMain::onCanvasCreate)

	// The canvas window has just been created
	ASSERT(hwnd != NULL);
	HDC hdc = ::GetDC(hwnd);
	ASSERT(hdc != NULL);
	HWND hdlg = ::GetParent(hwnd);
	ASSERT(hdlg != NULL);
	HFONT hfont = reinterpret_cast<HFONT>(::SendMessage(hdlg, WM_GETFONT, 0, 0));
	if (hfont != NULL)
		VERIFY(::SelectObject(hdc, hfont));
	VERIFY_(::SetTextColor(hdc, ::GetSysColor(COLOR_BTNTEXT)),!= CLR_INVALID);
	VERIFY_(::SetBkColor(hdc, ::GetSysColor(COLOR_3DFACE)),!= CLR_INVALID);
	VERIFY(::SetBkMode(hdc, OPAQUE));
	return;
}

void DialogMain::onCanvasPaint(void)
{
	MONITOR(DialogMain::onCanvasPaint)

	// Need to paint the canvas
	if (m_app == NULL)
		return;
	m_app->onCanvasPaint();
	return;
}

void DialogMain::onCanvasMove(void)
{
	MONITOR(DialogMain::onCanvasMove)

	// The canvas has changed position, shape or colour depth
	if (m_app != NULL)
		m_app->onCanvasChange(Renderer::ChangePosition);
	return;
}

void DialogMain::onCanvasResize(void)
{
	MONITOR(DialogMain::onCanvasResize)

	// The canvas has changed shape
	if (m_app != NULL)
		m_app->onCanvasChange(Renderer::ChangeSize);
	return;
}

void DialogMain::onCanvasMode(void)
{
	MONITOR(DialogMain::onCanvasMode)

	// The canvas has changed colour depth
	if (m_app != NULL)
		m_app->onCanvasChange(Renderer::ChangeMode);
	return;
}

void DialogMain::onCanvasClick(int mx, int my)
{
	MONITOR(DialogMain::onCanvasClick)

	// Mouse clicks within the canvas
	if (m_app == NULL)
		return;
	m_app->onCanvasClick(*this, mx, my);
	return;
}

void DialogMain::onCanvasThrow(void)
{
	MONITOR(DialogMain::onCanvasThrow)

	// Fetch the throw parameters and get the game to launch the stone
	if (m_app == NULL)
		return;
	setEnable(IDC_CANVAS, false);
	setEnable(IDC_ZOOM, false);
	setEnable(IDC_SPIN, false);
	setEnable(IDC_POWER, false);
	setEnable(IDC_AIM, false);
	m_app->onCanvasThrow(*this);
	return;
}

void DialogMain::showViolation(const char* violation)
{
	MONITOR(DialogMain::showViolation)

	// Show a violation string
	ASSERT(m_app != NULL);
	m_app->onDialogViolation(violation);
	return;
}

void DialogConfig::initialise(HWND hwnd, Configuration* config)
{
	MONITOR(DialogConfig::initialise)

	// Constant table
	static const char number[11][6] =
	{
		"None", "One", "Two", "Three", "Four", "Five", "Six", "Seven", "Eight", "Nine", "Ten"
	};
	static const struct
	{
		int id;
		int value;
	}
	table[] =
	{
		{ IDC_STONES,	4 },
		{ IDC_STONES,	6 },
		{ IDC_STONES,	8 },
		{ IDC_ENDS,		1 },
		{ IDC_ENDS,		3 },
		{ IDC_ENDS,		6 },
		{ IDC_ENDS,		10 },
		{ IDC_GUARDS,	0 },
		{ IDC_GUARDS,	3 },
		{ IDC_GUARDS,	4 },
	};

	// Initialise the config dialog object
	m_window = hwnd;
	m_config = config;
	if (m_window == NULL)
		return;
	VALIDATE(*this);

	// Limit the editbox sizes
	::SendDlgItemMessage(m_window, IDC_NAME1, EM_SETLIMITTEXT, dimof(m_config->player[0].name) - 1, 0);
	::SendDlgItemMessage(m_window, IDC_NAME2, EM_SETLIMITTEXT, dimof(m_config->player[1].name) - 1, 0);

	// Fill in the combo-boxes
	int i;
	for (i = 0; i < dimof(number); i++)
	{
		VERIFY_(::SendDlgItemMessage(m_window, IDC_HANDICAP1, CB_ADDSTRING, 0, reinterpret_cast<LPARAM>(number[i])),== i);
		VERIFY_(::SendDlgItemMessage(m_window, IDC_HANDICAP2, CB_ADDSTRING, 0, reinterpret_cast<LPARAM>(number[i])),== i);
		VERIFY_(::SendDlgItemMessage(m_window, IDC_HANDICAP1, CB_SETITEMDATA, i, i),!= CB_ERR);
		VERIFY_(::SendDlgItemMessage(m_window, IDC_HANDICAP2, CB_SETITEMDATA, i, i),!= CB_ERR);
	}
	for (i = 0; i < dimof(table); i++)
	{
		LONG index = ::SendDlgItemMessage(m_window, table[i].id, CB_ADDSTRING, 0, reinterpret_cast<LPARAM>(number[table[i].value]));
		ASSERT(index >= 0);
		VERIFY_(::SendDlgItemMessage(m_window, table[i].id, CB_SETITEMDATA, index, table[i].value),!= CB_ERR);
	}

	// Set slider ranges
	initialiseSlider(IDC_SKILL1);
	initialiseSlider(IDC_ACCURACY1);
	initialiseSlider(IDC_TACTICS1);
	initialiseSlider(IDC_STRATEGY1);
	initialiseSlider(IDC_SKILL2);
	initialiseSlider(IDC_ACCURACY2);
	initialiseSlider(IDC_TACTICS2);
	initialiseSlider(IDC_STRATEGY2);

	// Initialise the controls
	load(*m_config);
	return;
}

void DialogConfig::load(const Configuration& config)
{
	MONITOR(DialogConfig::load)

	// Load the state into the controls
	VALIDATE(*this);
	for (int p = 0; p < 2; p++)
	{
		// Load one player's attributes to the controls
		const Configuration::Player& player = config.player[p];
		setText(IDC_NAME1 + p, player.name);
		setCombo(IDC_HANDICAP1 + p, player.handicap);
		setCheck(IDC_NOVICE1 + p, player.human && !player.expert);
		setCheck(IDC_EXPERT1 + p, player.human && player.expert);
		setCheck(IDC_RANDOM1 + p, !player.human && !player.expert);
		setCheck(IDC_CUSTOM1 + p, !player.human && player.expert);
		setSlider(IDC_SKILL1 + p, player.skill);
		setSlider(IDC_ACCURACY1 + p, player.accuracy);
		setSlider(IDC_TACTICS1 + p, player.tactics);
		setSlider(IDC_STRATEGY1 + p, player.strategy);
	}
	setCombo(IDC_STONES, config.game.stones);
	setCombo(IDC_ENDS, config.game.ends);
	setCombo(IDC_GUARDS, config.game.guards);
	setCheck(IDC_RANDOM_ICE, config.game.random_ice);
	setCheck(IDC_QUICK, config.quick);
	setCheck(IDC_SOUND, config.sound);
	return;
}

void DialogConfig::save(Configuration& config, bool demo)
{
	MONITOR(DialogConfig::save)

	// Save the state of the controls
	VALIDATE(*this);
	if (demo)
	{
		// Standard player information
		config.player[0].set("Red", false, true, REAL(1.0), REAL(1.0));
		config.player[1].set("Yellow", false, true, REAL(1.0), REAL(1.0));
	}
	else
	{
		// Read player information from the dialog
		for (int p = 0; p < 2; p++)
		{
			// Save one player's attributes from the controls
			Configuration::Player& player = config.player[p];
			player.handicap = getCombo(IDC_HANDICAP1 + p);
			player.human = player.expert = false;
			if (getCheck(IDC_NOVICE1 + p))
				player.human = true;
			else if (getCheck(IDC_EXPERT1 + p))
				player.human = player.expert = true;
			else if (getCheck(IDC_CUSTOM1 + p))
				player.expert = true;
			else
				ASSERT(getCheck(IDC_RANDOM1 + p));
			if (getText(IDC_NAME1 + p, player.name, dimof(player.name)) <= 0)
				System::asciizCopy(player.name, player.human ? "Human" : "Computer", dimof(player.name));
			player.skill = getSlider(IDC_SKILL1 + p);
			player.accuracy = getSlider(IDC_ACCURACY1 + p);
			player.tactics = getSlider(IDC_TACTICS1 + p);
			player.strategy = getSlider(IDC_STRATEGY1 + p);
		}
	}
	config.game.stones = getCombo(IDC_STONES);
	config.game.ends = getCombo(IDC_ENDS);
	config.game.guards = getCombo(IDC_GUARDS);
	config.game.random_ice = getCheck(IDC_RANDOM_ICE);
	config.quick = getCheck(IDC_QUICK);
	config.sound = getCheck(IDC_SOUND);

	// Check for stupid names
	if (System::asciizCompare(config.player[0].name, config.player[1].name) == 0)
	{
		// Duplicate names (append 1 and 2)
		int length = Math::min(System::asciizLength(config.player[0].name), dimof(config.player[0].name) - 3);
		System::asciizCopy(config.player[0].name + length, " 1", 3);
		System::asciizCopy(config.player[1].name + length, " 2", 3);
	}
	return;
}

void DialogConfig::swap(void)
{
	MONITOR(DialogConfig::swap)

	// Swap the players over
	Configuration config;
	save(config);
	Memory::swap(config.player[0], config.player[1]);
	load(config);
	return;
}

BOOL DialogConfig::onMessage(UINT umsg, WPARAM wparam, LPARAM lparam)
{
	MONITOR(DialogConfig::onMessage)

	// Main message pump for config dialog box
	switch (umsg)
	{
		case WM_COMMAND:
			// Surely a button or menu item...
			if (HIWORD(wparam) == BN_CLICKED)
				switch (LOWORD(wparam))
				{
					case IDOK:
						// Save and end dialog
						save(*m_config);
						endDialog(IDOK);
						return TRUE;

					case IDCANCEL:
						// End dialog
						endDialog(IDCANCEL);
						return TRUE;

					case IDC_DEMO:
						// Set standard parameters and exit
						save(*m_config, true);
						endDialog(IDOK);
						return TRUE;

					case IDC_SWAP:
						// Swap the two player configurations
						swap();
						return TRUE;
				}
			break;

		case WM_HSCROLL:
			// Trackbar changes
			ASSERT(lparam != 0);
			switch (::GetDlgCtrlID(reinterpret_cast<HWND>(lparam)))
			{
				case IDC_SKILL1:
					// Player 1 random computer
					setCheck(IDC_NOVICE1, false);
					setCheck(IDC_EXPERT1, false);
					setCheck(IDC_CUSTOM1, false);
					setCheck(IDC_RANDOM1, true);
					break;

				case IDC_SKILL2:
					// Player 2 random computer
					setCheck(IDC_NOVICE2, false);
					setCheck(IDC_EXPERT2, false);
					setCheck(IDC_CUSTOM2, false);
					setCheck(IDC_RANDOM2, true);
					break;

				case IDC_ACCURACY1:
				case IDC_TACTICS1:
				case IDC_STRATEGY1:
					// Player 1 custom computer
					setCheck(IDC_NOVICE1, false);
					setCheck(IDC_EXPERT1, false);
					setCheck(IDC_RANDOM1, false);
					setCheck(IDC_CUSTOM1, true);
					break;

				case IDC_TACTICS2:
				case IDC_ACCURACY2:
				case IDC_STRATEGY2:
					// Player 2 custom computer
					setCheck(IDC_NOVICE2, false);
					setCheck(IDC_EXPERT2, false);
					setCheck(IDC_RANDOM2, false);
					setCheck(IDC_CUSTOM2, true);
					break;

				case IDC_RANDOM_ICE:
				case IDC_QUICK:
				case IDC_SOUND:
					// Random ice/sound
					break;

				default:
					// Oh dear
					UNREACHABLE();
					break;
			}
			return FALSE;

		case WM_CLOSE:
			// The user has selected "Close" in the system menu
			endDialog(IDCLOSE);
			return TRUE;
	}
	return FALSE;
}
