//----------------------------------------------------------------------------
//	HEADER:	Dialog.h
//	AUTHOR:	Ian Taylor
//	NOTES:	Curling game dialog classes
//----------------------------------------------------------------------------

#ifndef TEA_DIALOG_H
#define TEA_DIALOG_H
#ifdef TEA_DEBUG_INCLUDES
#pragma message("#include " __FILE__)
#endif

class Application;
struct Configuration;

class Dialog
{
	protected:

		// Private data
		Window m_window;		// The dialog window handle

		// Implementation
		void initialiseSlider(int id);

		// Internal clas
		class Slider
		{
			private:

				// Private data
				Window m_window;		// The slider window handle
				real m_value;			// The current value, if not dirty
				bool m_dirty;			// Whether the slider has been moved via the GUI

			public:

				// Methods
				void initialise(Window window, bool central = false, int range = 0);
				void makeDirty(void) { m_dirty = true; }
				void set(real value);
				real get(void) const;

				// Debugging
				bool isValid(void) const
				{
					return TEA::Memory::isValidRead(this) && (m_window != NULL);
				}
		};

	public:

		// Construction
		void shutdown(void);

		// Methods
		void setFocus(int id);
		void setVisible(int id, bool visible);
		void setEnable(int id, bool enable);
		void setMenuCheck(int id, bool value);
		void setText(int id, const char* value);
		void setInteger(int id, int value, const char* when_zero = NULL);
		void setCheck(int id, bool value);
		void setCombo(int id, int value);
		void setSlider(int id, real value);
		void wakeup(int milliseconds);
		void endDialog(int retcode);

		// Properties
		bool getMenuCheck(int id) const;
		int getText(int id, char* buffer, size_t size) const;
		bool getCheck(int id) const;
		int getCombo(int id) const;
		real getSlider(int id) const;
		Window handle(void) const { return m_window; }
		bool isWindow(Window window) const { return m_window == window; }

		// Debugging
		bool isValid(void) const
		{
			return TEA::Memory::isValidRead(this) && (m_window != NULL);
		}
};

class DialogMain : public Dialog
{
	private:

		// Data members
		Application* m_app;
		int m_minWidth;
		int m_minHeight;
		Slider m_zoom;
		Slider m_spin;
		Slider m_power;
		Slider m_aim;

	public:

		// Initialisation
		void initialise(Window window, Application* app);

		// Methods
		void resetSliders(void)
		{
			// Reset the values to defaults
			MONITOR(DialogMain::resetSliders)
			m_spin.set(REAL(0.5));
			m_power.set(REAL(0.0));
			m_aim.set(REAL(0.5));
		}
		void setSpin(real value)
		{
			m_spin.set((value - REAL(CURLING_SPIN_MIN)) / REAL(CURLING_SPIN_MAX - CURLING_SPIN_MIN));
		}
		void setPower(real value)
		{
			m_power.set((value - REAL(CURLING_POWER_MIN)) / REAL(CURLING_POWER_MAX - CURLING_POWER_MIN));
		}
		void setAim(real value)
		{
			m_aim.set((value - REAL(CURLING_AIM_MIN)) / REAL(CURLING_AIM_MAX - CURLING_AIM_MIN));
		}
		void showViolation(const char* violation = NULL);

		// Properties
		real getSpin(void) const
		{
			return m_spin.get() * REAL(CURLING_SPIN_MAX - CURLING_SPIN_MIN) + REAL(CURLING_SPIN_MIN);
		}
		real getPower(void) const
		{
			return m_power.get() * REAL(CURLING_POWER_MAX - CURLING_POWER_MIN) + REAL(CURLING_POWER_MIN);
		}
		real getAim(void) const
		{
			return m_aim.get() * REAL(CURLING_AIM_MAX - CURLING_AIM_MIN) + REAL(CURLING_AIM_MIN);
		}

		// Message handlers
		int onMessage(unsigned int umsg, unsigned int wparam, long lparam);
		int onCommand(int code, int id);
		void onSize(int width, int height);
		void onCanvasCreate(Window window);
		void onCanvasPaint(void);
		void onCanvasMove(void);
		void onCanvasResize(void);
		void onCanvasMode(void);
		void onCanvasClick(int mx, int my);
		void onCanvasThrow(void);
		void onAwaken(void);

		// Debugging
		bool isValid(void) const
		{
			return Dialog::isValid() && (m_app != NULL) &&
				   m_zoom.isValid() && m_spin.isValid() && m_power.isValid() && m_aim.isValid();
		}
};

class DialogConfig : public Dialog
{
	private:

		// Data members
		Configuration* m_config;

		// Implementation methods
		void load(const Configuration& config);
		void save(Configuration& config, bool demo = false);
		void swap(void);

	public:

		// Initialisation
		void initialise(Window window, Configuration* config);

		// Message handler
		int onMessage(unsigned int umsg, unsigned int wparam, long lparam);

		// Debugging
		bool isValid(void) const
		{
			return Dialog::isValid() && (m_config != NULL);
		}
};

#endif
