//----------------------------------------------------------------------------
//	SOURCE:	Dynamic.cpp
//	AUTHOR:	Ian Taylor
//	NOTES:	Curling world in motion class
//----------------------------------------------------------------------------

#include "Curling.h"
#include "Game.h"
#include "Dynamic.h"

TEA_NAMESPACE_USE

void DynamicWorld::spinSolve(real x, real y, real spin, real& power, real& aim)
{
	MONITOR(DynamicWorld::spinSolve)

	// Solve the power and aim parameters given a final position and spin
	power = y - REAL(CURLING_RINK_RELEASE);
	ASSERT(power > 0);
	aim = Math::atan(x, power);
	if (Math::isNearZero(spin))
		return;

	// Binary chop to find the aim
	real vy2 = REAL(2 * CURLING_DECELERATION) * power;
	real speed = Math::sqrt(vy2 - 2 * REAL(CURLING_DECELERATION) * REAL(CURLING_RINK_STRAIGHTEN - CURLING_RINK_RELEASE));
	ASSERT(speed > 0);
	real delta = REAL(CURLING_AIM_MAX);
	for (int i = 0; i < 10; i++)
	{
		// Simplistic root finder
		real sin_a, cos_a;
		Math::sincos(aim, sin_a, cos_a);
		real cx = cos_a / spin;
		real cy = REAL(CURLING_RINK_STRAIGHTEN - CURLING_RINK_RELEASE) + sin_a / spin;
		real x0 = cx - Math::sgn(spin) * Math::sqrt(1 / Math::sqr(spin) - Math::sqr(cy));
		real gx = x0 + (y - REAL(CURLING_RINK_STRAIGHTEN)) * Math::sin2tan(cy * spin);
		if (gx < x)
			aim += delta;
		else
			aim -= delta;
		delta /= 2;
	}
	return;
}

void DynamicWorld::reset(const GameWorld& world, int thrown, real spin, real power, real aim)
{
	MONITOR(DynamicWorld::reset)

	// Reset the world in motion
	VALIDATE(world);
	ASSERT((thrown >= 0) && (thrown < world.stones));

	// Copy relevant data to the dynamic view of the world
	Memory::zero(*this);
	timeshift = -CURLING_DRAWBACK_TIME;
	time = timeshift;
	stones = world.stones;
	for (int i = 0; i < stones; i++)
	{
		stone[i].x0 = world.stone[i].x;
		stone[i].y0 = world.stone[i].y;
		stone[i].rot0 = world.stone[i].rot;
	}

	// Calculate the new stone's initial velocity
	surreal vy = Math::sqrt(2 * CURLING_DECELERATION * power);
	stone[thrown].vy0 = vy;
	stone[thrown].vx0 = Math::tan(aim) * vy;
	if (!Math::isNearZero(spin))
	{
		// Curved release
		real sin_a, cos_a;
		Math::sincos(aim, sin_a, cos_a);
		curve.cr = 1 / spin;
		ASSERT(curve.cr != 0);
		curve.cx = curve.cr * cos_a;
		curve.cy = CURLING_RINK_RELEASE - curve.cr * sin_a;
		curver = thrown;
		stone[thrown].mrot0 = spin / CURLING_SPIN_MAX * CURLING_VISUAL_SPIN / vy;
	}
	else
	{
		// Straight release
		curver = -1;
	}
	VALIDATE(*this);
	return;
}

bool DynamicWorld::updateTime(void)
{
	MONITOR(DynamicWorld::updateTime)

	// Work out if the time has changed appreciably
	VALIDATE(*this);
	surreal new_time = System::timerMicroseconds(System::timerElapsed(start, System::timerGet())) * 0.000001 + timeshift;
	ASSERT(new_time >= time);
	if ((new_time - time) < CURLING_UPDATE_THROTTLE)
		return false;
	time = new_time;
	return true;
}

void DynamicWorld::updateView(GameView& view, GameSound& sound, DynamicWorld& upcoming, int thrown)
{
	MONITOR(DynamicWorld::updateView)

	// Work out if the world in motion has changed
	VALIDATE(*this);
	VALIDATE(view);
	ASSERT((thrown >= 0) && (thrown < stones));
	sound.knock_speed = 0;
	view.motion = false;

	// Work out which phase we're in
	ASSERT(stones > 0);
	if (time <= 0)
	{
		// Drawback (two phases)
		surreal drawback = stone[thrown].vy0 * CURLING_DRAWBACK_TIME / 4;
		ASSERT(drawback > 0);
		if (time <= (CURLING_DRAWBACK_TIME / -2))
		{
			// Drawing back (sinusoidal)
			surreal t1 = (time / CURLING_DRAWBACK_TIME + 1) * Math::Pi * 2;
			ASSERT(t1 >= 0);
			surreal s1, c1;
			Math::sincos(t1, s1, c1);
			view.world.stone[thrown].y = REAL(CURLING_RINK_RELEASE) + (real)(drawback * (c1 - 1) / 2);
			ASSERT(s1 >= 0);
			sound.fastest_stone = Math::max((real)(drawback * s1 * Math::Pi / CURLING_DRAWBACK_TIME), Math::Epsilon);
		}
		else
		{
			// Push off (linear acceleration)
			surreal t2 = time + REAL(CURLING_DRAWBACK_TIME / 2);
			ASSERT(t2 > 0);
			surreal a2 = stone[thrown].vy0 / REAL(CURLING_DRAWBACK_TIME / 2);
			ASSERT(a2 > 0);
			view.world.stone[thrown].y = REAL(CURLING_RINK_RELEASE) - (real)(drawback - a2 * Math::sqr(t2) / 2);
			sound.fastest_stone = Math::max((real)(a2 * t2), Math::Epsilon);
		}
		ASSERT((view.world.stone[thrown].y >= REAL(CURLING_RINK_HACK)) && (view.world.stone[thrown].y <= REAL(CURLING_RINK_RELEASE)));
		view.motion = true;
		return;
	}

	// Work out if we've crossed into the timeframe of the next upcoming event
	ASSERT(upcoming.time > 0);
	while ((upcoming.event != NoEvent) && (upcoming.time <= time))
	{
		// An event has occurred
		if (upcoming.knock_speed > sound.knock_speed)
			sound.knock_speed = upcoming.knock_speed;
		*this = upcoming;
		nextEvent(upcoming);
	}
	surreal ymin = -CURLING_ZOOM_MAX;
	if (curver >= 0)
	{
		// Curved release
		real sy = REAL(CURLING_RINK_RELEASE) + (real)((stone[thrown].vy0 - time * CURLING_DECELERATION / 2) * time);
		ASSERT(sy <= REAL(CURLING_RINK_STRAIGHTEN));
		surreal vy1 = stone[thrown].vy0 - time * REAL(CURLING_DECELERATION);
		ASSERT(vy1 > 0);
		sound.fastest_stone = (real)vy1;
		view.world.stone[thrown].x = (real)(curve.cx - Math::sgn(curve.cr) * Math::sqrt(Math::sqr(curve.cr) - Math::sqr(sy - curve.cy)));
		view.world.stone[thrown].y = sy;
		view.world.stone[thrown].rot = (real)(stone[thrown].rot0 + stone[thrown].mrot0 * (stone[thrown].vy0 + vy1) * time / 2);
		ymin = sy;
	}
	else
	{
		// Everything running in straight lines
		sound.fastest_stone = 0;
		for (int i = 0; i < stones; i++)
		{
			// See if this stone is moving
			surreal speed = Math::hypot(stone[i].vx0, stone[i].vy0);
			if (speed > 0)
			{
				// Work out the deceleration
				surreal dt = time - stone[i].t0;
				ASSERT(dt >= 0);
				surreal new_speed = Math::max(speed - dt * REAL(CURLING_DECELERATION), 0.0);
				surreal q = (1 + new_speed / speed) * dt / 2;
				ASSERT(q >= 0);
				view.world.stone[i].x = (real)(stone[i].x0 + stone[i].vx0 * q);
				view.world.stone[i].y = (real)(stone[i].y0 + stone[i].vy0 * q);
				view.world.stone[i].rot = (real)(stone[i].rot0 + stone[i].mrot0 * speed * q);
				if (new_speed > sound.fastest_stone)
					sound.fastest_stone = (real)new_speed;
			}
			if (view.world.stone[i].y < ymin)
				ymin = view.world.stone[i].y;
		}
	}

	// Update the camera position, if it's still moving
	surreal q = Math::power(time, CURLING_CAMERA_LAG) * camera.aho;
	ASSERT(q >= 0);
	if (q < Math::Pi)
	{
		// The camera's still moving
		view.motion = true;
		q = (Math::cos(q) + 1) / 2;
	}
	else
	{
		// The camera's not moving
		view.motion = (sound.fastest_stone > 0);
		q = 0;
	}
	ASSERT((q >= 0) && (q <= 1));
	if (ymin < CURLING_ZOOM_MIN)
		ymin = CURLING_ZOOM_MIN;
	surreal cy0 = (CURLING_ZOOM_MAX + ymin) / 2;
	surreal cz0 = GameCamera::lookDownZ((real)ymin, REAL(CURLING_ZOOM_MAX));
	view.camera.py = (real)((view.world.stone[thrown].y - CURLING_RINK_RELEASE + CURLING_RINK_HACK * 2) / 2 * q + cy0 * (1 - q));
	view.camera.pz = (real)(camera.ullman * q + cz0 * (1 - q));
	ASSERT(view.camera.pz > 0);
	if (view.world.stone[thrown].y > view.camera.py)
		view.camera.ly = (real)(view.world.stone[thrown].y * q + cy0 * (1 - q));
	else
		view.camera.ly = (real)(view.camera.py * q + cy0 * (1 - q));
	surreal cx0 = (CURLING_RINK_STRAIGHTEN - view.camera.py) / (CURLING_RINK_STRAIGHTEN - CURLING_RINK_HACK);
	ASSERT(cx0 <= 1);
	if (cx0 > 0)
		view.camera.px = (real)((1 - Math::cos(cx0 * Math::Pi)) * camera.sethi);
	else
		view.camera.px = 0;
	return;
}

surreal DynamicCurver::hitSide(surreal sx) const
{
	MONITOR(DynamicCurver::hitSide)

	// Work out where the curve path hits the sides
	VALIDATE(*this);
	surreal ycy = Math::sqr(cr) - Math::sqr(sx - cx);
	if (ycy < 0)
		return 0;
	ycy = Math::sqrt(ycy);
	surreal sy = cy - ycy;
	if (sy < REAL(CURLING_RINK_RELEASE))
		sy = cy + ycy;
	if ((sy >= REAL(CURLING_RINK_RELEASE)) && (sy <= REAL(CURLING_RINK_STRAIGHTEN)))
		return sy;
	return 0;
}

bool DynamicStone::wasApproaching(const DynamicStone& other) const
{
	MONITOR(DynamicStone::wasApproaching)

	// Determine if two stones are approaching each other at time t0
	VALIDATE(*this);
	VALIDATE(other);
	ASSERT(t0 == other.t0);
	surreal dx = other.x0 - x0;
	surreal dy = other.y0 - y0;
	surreal dvx = other.vx0 - vx0;
	surreal dvy = other.vy0 - vy0;
	return (dx * dvx + dy * dvy) < -Math::Epsilon;
}

void DynamicStone::releaseSpeedTime(surreal y, surreal& speed, surreal& time) const
{
	MONITOR(DynamicStone::releaseSpeedTime)

	// Work out the speed and time for a stone to reach a certain y-point after release
	VALIDATE(*this);
	ASSERT(y0 == REAL(CURLING_RINK_RELEASE));
	speed = Math::sqrtSafe(Math::sqr(vy0) - REAL(CURLING_DECELERATION * 2) * (y - y0));
	ASSERT(speed >= 0);
	time = (vy0 - speed) / REAL(CURLING_DECELERATION);
	ASSERT(time > 0);
	return;
}

void DynamicStone::evolveStone(const DynamicStone& from, surreal time, surreal speed0)
{
	MONITOR(DynamicStone::evolveStone)

	// Evolve a stone forward in time
	VALIDATE(from);
	ASSERT(time > 0);
	ASSERT(speed0 > 0);
	surreal t = time - from.t0;
	ASSERT(t > 0);
	surreal speed1 = speed0 - REAL(CURLING_DECELERATION) * t;
	ASSERT(speed1 > -Math::Epsilon);
	mrot0 = from.mrot0;
	rot0 = from.rot0 + from.mrot0 * (speed0 + speed1) * t / 2;
	surreal dx0 = REAL(CURLING_DECELERATION) * from.vx0 / speed0;
	surreal dy0 = REAL(CURLING_DECELERATION) * from.vy0 / speed0;
	vx0 = from.vx0 - dx0 * t;
	vy0 = from.vy0 - dy0 * t;
	x0 = from.x0 + (from.vx0 + vx0) * t / 2;
	y0 = from.y0 + (from.vy0 + vy0) * t / 2;
	t0 = time;
	VALIDATE(*this);
	return;
}

surreal DynamicWorld::hitLimitTime(surreal x, surreal t0, surreal x0, surreal vx0, surreal speed0) const
{
	MONITOR(DynamicWorld::hitLimitTime)

	// Calculate when a decelerating ordinate hits a limit
	if (((x - x0) * vx0) <= 0)
		return 9999;
	ASSERT(speed0 > 0);
	surreal dx0 = REAL(CURLING_DECELERATION) * vx0 / speed0;
	surreal vx2 = Math::sqr(vx0) - 2 * dx0 * (x - x0);
	if (vx2 < 0)
		return 9999;
	surreal t = (Math::abs(vx0) - Math::sqrt(vx2)) / Math::abs(dx0);
	ASSERT(t > 0);
	ASSERT((t0 + t) >= time);
	return t0 + t;
}

void DynamicCurver::straighten(const DynamicStone& before, surreal x, surreal y, DynamicStone& after) const
{
	MONITOR(DynamicWorld::straighten)

	// Straighten a curving stone
	surreal sv;
	before.releaseSpeedTime(y, sv, after.t0);
	ASSERT(!Math::isNearZero(cr));
	after.x0 = x;
	after.y0 = y;
	surreal sin_t = (y - cy) / cr;
	after.vx0 = sv * Math::sin2tan(sin_t);
	after.vy0 = sv;
	after.rot0 = before.mrot0 * (before.vy0 + sv) * after.t0 / 2;
	after.mrot0 = before.mrot0;
	return;
}

void DynamicWorld::nextEvent(DynamicWorld& result) const
{
	MONITOR(DynamicWorld::nextEvent)

	// Work out the next thing to happen in this world
	VALIDATE(*this);

	// Special case for curved release
	if (curver >= 0)
	{
		// Work out if we hit the sides before needing to straighten
		result = *this;
		surreal lx = -CURLING_INNER_X;
		surreal ly = curve.hitSide(lx);
		surreal rx = CURLING_INNER_X;
		surreal ry = curve.hitSide(rx);
		if (ly < ry)
		{
			// Hit the left limit
			curve.straighten(stone[curver], lx, ly, result.stone[curver]);
			ASSERT(result.stone[curver].vx0 < 0);
			if (result.stone[curver].vx0 > -0.1)
				result.stone[curver].vx0 = -0.1;
			result.event = HitLimit;
		}
		else if (ry < ly)
		{
			// Hit the right limit
			curve.straighten(stone[curver], rx, ry, result.stone[curver]);
			ASSERT(result.stone[curver].vx0 > 0);
			if (result.stone[curver].vx0 < 0.1)
				result.stone[curver].vx0 = 0.1;
			result.event = HitLimit;
		}
		else
		{
			// We hit the straighten point
			surreal sx = curve.cx - Math::sgn(curve.cr) * Math::sqrt(Math::sqr(curve.cr) - Math::sqr(REAL(CURLING_RINK_STRAIGHTEN) - curve.cy));
			surreal sy = REAL(CURLING_RINK_STRAIGHTEN);
			curve.straighten(stone[curver], sx, sy, result.stone[curver]);
			result.event = HitStraighten;
		}
		result.curver = -1;
		result.time = result.stone[curver].t0;
		result.hitters[0] = result.hitters[1] = 0;
		result.knock_speed = 0;
		return;
	}

	// Work out when each moving stone next interacts
	result.event = NoEvent;
	surreal speed0[GameMaxStones], vt, knock_time = 0;
	int i, j;
	for (i = 0; i < stones; i++)
	{
		// See if this stone is moving
		speed0[i] = Math::hypot(stone[i].vx0, stone[i].vy0);
		if (speed0[i] <= 0)
		{
			// Check for a moving stone hitting us
			if ((result.event != NoEvent) && stone[i].wasInPlay())
			{
				surreal t1 = Math::max(time, 0.001);
				for (j = 0; (j < i) && (t1 < result.time); j++)
					if ((speed0[j] > 0) && stone[j].wasInPlay() && ((hitters[0] != j) || (hitters[1] != i)))
					{
						vt = hitStones(t1, result.time, stone[j], speed0[j], stone[i], 0, knock_time);
						if (vt > 0)
						{
							// These stones collide next
							ASSERT((vt >= time) && (vt < result.time));
							result = *this;
							result.time = vt;
							result.event = HitStone;
							result.hitters[0] = j;
							result.hitters[1] = i;
						}
					}
			}
			continue;
		}

		// Work out when it stops moving
		vt = stone[i].t0 + speed0[i] / REAL(CURLING_DECELERATION);
		ASSERT(vt >= time);
		if ((result.event == NoEvent) || (vt < result.time))
		{
			// This stone stopping is the next event, thus far
			result = *this;
			result.time = vt;
			result.stone[i].t0 = vt;
			vt -= stone[i].t0;
			result.stone[i].x0 = stone[i].x0 + stone[i].vx0 * vt / 2;
			result.stone[i].y0 = stone[i].y0 + stone[i].vy0 * vt / 2;
			result.stone[i].vx0 = 0;
			result.stone[i].vy0 = 0;
			result.event = HitNothing;
			result.hitters[0] = result.hitters[1] = i;
		}

		// Work out when we hit the inner limits
		vt = hitLimitTime(CURLING_INNER_Y, stone[i].t0, stone[i].y0, stone[i].vy0, speed0[i]);
		if (vt < result.time)
		{
			// This stone going out of the back of the house is the next event, thus far
			result = *this;
			result.time = vt;
			result.stone[i].evolveStone(stone[i], vt, speed0[i]);
			result.stone[i].y0 = CURLING_INNER_Y + Math::Epsilon;
			result.event = HitLimit;
			result.hitters[0] = result.hitters[1] = i;
		}
		vt = hitLimitTime(-CURLING_INNER_X, stone[i].t0, stone[i].x0, stone[i].vx0, speed0[i]);
		if (vt < result.time)
		{
			// This stone going out left is the next event, thus far
			result = *this;
			result.time = vt;
			result.stone[i].evolveStone(stone[i], vt, speed0[i]);
			result.stone[i].x0 = -CURLING_INNER_X - Math::Epsilon;
			result.event = HitLimit;
			result.hitters[0] = result.hitters[1] = i;
		}
		vt = hitLimitTime(CURLING_INNER_X, stone[i].t0, stone[i].x0, stone[i].vx0, speed0[i]);
		if (vt < result.time)
		{
			// This stone going out right is the next event, thus far
			result = *this;
			result.time = vt;
			result.stone[i].evolveStone(stone[i], vt, speed0[i]);
			result.stone[i].x0 = CURLING_INNER_X + Math::Epsilon;
			result.event = HitLimit;
			result.hitters[0] = result.hitters[1] = i;
		}

		// Work out when we hit the outer limits
		vt = hitLimitTime(CURLING_OUTER_Y, stone[i].t0, stone[i].y0, stone[i].vy0, speed0[i]);
		if (vt < result.time)
		{
			// This stone hitting the end is the next event, thus far
			result = *this;
			result.time = vt;
			result.stone[i].evolveStone(stone[i], vt, speed0[i]);
			result.stone[i].y0 = CURLING_OUTER_Y;
			result.stone[i].vx0 = 0;
			result.stone[i].vy0 = 0;
			result.event = HitBoard;
			result.hitters[0] = result.hitters[1] = i;
		}
		vt = hitLimitTime(-CURLING_OUTER_X, stone[i].t0, stone[i].x0, stone[i].vx0, speed0[i]);
		if (vt < result.time)
		{
			// This stone hitting the left side is the next event, thus far
			result = *this;
			result.time = vt;
			result.stone[i].evolveStone(stone[i], vt, speed0[i]);
			result.stone[i].x0 = -CURLING_OUTER_X + Math::Epsilon;
			result.stone[i].vx0 = 0;
			result.stone[i].vy0 /= 2;
			result.stone[i].mrot0 -= CURLING_VISUAL_SPIN;
			result.event = HitBoard;
			result.hitters[0] = result.hitters[1] = i;
		}
		vt = hitLimitTime(CURLING_OUTER_X, stone[i].t0, stone[i].x0, stone[i].vx0, speed0[i]);
		if (vt < result.time)
		{
			// This stone hitting the right side is the next event, thus far
			result = *this;
			result.time = vt;
			result.stone[i].evolveStone(stone[i], vt, speed0[i]);
			result.stone[i].x0 = CURLING_OUTER_X - Math::Epsilon;
			result.stone[i].vx0 = 0;
			result.stone[i].vy0 /= 2;
			result.stone[i].mrot0 += CURLING_VISUAL_SPIN;
			result.event = HitBoard;
			result.hitters[0] = result.hitters[1] = i;
		}

		// Work out if we hit a stone (but not the stone we've just collided with)
		ASSERT(speed0[i] > 0);
		if (stone[i].wasInPlay())
		{
			surreal t1 = Math::max(time, 0.001);
			for (j = 0; (j < i) && (t1 < result.time); j++)
			{
				if (stone[j].wasInPlay() && ((hitters[0] != j) || (hitters[1] != i)))
				{
					vt = hitStones(t1, result.time, stone[i], speed0[i], stone[j], speed0[j], knock_time);
					if (vt > 0)
					{
						// These stones collide next
						ASSERT((vt >= time) && (vt < result.time));
						result = *this;
						result.time = vt;
						result.event = HitStone;
						result.hitters[0] = j;
						result.hitters[1] = i;
					}
				}
			}
		}
	}
	VALIDATE(*this);
	if (result.event == HitStone)
	{
		// Work out the effect of the hit
		ASSERT(knock_time > 0);
		ASSERT(result.hitters[0] < result.hitters[1]);
		result.knock_speed = result.knockStones(knock_time, result.time, result.stone[result.hitters[0]], result.stone[result.hitters[1]]);
		ASSERT(result.time >= time);
	}
	else
	{
		// No knocking
		result.knock_speed = 0;
	}
	VALIDATE(*this);
	return;
}

surreal DynamicWorld::hitStones(surreal t1, surreal t2, const DynamicStone& a, surreal sa, const DynamicStone& b, surreal sb, surreal& knock_time)
{
	MONITOR(DynamicWorld::hitStones)

	// Work out if two moving stones, 'a' and 'b', hit between times 't1' and 't2'
	VALIDATE(a);
	VALIDATE(b);
	ASSERT(t1 > 0);
	ASSERT(t1 < t2);
	ASSERT(a.t0 <= t1);
	ASSERT(b.t0 <= t1);
	ASSERT(sa > 0);
	ASSERT(sb >= 0);

	// Work out the motion of 'a' as 'x(a) = ax0 + ax1*(t-t0) + ax2*sqr(t-t0)' etc
	surreal ax0 = a.x0;
	surreal ax1 = a.vx0;
	surreal ax2 = REAL(CURLING_DECELERATION / -2) * (a.vx0 / sa);
	surreal ay0 = a.y0;
	surreal ay1 = a.vy0;
	surreal ay2 = REAL(CURLING_DECELERATION / -2) * (a.vy0 / sa);

	// Work out the motion of 'b' as 'y(b) = by0 + by1*(t-t0) + by2*sqr(t-t0)' etc
	surreal bx0 = b.x0;
	surreal bx1 = b.vx0;
	surreal bx2 = (sb > 0) ? REAL(CURLING_DECELERATION / -2) * (b.vx0 / sb) : 0;
	surreal by0 = b.y0;
	surreal by1 = b.vy0;
	surreal by2 = (sb > 0) ? REAL(CURLING_DECELERATION / -2) * (b.vy0 / sb) : 0;

	// Now express 'x(a) - x(b)' as 'dx(t) = dx0 + dx1*t + dx2*sqr(t)' etc
	surreal dx0 = (a.t0 * (a.t0 * ax2 - ax1) + ax0) - (b.t0 * (b.t0 * bx2 - bx1) + bx0);
	surreal dx1 = (ax1 - 2 * a.t0 * ax2) - (bx1 - 2 * b.t0 * bx2);
	surreal dx2 = ax2 - bx2;
	surreal dy0 = (a.t0 * (a.t0 * ay2 - ay1) + ay0) - (b.t0 * (b.t0 * by2 - by1) + by0);
	surreal dy1 = (ay1 - 2 * a.t0 * ay2) - (by1 - 2 * b.t0 * by2);
	surreal dy2 = ay2 - by2;

	// Now express 'sqr(dx(t))' as 'tx0 + tx1*t + tx2*t*t + tx3*t*t*t + tx4*t*t*t*t' etc
	surreal tx0 = dx0 * dx0;
	surreal tx1 = 2 * dx0 * dx1;
	surreal tx2 = dx1 * dx1 + 2 * dx0 * dx2;
	surreal tx3 = 2 * dx1 * dx2;
	surreal tx4 = dx2 * dx2;
	surreal ty0 = dy0 * dy0;
	surreal ty1 = 2 * dy0 * dy1;
	surreal ty2 = dy1 * dy1 + 2 * dy0 * dy2;
	surreal ty3 = 2 * dy1 * dy2;
	surreal ty4 = dy2 * dy2;

	// Now prepare to solve 'sqr(dx(t)) + sqr(dy(t)) = sqr(D)'
	surreal negative[2][3];
	int n;
	surreal qa = tx4 + ty4;
	surreal qb = tx3 + ty3;
	surreal qc = tx2 + ty2;
	surreal qd = tx1 + ty1;
	surreal qe = tx0 + ty0 - CURLING_STONE_DIAMETER * CURLING_STONE_DIAMETER;
	ASSERT(qa >= 0);
	if (qa < 0.000001)
	{
		// Solution is a cubic
		n = Math::negativeCubic(qb, qc, qd, qe, negative[0]);
	}
	else
	{
		// Solution is a quartic
		n = Math::negativeQuartic(qb / qa, qc / qa, qd / qa, qe / qa, negative);
	}

	// Ignore past collisions
	int i = 0;
	while ((i < n) && (t1 >= Math::min(negative[i][0] + 0.01, negative[i][1])))
		i++;
	if ((i >= n) || (t2 <= negative[i][0]))
		return 0;
	knock_time = negative[i][0];
	ASSERT(knock_time >= (t1 - 1));
	return Math::max(knock_time, t1);
}

real DynamicWorld::knockStones(surreal knock_time, surreal start_time, DynamicStone& a, DynamicStone& b)
{
	MONITOR(DynamicWorld::knockStones)

	// Two stones have collided; either may be stationary
	ASSERT(knock_time > 0);
	ASSERT(start_time >= knock_time);
	ASSERT((knock_time - start_time) < 1);
	surreal t, d, dd;

	// Evolve stone 'a' to the collision point
	VALIDATE(a);
	ASSERT(a.t0 <= start_time);
	t = knock_time - a.t0;
	surreal sa = Math::hypot(a.vx0, a.vy0);
	if (sa > 0)
	{
		// We've moving
		d = 1 - CURLING_DECELERATION * t / sa;
		dd = (1 + d) * t / 2;
		a.x0 += a.vx0 * dd;
		a.y0 += a.vy0 * dd;
		a.vx0 *= d;
		a.vy0 *= d;
		a.rot0 += a.mrot0 * sa * dd;
		sa *= d;
		ASSERT(Math::isNear(sa, Math::hypot(a.vx0, a.vy0)));
	}
	a.t0 = start_time;
	VALIDATE(a);

	// Evolve stone 'b' to the collision point
	VALIDATE(b);
	ASSERT(a.t0 <= start_time);
	t = knock_time - b.t0;
	surreal sb = Math::hypot(b.vx0, b.vy0);
	if (sb > 0)
	{
		// We've moving
		d = 1 - CURLING_DECELERATION * t / sb;
		dd = (1 + d) * t / 2;
		b.x0 += b.vx0 * dd;
		b.y0 += b.vy0 * dd;
		b.vx0 *= d;
		b.vy0 *= d;
		b.rot0 += b.mrot0 * sb * dd;
		sb *= d;
		ASSERT(Math::isNear(sb, Math::hypot(b.vx0, b.vy0)));
	}
	b.t0 = start_time;
	VALIDATE(b);

	// Make sure the stones are the correct distance apart
	surreal dx = b.x0 - a.x0;
	surreal dy = b.y0 - a.y0;
	surreal exy = (Math::sqr(dx) + Math::sqr(dy)) / Math::sqr(CURLING_STONE_DIAMETER);
	if (exy < 1)
	{
		// Separate the stones in proportion to their speeds
		ASSERT((sa + sb) > 0);
		ASSERT(exy > 0.7);
		exy = Math::sqrt(exy);
		surreal ex = (dx / exy - dx) / (sa + sb);
		a.x0 -= ex * sa;
		b.x0 += ex * sb;
		dx /= exy;
		surreal ey = (dy / exy - dy) / (sa + sb);
		a.y0 -= ey * sa;
		b.y0 += ey * sb;
		dy /= exy;
	}
	ASSERT(Math::hypot(dx, dy) >= (CURLING_STONE_DIAMETER - Math::Epsilon));

	// Work out the approach speed
	real knock_speed = 0;
	surreal sin_a, cos_a;
	Math::tan2sincos(dx, dy, sin_a, cos_a);
	surreal dvx = b.vx0 - a.vx0;
	surreal dvy = b.vy0 - a.vy0;
	surreal approach = -(dvx * sin_a + dvy * cos_a);
	if (approach > 0)
	{
		// Adjust the velocities by applying restitution
		surreal f = (1 + CURLING_RESTITUTION) * approach / 2;
		ASSERT(f > 0);
		a.vx0 -= f * sin_a;
		a.vy0 -= f * cos_a;
		b.vx0 += f * sin_a;
		b.vy0 += f * cos_a;
		knock_speed = (real)approach;

		// Work out the angle of hitting to adjust the rotations (unscientific)
		surreal twist = Math::sin(Math::atan(dx, dy) - Math::atan(dvx, dvy)) * CURLING_VISUAL_SPIN * 3;
		surreal arot0 = a.mrot0 - twist - b.mrot0 / 2;
		b.mrot0 -= twist + a.mrot0 / 2;
		a.mrot0 = arot0;
	}
	ASSERT(!a.wasApproaching(b));
	return knock_speed;
}
