//----------------------------------------------------------------------------
//	HEADER:	Dynamic.h
//	AUTHOR:	Ian Taylor
//	NOTES:	Curling dynamic motion classes
//----------------------------------------------------------------------------

#ifndef TEA_DYNAMIC_H
#define TEA_DYNAMIC_H
#ifdef TEA_DEBUG_INCLUDES
#pragma message("#include " __FILE__)
#endif

using TEA::surreal;

struct DynamicCamera
{
	// Camera position parameterisation
	surreal aho;
	surreal sethi;
	surreal ullman;

	// Debugging
	bool isValid(void) const
	{
		return TEA::Memory::isValidRead(this);
	}	
};

struct DynamicStone
{
	// Stone position
	surreal t0;							// The start time
	surreal x0;							// The stone's x-ordinate at t0
	surreal y0;							// The stone's y-ordinate at t0
	surreal rot0;						// The stone's rotation in radians at t0
	surreal vx0;						// The stone's velocity in the x-direction at t0
	surreal vy0;						// The stone's velocity in the y-direction at t0
	surreal mrot0;						// The stone's rotational speed at t0 as a multiple of linear speed

	// Methods
	void releaseSpeedTime(surreal y, surreal& speed, surreal& time) const;
	void evolveStone(const DynamicStone& from, surreal time, surreal speed0);

	// Properties
	bool wasInPlay(void) const { return (x0 > -CURLING_INNER_X) && (x0 < CURLING_INNER_X) && (y0 < CURLING_INNER_Y); }
	bool wasApproaching(const DynamicStone& other) const;

	// Debugging
	bool isValid(void) const
	{
		return TEA::Memory::isValidRead(this) && (t0 >= 0);
	}	
};

struct DynamicCurver
{
	// A stone on a curved path
	surreal cx;					// The x-ordinate of the centre of the circular arc
	surreal cy;					// The y-ordinate of the centre of the circular arc
	surreal cr;					// The radius of the circular arc

	// Methods
	surreal hitSide(surreal sx) const;
	void straighten(const DynamicStone& before, surreal x, surreal y, DynamicStone& after) const;

	// Debugging
	bool isValid(void) const
	{
		return TEA::Memory::isValidRead(this);
	}	
};

struct DynamicWorld
{
	enum Event
	{
		NoEvent = 0,
		HitNothing,
		HitLimit,
		HitStone,
		HitBoard,
		HitStraighten
	};

	// Game world in motion
	TEA::System::Timer start;				// The timer value for the start of the throw
	surreal timeshift;						// The world time shift for quick cameras
	surreal time;							// The world time (0 = release from hack)
	int stones;								// The number of stones in play
	DynamicStone stone[GameMaxStones];		// The list of stones in order of delivery
	int curver;								// The index of the stone following a curved path or -1
	DynamicCurver curve;					// The parameters of the curved path
	DynamicCamera camera;					// The dynamic positioning of the camera
	Event event;							// The event that started this time segment
	real knock_speed;						// The speed of the latest knock
	int hitters[2];							// The indices of the two stones that hit

	// Methods
	static void spinSolve(real x, real y, real spin, real& power, real& aim);
	void reset(const GameWorld& world, int thrown, real spin, real power, real aim);
	bool updateTime(void);
	void updateView(GameView& view, GameSound& sound, DynamicWorld& upcoming, int thrown);
	void nextEvent(DynamicWorld& result) const;
	surreal hitLimitTime(surreal x, surreal t0, surreal x0, surreal vx0, surreal speed0) const;
	static surreal hitStones(surreal t1, surreal t2, const DynamicStone& a0, surreal sa, const DynamicStone& b0, surreal sb, surreal& knock_time);
	static real knockStones(surreal knock_time, surreal start_time, DynamicStone& a, DynamicStone& b);

	// Debugging
	bool isValid(void) const
	{
		for (int i = 0; i < stones; i++)
			if (!stone[i].isValid())
				return false;
		return TEA::Memory::isValidRead(this) && curve.isValid() && camera.isValid() &&
			   (hitters[0] >= 0) && (hitters[1] >= hitters[0]) && (hitters[0] < stones) &&
			   (stones >= 0) && (stones <= dimof(stone));
	}
};

#endif
