//----------------------------------------------------------------------------
//	SOURCE:	Game.cpp
//	AUTHOR:	Ian Taylor
//	NOTES:	Curling game class
//----------------------------------------------------------------------------

#include "Curling.h"
#include "App.h"
#include "Game.h"
#include "Dynamic.h"
#include "Config.h"
#include "Dialog.h"
#include "Resource.h"

TEA_NAMESPACE_USE

void GameWorld::sanitise(void)
{
	MONITOR(GameWorld::sanitise)

	// Remove stones that are out of play
	VALIDATE(*this);
	int i, j = 0;
	for (i = 0; i < stones; i++)
		if (stone[i].isInPlay())
		{
			if (i != j)
				stone[j] = stone[i];
			j++;
		}
	stones = j;

	// Sanitise all the stones so they aren't touching
	bool nudged;
	do
	{
		nudged = false;
		for (i = 1; i < stones; i++)
			for (j = 0; j < i; j++)
			{
				// If they're touch (or worse) move them a millimeter apart
				static const surreal nudge = CURLING_STONE_DIAMETER + CURLING_NUDGE;
				surreal dx = stone[i].x - stone[j].x;
				surreal dy = stone[i].y - stone[j].y;
				surreal dxy = Math::sqr(dx) + Math::sqr(dy);
				if (dxy < Math::sqr(nudge))
				{
					// Nudge the two stones apart
					dxy = Math::max(nudge / Math::sqrt(dxy) - 1, CURLING_NUDGE) / 2;
					ASSERT(dxy > 0);
					stone[i].nudge((real)(dx * dxy), (real)(dy * dxy));
					stone[j].nudge((real)(dx * -dxy), (real)(dy * -dxy));
					nudged = true;
				}
				ASSERT(Math::hypot(stone[i].x - stone[j].x, stone[i].y - stone[j].y) > REAL(CURLING_STONE_DIAMETER));
			}
	}
	while (nudged);
	return;
}

int GameWorld::scoredBy(GameStone::Colour colour) const
{
	MONITOR(GameWorld::scoredBy)

	// Work out the scoring for this colour
	VALIDATE(*this);
	ASSERT((colour >= 0) || (colour < 2));
	surreal s[GameMaxStones], closest[2] =
	{
		CURLING_HOUSE_R * CURLING_HOUSE_R,
		CURLING_HOUSE_R * CURLING_HOUSE_R
	};
	int i, found = 0;
	for (i = 0; i < stones; i++)
	{
		// Find the closest red and yellow stones
		s[i] = Math::sqr(stone[i].x) + Math::sqr(stone[i].y);
		if (s[i] < closest[stone[i].colour])
			closest[stone[i].colour] = s[i];
	}
	if (closest[colour] < closest[1 - colour])
	{
		// We've got a positive score
		for (i = 0; i < stones; i++)
			if ((stone[i].colour == colour) && (s[i] < closest[1 - colour]))
				found++;
	}
	else
	{
		// We've got a negative score
		for (i = 0; i < stones; i++)
			if ((stone[i].colour != colour) && (s[i] <= closest[colour]))
				found--;
	}
	return found;
}

void GameStone::nudge(real dx, real dy)
{
	MONITOR(GameStone::nudge)

	// Nudge a stone, but don't let it leave play
	VALIDATE(*this);
	ASSERT(isInPlay());
	if (isInPlay(x + dx, y + dy))
	{
		// Simple nudge
		x += dx;
		y += dy;
	}
	else
	{
		// Only nudge towards the centre of the house
		if ((Math::sgn(x) * Math::sgn(dx)) < 0)
			x += dx;
		if ((Math::sgn(y) * Math::sgn(dy)) < 0)
			y += dy;
	}
	ASSERT(isInPlay());
	return;
}

void GameCamera::toWorld(real& x, real& y) const
{
	MONITOR(GameCamera::toWorld)

	// Convert windows coords (x,[-1,1]) to world coords, assuming looking straight down
	VALIDATE(*this);
	ASSERT((y >= -1) && (y < 1));
	ASSERT(px == 0);
	ASSERT(pz > 0);
	ASSERT(lx == 0);
	ASSERT(ly == py);
	ASSERT(lz < pz);
	real scale = pz / REAL(CURLING_COT_FOV_BY_2);
	x = x * scale;
	y = py - y * scale;
	return;
}

void GameCamera::lookDown(real ymin, real ymax)
{
	MONITOR(GameCamera::lookDown)

	// Configure the camera to look straight down with the extents in view (plus margin)
	VALIDATE(*this);
	ASSERT(ymax > ymin);
	Memory::zero(*this);
	py = ly = (ymax + ymin) / 2;
	pz = lookDownZ(ymin, ymax);
	return;
}

void GameSound::updateSound(Sound& sound) const
{
	MONITOR(GameSound::updateSound)

	// Update the sounds of the dynamic world, if necessary
	VALIDATE(*this);
	VALIDATE(sound);
	int speed;
	if (knock_speed > 0)
	{
		const real max_event_speed = Math::sqrt(REAL(2 * CURLING_DECELERATION * (CURLING_POWER_MAX + CURLING_RINK_RELEASE - CURLING_RINK_STRAIGHTEN)));
		speed = Math::round(knock_speed / max_event_speed * 127);
		ASSERT((speed >= 0) && (speed <= 127));
		if (speed > 0)
			VERIFY(sound.knock(speed));
	}
	const real max_start_speed = Math::sqrt(REAL(2 * CURLING_DECELERATION * CURLING_POWER_MAX));
	speed = Math::roundUp(fastest_stone / max_start_speed * 127);
	ASSERT((speed >= 0) && (speed <= 127));
	VERIFY(sound.speed(speed));
	return;
}

void GameComputer::setRandom(real skill, TEA::Random& random)
{
	MONITOR(GameComputer::setRandom)

	// Set a computer player to random attributes much confined to an overall skill
	if (skill < Math::Epsilon)
	{
		// No skill whatsoever
		accuracy = 0;
		tactics = 0;
		strategy = 0;
	}
	else
	{
		// Start with uniform randoms
		accuracy = random.uniform();
		tactics = random.uniform();
		strategy = random.uniform();
		real scale = (accuracy + tactics + strategy) / (skill * 3);
		if (scale >= 1)
		{
			// Scale the values towards zero
			accuracy /= scale;
			tactics /= scale;
			strategy /= scale;
		}
		else
		{
			// Scale the values towards one
			scale = (3 - skill * 3) / (3 - accuracy - tactics - strategy);
			ASSERT(scale < 1);
			accuracy = 1 - (1 - accuracy) * scale;
			tactics = 1 - (1 - tactics) * scale;
			strategy = 1 - (1 - strategy) * scale;
		}
	}
	ASSERT(Math::isNear(accuracy + tactics + strategy, skill * 3));
	VALIDATE(*this);
	return;
}

bool Game::freeGuardViolation(void) const
{
	MONITOR(Game::freeGuardViolation)

	// See if the FGZ rule is violated
	VALIDATE(*this);
	ASSERT(thrown >= view.world.stones);
	if (thrown >= config->game.guards)
		return false;

	// Check for any stones that were in the FGZ but now out of play
	for (int i = 0; i < previous.stones; i++)
		if (previous.stone[i].isInFreeGuardZone() && !view.world.stone[i].isInPlay())
			return true;
	return false;
}

void Game::newGame(const Configuration& cfg, Random& rng)
{
	MONITOR(Game::newGame)

	// Initialise a new game
	Memory::zero(*this);
	config = &cfg;
	random = &rng;
	for (int p = 0; p < 2; p++)
	{
		// Initialise the two players
		total[p] = cfg.player[p].handicap;
		if (!cfg.player[p].human)
		{
			// The player is a computer, so we need its attributes
			computer[p].me = static_cast<GameComputer::Colour>(p);
			computer[p].config = config;
			if (cfg.player[p].expert)
			{
				// Copy them across
				computer[p].accuracy = cfg.player[p].accuracy;
				computer[p].tactics = cfg.player[p].tactics;
				computer[p].strategy = cfg.player[p].strategy;
			}
			else
			{
				// Calculate random attributes
				computer[p].setRandom(cfg.player[p].skill, rng);
			}
		}
	}
	VALIDATE(*this);

	// Position the camera to show the required extents
	view.camera.lookDown(CURLING_ZOOM_MIN, CURLING_ZOOM_MAX);
	return;
}

void Game::updateControls(DialogMain& main, GameStone::Colour colour) const
{
	MONITOR(Game::updateControls)

	// Update the dialog based on the game state
	VALIDATE(*this);
	VALIDATE(main);
	if (thrown == 0)
	{
		// This is a new end
		if (end == 0)
		{
			// This is a new game
			main.setText(IDC_PLAYER1, config->player[0].name);
			main.setText(IDC_PLAYER2, config->player[1].name);
		}
		char buffer[256];
		char* dst = buffer;
		size_t size = dimof(buffer);
		char number[16];
		Format::concat(dst, size, "End ");
		Format::concat(dst, size, Format::fromUnsigned(end + 1, number, dimof(number)));
		Format::concat(dst, size, " of ");
		Format::concat(dst, size, Format::fromUnsigned(config->game.ends, number, dimof(number)));
		main.setText(IDC_END, buffer);
	}

	// Work out who the current player is
	main.setVisible(IDC_ICON1, colour == GameStone::Red);
	main.setVisible(IDC_ICON2, colour == GameStone::Yellow);
	main.setInteger(IDC_LEFT1, config->game.stones - (thrown + 1 - first) / 2);
	main.setInteger(IDC_LEFT2, config->game.stones - (thrown + first) / 2);
	main.setInteger(IDC_SCORING1, scoring[0], "-");
	main.setInteger(IDC_SCORING2, scoring[1], "-");
	main.setInteger(IDC_TOTAL1, total[0]);
	main.setInteger(IDC_TOTAL2, total[1]);
	return;
}

void Game::updateMessage(DialogMain& main, int player, const char* suffix) const
{
	MONITOR(Game::updateMessage)

	// Update the dialog message
	VALIDATE(*this);
	VALIDATE(main);
	ASSERT((player == 0) || (player == 1)); 
	ASSERT(suffix != NULL);
	char buffer[128];
	char* dst = buffer;
	size_t size = dimof(buffer);
	VERIFY(TEA::Format::concat(dst, size, config->player[player].name));
	VERIFY(TEA::Format::concat(dst, size, suffix));
	main.setText(IDC_MESSAGE, buffer);
	return;
}

bool Game::onIdle(void)
{
	MONITOR(Game::onIdle)

	// Work out if things are moving
	if ((dynamic == NULL) || !dynamic->updateTime())
		return false;
	ASSERT(upcoming != NULL);
	dynamic->updateView(view, sound, *upcoming, view.world.stones - 1);
	return true;
}

void Game::onNewStone(GameStone::Colour colour, Sound& sound)
{
	MONITOR(Game::onNewStone)

	// Play an appropriate sounds
	if (config->player[colour].human)
	{
		// Play a bell
		int score = scoring[colour];
		if (score < 0)
			score = score * 2 + 1;
		else if (score > 0)
			score = score * 2 + 12;
		VERIFY(sound.bell(score));
	}
	return;
}

void Game::onClick(DialogMain& main, real x, real y)
{
	MONITOR(Game::onClick)

	// A human player has clicked on the canvas to aim the stone
	if (!config->player[colour()].expert)
	{
		// This is a novice, so set the aiming parameters
		real power, aim;
		DynamicWorld::spinSolve(x, y, main.getSpin(), power, aim);
		main.setPower(power);
		main.setAim(aim);
	}
	main.setFocus(IDC_ZOOM);
	return;
}

void Game::onThrow(DialogMain& main)
{
	MONITOR(Game::onThrow)

	// Only important if not in motion
	VALIDATE(*this);
	ASSERT(dynamic == NULL);
	real spin = main.getSpin();
	real power = main.getPower();
	real aim = main.getAim();
	const char* what = " throws";
	real qy = REAL(CURLING_RINK_RELEASE) + power;
	if (qy < REAL(-CURLING_RINK_BACK))
		what = " throws a guard";
	else if (qy > REAL(CURLING_RINK_BACK))
		what = " throws with weight";
	updateMessage(main, player(), what);
	if (config->game.random_ice)
		applyInaccuracy(spin, power, aim, REAL(CURLING_RANDOM_ICE));
	onThrow(spin, power, aim);
	ASSERT(dynamic != NULL);
	ASSERT(upcoming != NULL);
	if (main.getMenuCheck(IDM_OPTIONS_QUICK))
	{
		// Work out how much to timeshift the quick camera
		surreal qt = upcoming->time;
		if (upcoming->event == DynamicWorld::HitStraighten)
		{
			// Go to the next event
			DynamicWorld after;
			upcoming->nextEvent(after);
			ASSERT(after.time > qt);
			qt = after.time;
			if (after.event == DynamicWorld::HitNothing)
				qt -= CURLING_QUICK_CAMERA;
		}
		qt -= CURLING_QUICK_CAMERA;
		dynamic->timeshift = qt;
		upcoming->timeshift = qt;
		VERIFY(dynamic->updateTime());
	}
	dynamic->updateView(view, sound, *upcoming, view.world.stones - 1);
	return;
}

void Game::onThrow(real spin, real power, real aim)
{
	MONITOR(Game::onThrow)

	// Throw a new stone
	VALIDATE(*this);
	previous = view.world;
	int thrown = view.world.stones++;
	GameStone& new_stone = view.world.stone[thrown];
	new_stone.colour = colour();
	new_stone.x = 0;
	new_stone.y = REAL(CURLING_RINK_RELEASE);
	new_stone.rot = 0;

	// Reset the dynamic world (WARNING: Singletons)
	ASSERT(dynamic == NULL);
	static DynamicWorld s_dynamic, s_upcoming;
	s_dynamic.reset(view.world, thrown, spin, power, aim);
	s_dynamic.start = System::timerGet();
	dynamic = &s_dynamic;

	// Set the camera parameters
	real cx, cz, cr = random->uniform(REAL(1.5), REAL(2.5));
	Math::sincos(random->uniform(REAL(0.1), REAL(0.6)), cz, cx);
	cx *= (colour() == GameStone::Red) ? cr : -cr;
	cz *= cr;
	view.camera.px = cx;
	view.camera.py = REAL(CURLING_RINK_HACK);
	view.camera.pz = cz;
	view.camera.lx = 0;
	view.camera.ly = REAL(CURLING_RINK_RELEASE);
	view.camera.lz = 0;
	surreal cv, ct, cy = Math::min(power + CURLING_RINK_RELEASE, 0.0);
	s_dynamic.stone[thrown].releaseSpeedTime(cy, cv, ct);
	if (config->quick)
		ct -= REAL(CURLING_QUICK_CAMERA * 2);
	s_dynamic.camera.aho = Math::Pi / Math::power(ct, CURLING_CAMERA_LAG);
	s_dynamic.camera.sethi = cx / 2;
	s_dynamic.camera.ullman = cz;

	// Now find the next upcoming event
	s_dynamic.nextEvent(s_upcoming);
	upcoming = &s_upcoming;
	return;
}

void Game::onMotionComplete(DialogMain& main, Sound& sound)
{
	MONITOR(Game::onMotionComplete)

	// All stones and the camera have stopped moving
	VALIDATE(*this);
	VALIDATE(main);
	ASSERT(!view.motion);
	dynamic = upcoming = NULL;
	thrown++;
	if (freeGuardViolation())
	{
		// Need to restore the previous positions
		if (config->player[0].human || config->player[1].human)
			main.showViolation();
		view.world = previous;
	}

	// Sanitise the stones and count the closest stones
	view.world.sanitise();
	int reds = view.world.scoredBy(GameStone::Red);
	scoring[0] = Math::max(reds, 0);
	scoring[1] = Math::max(-reds, 0);

	// Re-enable the zoom
	main.setSlider(IDC_ZOOM, view.camera.zoomFactor(view.camera.pz));
	main.setEnable(IDC_ZOOM, true);
	main.setFocus(IDC_ZOOM);

	// Work out which colour throws next
	if (thrown < (config->game.stones * 2))
	{
		// There's more stones in this end
		updateControls(main, colour());
		onNewStone(colour(), sound);
		main.wakeup(0);
	}
	else
	{
		// That's the end of this end
		if (scoring[0] > scoring[1])
			updateMessage(main, 0, " wins the end");
		else if (scoring[1] > scoring[0])
			updateMessage(main, 1, " wins the end");
		else
			main.setText(IDC_MESSAGE, "The end is blanked");
		updateControls(main, GameStone::Neither);
		VERIFY(sound.orchestra());
		main.wakeup(5000);
	}
	return;
}

void Game::applyInaccuracy(real&, real& power, real& aim, real inaccuracy)
{
	MONITOR(Game::nextStone)

	// Apply an inaccuracy to throw parameters
	VALIDATE(*this);
	ASSERT(inaccuracy > 0);
	real ex, ey, er = inaccuracy * random->uniform();
	Math::sincos(random->uniform() * Math::Pi * 2, ex, ey);
	power += ey * er;
	aim += Math::atan(ex * er, -REAL(CURLING_RINK_RELEASE));
	return;
}

bool Game::nextStone(DialogMain& main)
{
	MONITOR(Game::nextStone)

	// Start the next stone
	VALIDATE(*this);
	VALIDATE(main);
	if (thrown >= (config->game.stones * 2))
	{
		// Reached the end of the end
		if (!nextEnd(main))
			return false;
	}
	if (config->player[colour()].human)
		goHuman(main);
	else
		goComputer(main);
	return true;
}

bool Game::nextEnd(DialogMain& main)
{
	MONITOR(Game::nextEnd)

	// Start the next 
	VALIDATE(*this);
	VALIDATE(main);
	total[0] += scoring[0];
	total[1] += scoring[1];
	if (scoring[0] < scoring[1])
		first = 1;
	else if (scoring[0] > scoring[1])
		first = 0;
	scoring[0] = 0;
	scoring[1] = 0;
	if (((end + 1) >= config->game.ends) && (total[0] != total[1]))
	{
		// The game has finished
		updateMessage(main, (total[0] > total[1]) ? 0 : 1, " wins the game");
		updateControls(main, GameStone::Neither);
		return false;
	}

	// The next end should start
	thrown = 0;
	end++;
	Memory::zero(view.world);
	updateControls(main, colour());
	return true;
}

void Game::goHuman(DialogMain& main)
{
	MONITOR(Game::goHuman)

	// A human is to throw the next stone
	VALIDATE(*this);
	VALIDATE(main);
	updateMessage(main, player(), " to throw next...");
	main.resetSliders();
	if (!config->player[colour()].expert)
		main.setEnable(IDC_CANVAS, true);
	main.setEnable(IDC_SPIN, true);
	main.setEnable(IDC_POWER, true);
	main.setEnable(IDC_AIM, true);
	return;
}

void Game::goComputer(DialogMain& main)
{
	MONITOR(Game::nextEnd)

	// The computer is to throw the next stone
	VALIDATE(*this);
	VALIDATE(main);
	main.setEnable(IDC_ZOOM, false);
	updateMessage(main, player(), " is thinking...");
	const GameComputer& brain = computer[colour()];
	real spin, power, aim, inaccuracy = 1 - brain.accuracy;
	brain.choose(*this, spin, power, aim);
	ASSERT((spin >= REAL(CURLING_SPIN_MIN)) && (spin <= REAL(CURLING_SPIN_MAX)));
	ASSERT((power >= REAL(CURLING_POWER_MIN)) && (power <= REAL(CURLING_POWER_MAX)));
	ASSERT((aim >= REAL(CURLING_AIM_MIN)) && (aim <= REAL(CURLING_AIM_MAX)));
	if (inaccuracy > 0)
		applyInaccuracy(spin, power, aim, inaccuracy * REAL(CURLING_INACCURACY_MAX));
	main.setSpin(spin);
	main.setPower(power);
	main.setAim(aim);
	onThrow(main);
	return;
}

bool Game::isValid(void) const
{
	// Exhaustive validation
	if (!Memory::isValidRead(this) ||
		!view.isValid() || !sound.isValid() || !previous.isValid() ||
		(end < 0) ||
		(thrown < 0) || (thrown > GameMaxStones) ||
		((first & ~1) != 0) ||
		(scoring[0] < 0) || (scoring[1] < 0) || ((scoring[0] * scoring[1]) != 0) ||
		(total[0] < 0) || (total[1] < 0) ||
		!computer[0].isValid() || !computer[1].isValid())
		return false;
	if ((config != NULL) && !config->isValid())
		return false;
	if ((random != NULL) && !random->isValid())
		return false;
	if ((dynamic != NULL) && !dynamic->isValid())
		return false;
	if ((upcoming != NULL) && !upcoming->isValid())
		return false;
	return true;
}
