//----------------------------------------------------------------------------
//	HEADER:	Game.h
//	AUTHOR:	Ian Taylor
//	NOTES:	Curling game class
//----------------------------------------------------------------------------

#ifndef TEA_GAME_H
#define TEA_GAME_H
#ifdef TEA_DEBUG_INCLUDES
#pragma message("#include " __FILE__)
#endif

struct Game;
struct Configuration;
struct DynamicWorld;
class DialogMain;
class Application;
class Sound;

enum GameConstants
{
	GameMaxStones = 8 * 2
};

// Derived constants
#define CURLING_ZOOM_MIN	REAL(CURLING_RINK_HOG)
#define CURLING_ZOOM_MAX	REAL(CURLING_RINK_BACK)
#define CURLING_INNER_X		REAL((CURLING_RINK_WIDTH - CURLING_STONE_DIAMETER) / 2)
#define CURLING_INNER_Y		REAL(CURLING_RINK_BACK + CURLING_STONE_DIAMETER / 2)
#define CURLING_OUTER_X		REAL((CURLING_RINK_WIDTH - CURLING_STONE_DIAMETER) / 2 + CURLING_GUTTER_WIDTH)
#define CURLING_OUTER_Y		REAL(CURLING_RINK_END - CURLING_STONE_DIAMETER / 2)
#define CURLING_HOG_Y		REAL(CURLING_RINK_HOG + CURLING_STONE_DIAMETER / 2)
#define CURLING_HOUSE_R		REAL(CURLING_RINK_BACK + CURLING_STONE_DIAMETER / 2)

struct GameCamera
{
	// Camera position
	real px, py, pz;					// The camera's coordinate
	real lx, ly, lz;					// Where the camera is looking

	// Methods
	void toWorld(real& x, real& y) const;
	void lookDown(real ymin, real ymax);
	static real lookDownZ(real ymin, real ymax)
	{
		// Calculate the camera height to look down and include extents
		MONITOR(GameCamera::lookDownZ)
		ASSERT(ymax > ymin);
		return (ymax - ymin) * REAL(CURLING_COT_FOV_BY_2 * 0.6);
	}
	static real zoomFactor(real height)
	{
		// Calculate the zoom factor based on the camera height
		MONITOR(GameCamera::zoomFactor)
		ASSERT(height > 0);
		real ymin = CURLING_ZOOM_MAX - height / REAL(CURLING_COT_FOV_BY_2 * 0.6);
		ASSERT(ymin >= CURLING_ZOOM_MIN);
		real zoom = (CURLING_ZOOM_MIN - ymin) / (CURLING_ZOOM_MIN + CURLING_ZOOM_MAX);
		return zoom;
	}

	// Debugging
	bool isValid(void) const
	{
		return TEA::Memory::isValidRead(this) && (lz == 0);
	}
};

struct GameStone
{
	// Stone position
	enum Colour
	{
		Red = 0,
		Yellow = 1,
		Neither = 2
	}
	colour;
	real x;								// The stone's x-ordinate
	real y;								// The stone's y-ordinate
	real rot;							// The stone's rotation in radians

	// Methods
	void nudge(real dx, real dy);

	// Properties
	static bool isInPlay(real x, real y) { return (x > -CURLING_INNER_X) && (x < CURLING_INNER_X) && (y > CURLING_HOG_Y) && (y < CURLING_INNER_Y); }
	static bool isInHouse(real x, real y) { return (x * x + y * y) < (CURLING_HOUSE_R * CURLING_HOUSE_R); }
	static bool isInFreeGuardZone(real x, real y) { return (x > -CURLING_INNER_X) && (x < CURLING_INNER_X) && (y > CURLING_HOG_Y) && (y < REAL(CURLING_STONE_DIAMETER / -2)) && !isInHouse(x, y); }
	bool isInPlay(void) const { return isInPlay(x, y); }
	bool isInHouse(void) const { return isInHouse(x, y); }
	bool isInFreeGuardZone(void) const { return isInFreeGuardZone(x, y); }

	// Debugging
	bool isValid(void) const
	{
		return TEA::Memory::isValidRead(this);
	}
};

struct GameWorld
{
	// Game world
	int stones;							// The number of stones to be displayed
	GameStone stone[GameMaxStones];		// The complete list of stones in order of delivery

	// Methods
	void sanitise(void);

	// Properties
	int scoredBy(GameStone::Colour colour) const;

	// Debugging
	bool isValid(void) const
	{
		for (int i = 0; i < stones; i++)
			if (!stone[i].isValid())
				return false;
		return TEA::Memory::isValidRead(this) &&
			   (stones >= 0) && (stones <= dimof(stone));
	}
};

struct GameView
{
	// Game view
	GameCamera camera;					// The camera information
	GameWorld world;					// The world information
	bool motion;						// Either the camera or the world is in motion

	// Debugging
	bool isValid(void) const
	{
		return TEA::Memory::isValidRead(this) && camera.isValid() && world.isValid();
	}
};

struct GameSound
{
	real fastest_stone;					// The speed of the fastest stone
	real knock_speed;					// The speed of the latest knock

	// Methods
	void updateSound(Sound& sound) const;

	// Debugging
	bool isValid(void) const
	{
		return TEA::Memory::isValidRead(this) && (fastest_stone >= 0) && (knock_speed >= 0);
	}
};

struct GameComputer
{
	// Colours are colours are colours
	typedef GameStone::Colour Colour;

	// Game computer player
	Colour me;
	real accuracy;
	real tactics;
	real strategy;
	const Configuration* config;

	// Methods
	void setRandom(real skill, TEA::Random& random);
	void choose(const Game& game, real& spin, real& power, real& aim) const;

	// Debugging
	bool isValid(void) const
	{
		return TEA::Memory::isValidRead(this) &&
			   (accuracy >= 0) && (accuracy <= 1) &&
			   (tactics >= 0) && (tactics <= 1) &&
			   (strategy >= 0) && (strategy <= 1);
	}
};

struct Game
{
	// Game state
	const Configuration* config;	// The current configuration
	TEA::Random* random;			// The random number generator for this game
	GameView view;					// The current view
	GameSound sound;				// The current sounds
	GameWorld previous;				// The previous world
	int end;						// The current end [0..N-1]
	int first;						// Which player [0..1] went first
	int thrown;						// The number of stones thrown (including those removed)
	int scoring[2];					// The number of stones scoring in this end for each player
	int total[2];					// The total score of each player
	GameComputer computer[2];		// The configuration of computer players
	DynamicWorld* dynamic;			// The world in motion
	DynamicWorld* upcoming;			// The next event to occur

	// Methods
	void newGame(const Configuration& cfg, TEA::Random& rng);
	bool nextStone(DialogMain& main);
	bool nextEnd(DialogMain& main);
	void goHuman(DialogMain& main);
	void goComputer(DialogMain& main);
	void applyInaccuracy(real& spin, real& power, real& aim, real inaccuracy);
	void updateControls(DialogMain& main, GameStone::Colour colour) const;
	void updateMessage(DialogMain& main, int player, const char* suffix) const;

	// Message handlers
	bool onIdle(void);
	void onNewStone(GameStone::Colour colour, Sound& sound);
	void onClick(DialogMain& main, real x, real y);
	void onThrow(DialogMain& main);
	void onThrow(real spin, real power, real aim);
	void onMotionComplete(DialogMain& main, Sound& sound);

	// Properties
	int player(void) const { return (first ^ thrown) & 1; }
	GameStone::Colour colour(void) const { return static_cast<GameStone::Colour>(player()); }
	bool freeGuardViolation(void) const;

	// Debugging
	bool isValid(void) const;
};

#endif
