//----------------------------------------------------------------------------
//	SOURCE:	Generate.cpp
//	AUTHOR:	Ian Taylor
//	NOTES:	Curling stone model generation
//----------------------------------------------------------------------------

#include "Curling.h"

// Only in debug builds
#if TEA_DEBUG_LEVEL > 0

TEA_NAMESPACE_USE

// Automatically generate the shadow and stone definitions
static struct GenerateModel
{
	GenerateModel(void);
	static void rotate(real sin, real cos, real& x, real& y)
	{
		real ox = x;
		real oy = y;
		x = ox * cos - oy * sin;
		y = oy * cos + ox * sin;
	}
	static void newline(void) { System::debugOutput("\r\n"); System::debugFlush(); }
	static void output(const char* message) { System::debugFormat("%s\r\n", message); }
	void output(const char* prefix, real values[], int count, const char* suffix);
	static void outputShadow(int shade) { System::debugFormat("MODEL_SHADE(%d)\r\n", shade); }
	void outputCoord1(real x, real y, real z, real nx = 0, real ny = 0, real nz = 1) { real xyz[6] = { x, y, z, nx, ny, nz }; output("MODEL_COORD(", xyz, 6, ")"); }
	void outputCoords(real cx, real cy, real cz, real radius, real dz, int rot, int slices, int twist);
	void outputIndex1(int offset) { System::debugFormat("%3d,", vertices + offset); }
	void outputIndices(bool join, int slices, bool clockwise);
	int vertices;
}
_model;

GenerateModel::GenerateModel(void)
{
	// Generate the shadow and stone models
	const int slices1 = 16;
	const int slices2 = 8;
	vertices = 0;

	// Shadow	 cx		 cy		 cz		 radius	  dz	  rot	 n	 twist
	output("// Shadow vertices");
	outputCoords(0.000f, 0.000f, 0.000f, 1.000f,  1.000f,  0, slices1, 1);
	outputShadow(30);
	outputCoords(0.000f, 0.000f, 0.000f, 0.600f,  1.000f,  0, slices1, 2);
	int phase1 = vertices;

	// Stone	 cx		 cy		 cz		 radius	  dz	  rot	 n	 twist
	output("// Stone vertices");
	outputCoords(0.000f, 0.000f, 0.000f, 0.600f, -1.000f,  0, slices1, 2);
	outputCoords(0.000f, 0.000f, 0.054f, 0.800f, -0.866f,  0, slices1, 3);
	outputCoords(0.000f, 0.000f, 0.200f, 0.946f, -0.500f,  0, slices1, 2);
	outputCoords(0.000f, 0.000f, 0.400f, 1.000f,  0.000f,  0, slices1, 3);
	outputCoords(0.000f, 0.000f, 0.600f, 1.000f,  0.000f,  0, slices1, 2);
	outputCoords(0.000f, 0.000f, 0.800f, 0.946f,  0.500f,  0, slices1, 3);
	outputCoords(0.000f, 0.000f, 0.946f, 0.800f,  0.866f,  0, slices1, 2);
	outputShadow(100);
	outputCoords(0.000f, 0.000f, 0.946f, 0.800f,  0.866f,  0, slices1, 2);
	outputCoords(0.000f, 0.000f, 1.000f, 0.600f,  1.000f,  0, slices1, 3);

	// Fiddly top bit (see graph paper!)
	const real z = 1.000f;
	outputCoord1(-0.185f,  0.477f, z);
	outputCoord1(-0.185f,  0.323f, z);
	outputCoord1(-0.100f,  0.215f, z);
	outputCoord1(-0.100f,  0.000f, z);
	outputCoord1(-0.100f, -0.200f, z);
	outputCoord1(-0.100f, -0.400f, z);
	outputCoord1(-0.071f, -0.571f, z);
	outputCoord1( 0.071f, -0.571f, z);
	outputCoord1( 0.100f, -0.400f, z);
	outputCoord1( 0.100f, -0.200f, z);
	outputCoord1( 0.100f,  0.000f, z);
	outputCoord1( 0.100f,  0.215f, z);
	outputCoord1( 0.185f,  0.323f, z);
	outputCoord1( 0.185f,  0.477f, z);
	outputShadow(60);
	outputCoord1( 0.000f,  0.215f, z);
	outputCoord1( 0.000f,  0.100f, z);
	outputCoord1( 0.000f, -0.100f, z);
	outputCoord1( 0.000f, -0.300f, z);
	outputCoord1( 0.000f, -0.500f, z);
	outputShadow(100);

	// Handle	 cx		  cy	  cz	  radius  dz	  rot	  n	   twist
	outputCoords(0.000f,  0.400f, 1.029f, 0.129f, 0.707f,   0, slices2, -2);
	outputCoords(0.000f,  0.400f, 1.100f, 0.100f, 0.000f,   0, slices2, -1);
	outputCoords(0.000f,  0.373f, 1.200f, 0.100f, 0.000f, -30, slices2, -2);
	outputCoords(0.000f,  0.300f, 1.273f, 0.100f, 0.000f, -60, slices2, -1);
	outputCoords(0.000f,  0.200f, 1.300f, 0.100f, 0.000f, -90, slices2, -2);
	outputCoords(0.000f, -0.100f, 1.300f, 0.100f, 0.000f, -90, slices2, -1);
	outputCoords(0.000f, -0.400f, 1.300f, 0.100f, 0.000f, -90, slices2, -2);
	outputCoords(0.000f, -0.450f, 1.300f, 0.087f, 0.500f, -90, slices2, -1);
	outputCoords(0.000f, -0.487f, 1.300f, 0.050f, 0.866f, -90, slices2, -2);
	outputCoord1(0.000f, -0.500f, 1.300f, 0.000f, -1.000f, 0.000f);

	// The number of raw vertices
	int total = vertices * 2 - phase1;
	System::debugFormat("vertices = %d\r\n", total);

	// Generate the indices for the shadow
	output("// Shadow indices");
	vertices = 0;
	outputIndices(false, slices1, false);

	// Generate the indices for the stone
	output("// Stone indices");
	vertices = 0;
	outputIndices(false, slices1, false);
	outputIndices(true, slices1, true);
	outputIndices(true, slices1, false);
	outputIndices(true, slices1, true);
	outputIndices(true, slices1, false);
	outputIndices(true, slices1, true);
	vertices += slices1;
	outputIndices(true, slices1, false);

	// Fiddly top bit
	const int r0 = 0;
	const int k = r0 + slices1;
	const int c = k + 14;
	const int r1 = c + 5;
	int i;
	outputIndex1(r0 + 0);
	outputIndex1(r0 + slices1 - 1);
	for (i = 0; i < 14; i++)
	{
		outputIndex1(r0 + i);
		outputIndex1(k + i);
	}
	newline();

	// Where the handle joins the top
	outputIndex1(r0 + i);
	outputIndex1(r1);
	outputIndex1(r0 + i);
	outputIndex1(r1 + 1);
	outputIndex1(r0 + slices1 - 1);
	outputIndex1(r1 + 2);
	outputIndex1(k + 0);
	outputIndex1(r1 + 3);
	outputIndex1(k + 1);
	outputIndex1(r1 + 4);
	outputIndex1(k + 2);
	outputIndex1(r1 + 5);
	outputIndex1(c + 0);

	// Shadow under the handle
	for (i = 0; i <= 4; i++)
	{
		outputIndex1(c + i);
		outputIndex1(k + 2 + i);
	}
	outputIndex1(k + 7);
	for (i = 0; i <= 4; i++)
	{
		outputIndex1(k + 7 + i);
		outputIndex1(c + 4 - i);
	}

	// The rest of where the handle meets the top
	outputIndex1(r1 + 5);
	outputIndex1(r1 + 5);
	outputIndex1(k + 11);
	outputIndex1(r1 + 6);
	outputIndex1(k + 12);
	outputIndex1(r1 + 7);
	outputIndex1(k + 13);
	outputIndex1(r1 + 0);
	newline();

	// The handle itself
	vertices += r1;
	outputIndices(false, slices2, false);
	outputIndices(true, slices2, true);
	outputIndices(true, slices2, false);
	outputIndices(true, slices2, true);
	outputIndices(true, slices2, false);
	outputIndices(true, slices2, true);
	outputIndices(true, slices2, false);
	outputIndices(true, slices2, true);
	outputIndex1(0);
	for (i = 0; i < slices2; i++)
	{
		outputIndex1(i);
		outputIndex1(slices2);
	}
	outputIndex1(0);
	newline();
	return;
}

void GenerateModel::output(const char* prefix, real values[], int count, const char* suffix)
{
	// Output a line of reals
	for (int i = 0; i < count; i++)
	{
		int v = Math::round(values[i] * 10000);
		int a = Math::abs(v) / 10000;
		int b = Math::abs(v) - a * 10000;
		System::debugFormat("%s%s%d.%04d", prefix, (v < 0) ? "-" : "+", a, b);
		prefix = ", ";
	}
	System::debugOutput(suffix);
	newline();
	vertices++;
	return;
}

void GenerateModel::outputCoords(real cx, real cy, real cz, real radius, real dz, int rot, int slices, int twist)
{
	// Output a circle of points
	real rsin, rcos;
	Math::sincos(Math::deg2rad(real(rot)), rsin, rcos);
	real normal = Math::sqrt(1 - dz * dz);
	for (int i = 0; i < slices; i++)
	{
		real sin, cos, theta = (Math::Pi * (i * 2 + twist)) / slices;
		Math::sincos(-theta, sin, cos);
		real value[6] =
		{
			radius * sin,	// Offset x
			radius * cos,	// Offset y
			0.0f,			// Offset z
			normal * sin,	// Normal dx
			normal * cos,	// Normal dy
			dz,				// Normal dz
		};
		rotate(rsin, rcos, value[2], value[1]);
		rotate(rsin, rcos, value[5], value[4]);
		value[0] += cx;
		value[1] += cy;
		value[2] += cz;
		output("MODEL_COORD(", value, 6, ")");
	}
	return;
}

void GenerateModel::outputIndices(bool join, int slices, bool clockwise)
{
	// Output a list of indices to form a ring
	if (join)
	{
		// Fillet with the previous ring
		outputIndex1(0);
	}
	if (clockwise)
	{
		// Clockwise winding
		outputIndex1(0);
		outputIndex1(slices);
		for (int i = slices - 1; i >= 0; i--)
		{
			outputIndex1(i);
			outputIndex1(i + slices);
		}
	}
	else
	{
		// Anticlockwise winding
		for (int i = 0; i < slices; i++)
		{
			outputIndex1(i);
			outputIndex1(i + slices);
		}
		outputIndex1(0);
		outputIndex1(slices);
	}
	newline();
	vertices += slices;
	return;
}

#endif
