//----------------------------------------------------------------------------
//	SOURCE:	Midi.cpp
//	AUTHOR:	Ian Taylor
//	NOTES:	Curling game MIDI implementation
//----------------------------------------------------------------------------

#include "Curling.h"
#include "Midi.h"

// Multimedia uses nameless unions
#include <windows.h>
#pragma warning(disable: 4201)
#include <mmsystem.h>

// Useful macro
#define VERIFY_OK(expr)		VERIFY_(expr,== MMSYSERR_NOERROR)

TEA_NAMESPACE_USE

// Device singleton data
static CRITICAL_SECTION g_critsec;
static HMIDIOUT g_device;
static HANDLE g_thread, g_event;
static volatile MidiDevice::Msg g_oob, g_queue[256];
static volatile BYTE g_head, g_tail;

static inline bool MidiDevice__playSync(MidiDevice::Msg msg)
{
	// Play a single MIDI message immediately
	for (;;)
	{
		switch (::midiOutShortMsg(g_device, msg))
		{
			case MMSYSERR_NOERROR:
				// Success
				return true;

			case MIDIERR_NOTREADY:
				// Wait before retrying
				::Sleep(10);
				break;

			default:
				// Failure
				return false;
		}
	}
}

static DWORD WINAPI MidiThread(LPVOID)
{
	// MIDI worker thread
	MidiDevice::Msg msg;
	while (::WaitForSingleObject(g_event, INFINITE) == WAIT_OBJECT_0)
	{
		// Send each message to the device
		::EnterCriticalSection(&g_critsec);
		while (g_head != g_tail)
		{
			// Send the next message
			if (g_oob)
			{
				// Out of band
				msg = g_oob;
				g_oob = 0;
			}
			else
			{
				// Direct from queue
				msg = g_queue[g_head++];
			}
			::LeaveCriticalSection(&g_critsec);
			if (!MidiDevice__playSync(msg))
				return 0;
			::EnterCriticalSection(&g_critsec);
		}
		msg = g_oob;
		g_oob = 0;
		::LeaveCriticalSection(&g_critsec);
		if (msg && !MidiDevice__playSync(msg))
			return 0;
	}
	return 1;
}

bool MidiDevice::initialise(void)
{
	MONITOR(MidiDevice::initialise)

	// Initialise the MIDI device
	::InitializeCriticalSection(&g_critsec);
	g_oob = 0;
	g_head = g_tail = 0;
	if (::midiOutOpen(&g_device, MIDI_MAPPER, 0, 0, CALLBACK_NULL) == MMSYSERR_NOERROR)
	{
		g_event = ::CreateEvent(NULL, FALSE, FALSE, NULL);
		if (g_event != NULL)
		{
			DWORD threadid;
			g_thread = ::CreateThread(NULL, 0, MidiThread, NULL, 0, &threadid);
			if (g_thread != NULL)
			{
				// Increase the priority slightly to improve responsiveness
				VERIFY(::SetThreadPriority(g_thread, THREAD_PRIORITY_HIGHEST));
				return true;
			}
			VERIFY(::CloseHandle(g_event));
		}
	}
	VERIFY_OK(::midiOutClose(g_device));
	g_device = NULL;
	return false;
}

void MidiDevice::shutdown(void)
{
	MONITOR(MidiDevice::shutdown)

	// Shutdown the MIDI services
	if (g_thread != NULL)
	{
		// Close the thread down
		VERIFY(::TerminateThread(g_thread, 0xDEAD));
		VERIFY(::CloseHandle(g_thread));
	}
	if (g_event != NULL)
	{
		// Close the event down
		VERIFY(::CloseHandle(g_event));
	}
	if (g_device != NULL)
	{
		// Shutdown MIDI
		VERIFY_OK(::midiOutReset(g_device));
		VERIFY_OK(::midiOutClose(g_device));
	}
	g_device = NULL;
	return;
}

bool MidiDevice::reset(void)
{
	MONITOR(MidiDevice::reset)

	// Quieten all voices and empty the queue
	if (g_device == NULL)
		return true;
	::EnterCriticalSection(&g_critsec);
	bool okay = (::midiOutReset(g_device) == MMSYSERR_NOERROR);
	g_oob = 0;
	g_head = g_tail = 0;
	::LeaveCriticalSection(&g_critsec);
	return okay;
}

bool MidiDevice::playAsync(const Msg* msg, int count)
{
	MONITOR(MidiDevice::playAsync)

	// Play a sequence of MIDI messages
	if (g_device == NULL)
		return true;
	ASSERT(msg != NULL);
	ASSERT(count > 0);
	ASSERT(g_thread != NULL);
	ASSERT(g_event != NULL);
	::EnterCriticalSection(&g_critsec);
	bool okay = true;
	do
	{
		// Place the remainder on the queue
		BYTE next = g_tail;
		if (++next == g_head)
		{
			// The queue is full
			okay = false;
			break;
		}
		g_queue[g_tail] = *(msg++);
		g_tail = next;
	}
	while (--count > 0);
	::LeaveCriticalSection(&g_critsec);
	VERIFY(::SetEvent(g_event));
	return okay;
}

bool MidiDevice::playSync(Msg msg)
{
	MONITOR(MidiDevice::playAsync)

	// Play a single MIDI message immediately
	if (g_device == NULL)
		return true;
	ASSERT((msg & 0x80808080) == 0x80);
	return MidiDevice__playSync(msg);
}

bool MidiDevice::playOutOfBounds(Msg msg)
{
	MONITOR(MidiDevice::playOutOfBounds)

	// Play a single MIDI message as soon as possible (queue of one)
	if (g_device == NULL)
		return true;
	ASSERT(g_thread != NULL);
	ASSERT(g_event != NULL);
	ASSERT((msg & 0x80808080) == 0x80);
	::EnterCriticalSection(&g_critsec);
	bool wakeup = (g_oob == 0);
	g_oob = msg;
	::LeaveCriticalSection(&g_critsec);
	if (wakeup)
		VERIFY(::SetEvent(g_event));
	return true;
}
