//----------------------------------------------------------------------------
//	SOURCE:	Render.cpp
//	AUTHOR:	Ian Taylor
//	NOTES:	Curling game renderer interface
//----------------------------------------------------------------------------

#include "Curling.h"
#include <windows.h>
#include "Render.h"

// Conversion to texture pixels (metres and 12-units)
#define M2TEX(m)	int((m) * CURLING_TEXTURE_SIZE / (CURLING_RINK_WIDTH * 2) + 0.5)
#define H2TEX(h)	(((h) * CURLING_TEXTURE_SIZE) / 12)

TEA_NAMESPACE_USE

// Local structures
struct Hoarding
{
	const char* text;
	const char* face;
	COLORREF fg;
	COLORREF bg;

	// Methods
	void render(HDC hdc, LOGFONT& font, RECT& rect, const char* message) const;
};

Renderer* Renderer::create(Renderer::Type type, HWND hwnd)
{
	MONITOR(Renderer::create)

	// Call the appropriate factory
	switch (type)
	{
		case DirectX7:
			// DirectX 7.0
			return createDirectX7(hwnd);
	}
	return NULL;
}

void Renderer::errorPaint(HWND hwnd, const char* message)
{
	MONITOR(Renderer::errorPaint)

	// Paint an error message in the window
	ASSERT(hwnd != NULL);
	RECT rect;
	VERIFY(::GetClientRect(hwnd, &rect));
	HDC hdc = ::GetDC(hwnd);
	ASSERT(hdc != NULL);
	VERIFY(::FillRect(hdc, &rect, reinterpret_cast<HBRUSH>(COLOR_3DFACE + 1)));
	if (message != NULL)
		::DrawText(hdc, message, -1, &rect, DT_CENTER|DT_NOPREFIX|DT_WORDBREAK);
	return;
}

void Hoarding::render(HDC hdc, LOGFONT& font, RECT& rect, const char* message) const
{
	MONITOR(Hoarding::render)

	// Draw text in a particular font and colour scheme
	ASSERT(hdc != NULL);
	if (text != NULL)
		message = text;
	ASSERT(message != NULL);
	ASSERT(face != NULL);
	System::asciizCopy(font.lfFaceName, face, dimof(font.lfFaceName));
	HFONT hfont = ::CreateFontIndirect(&font);
	ASSERT(hfont != NULL);
	HGDIOBJ ofont = NULL;
	if (hfont != NULL)
		ofont = ::SelectObject(hdc, hfont);
	::SetTextColor(hdc, fg);
	::SetBkColor(hdc, bg);
	VERIFY(::ExtTextOut(hdc, (rect.left + rect.right) / 2, rect.bottom - 1, ETO_CLIPPED|ETO_OPAQUE,
						&rect, message, System::asciizLength(message), NULL));
	if (hfont != NULL)
		VERIFY(::DeleteObject(::SelectObject(hdc, ofont)));
	rect.top += H2TEX(1);
	rect.bottom += H2TEX(1);
	return;
}

HRESULT Renderer::textureLoad(HDC hdc, const char* message)
{
	MONITOR(Renderer::textureLoad)

	// Load a texture with the standard stuff
	ASSERT(hdc != NULL);
	HBRUSH white = ::CreateSolidBrush(RGB(0xFF,0xFF,0xFF));
	ASSERT(white != NULL);
	HBRUSH black = ::CreateSolidBrush(RGB(0x00,0x00,0x00));
	ASSERT(black != NULL);
	HBRUSH red = ::CreateSolidBrush(RGB(0xFF,0x00,0x00));
	ASSERT(red != NULL);
	HBRUSH blue = ::CreateSolidBrush(RGB(0x00,0x00,0xC0));
	ASSERT(blue != NULL);
	HBRUSH cyan = ::CreateSolidBrush(RGB(0xC0,0xFF,0xFF));
	ASSERT(cyan != NULL);

	// Antialias
	// Start with all white
	RECT rect = { 0, 0, TextureWidth, TextureHeight };
	VERIFY(::FillRect(hdc, &rect, white));
	const int mx = TextureWidth / 2;
	const int my = TextureHeight / 2;
	const int r0 = M2TEX(CURLING_RINK_BACK);
	const int r1 = M2TEX(1.2192);
	const int r2 = M2TEX(0.6096);
	const int r3 = M2TEX(0.1524);
	const int cx = H2TEX(3);
	const int cy = H2TEX(9);
	HGDIOBJ hpbrush = ::SelectObject(hdc, blue);
	ASSERT(hpbrush != NULL);
	HGDIOBJ hppen = ::SelectObject(hdc, ::GetStockObject(NULL_PEN));
	ASSERT(hppen != NULL);

	// See if we can use blt stretching to antialias the circles
	bool antialiased = false;
	if ((::GetDeviceCaps(hdc, RASTERCAPS) & RC_STRETCHBLT) && ::SetStretchBltMode(hdc, HALFTONE))
	{
		// Draw the big cyan guide
		rect.left = mx;
		rect.top = my - H2TEX(2);
		rect.bottom = my + H2TEX(2);
		VERIFY(::FillRect(hdc, &rect, cyan));

		// Fill in the guide and big circles
		VERIFY(::Ellipse(hdc, mx - r0 * 2, my - r0 * 2, mx + r0 * 2, my + r0 * 2));
		::SelectObject(hdc, white);
		VERIFY(::Ellipse(hdc, mx - r1 * 2, my - r1 * 2, mx + r1 * 2, my + r1 * 2));
		::SelectObject(hdc, red);
		VERIFY(::Ellipse(hdc, mx - r2 * 2, my - r2 * 2, mx + r2 * 2, my + r2 * 2));
		::SelectObject(hdc, white);
		VERIFY(::Ellipse(hdc, mx - r3 * 2, my - r3 * 2, mx + r3 * 2, my + r3 * 2));

		// Stretch it to the correct place and blank out the remainder
		VERIFY(::StretchBlt(hdc, 0, my, mx, my, hdc, 0, 0, TextureWidth, TextureHeight, SRCCOPY));
		rect.left = 0;
		rect.top = 0;
		rect.bottom = my;
		VERIFY(::FillRect(hdc, &rect, white));
		rect.left = mx;
		rect.top = my;
		rect.bottom = TextureHeight;
		VERIFY(::FillRect(hdc, &rect, white));
		rect.left = mx;
		rect.top = cy - H2TEX(1);
		rect.bottom = cy + H2TEX(1);
		VERIFY(::FillRect(hdc, &rect, cyan));
		antialiased = true;
	}

	// Fill in most of the target the normal way
	if (!antialiased)
	{
		// Draw the cyan guide
		rect.left = cx;
		rect.top = cy - H2TEX(1);
		rect.bottom = cy + H2TEX(1);
		VERIFY(::FillRect(hdc, &rect, cyan));

		// Fill in the circles
		VERIFY(::Ellipse(hdc, cx - r0, cy - r0, cx + r0, cy + r0));
		::SelectObject(hdc, white);
		VERIFY(::Ellipse(hdc, cx - r1, cy - r1, cx + r1, cy + r1));
		::SelectObject(hdc, red);
		VERIFY(::Ellipse(hdc, cx - r2, cy - r2, cx + r2, cy + r2));
		::SelectObject(hdc, white);
		VERIFY(::Ellipse(hdc, cx - r3, cy - r3, cx + r3, cy + r3));
	}

	// Draw the black lines
	rect.left = cx - M2TEX(CURLING_RINK_HOG) - CURLING_LINE_WIDTH / 2;
	rect.right = rect.left + CURLING_LINE_WIDTH;
	rect.top = H2TEX(6) + 1;
	rect.bottom = H2TEX(12) - 1;
	VERIFY(::FillRect(hdc, &rect, black));
	rect.right = cx - r0;
	rect.left = rect.right - CURLING_LINE_WIDTH;
	VERIFY(::FillRect(hdc, &rect, black));
	rect.left = cx - CURLING_LINE_WIDTH / 2;
	rect.right = rect.left + CURLING_LINE_WIDTH;
	rect.bottom = cy - r3 + 1;
	VERIFY(::FillRect(hdc, &rect, black));
	rect.top = cy + r3 - 1;
	rect.bottom = H2TEX(12) - 1;
	VERIFY(::FillRect(hdc, &rect, black));
	rect.left = cx - r0;
	rect.right = cx - r3 + 1;
	rect.top = cy - CURLING_LINE_WIDTH / 2;
	rect.bottom = rect.top + CURLING_LINE_WIDTH;
	VERIFY(::FillRect(hdc, &rect, black));
	rect.left = cx + r3 - 1;
	rect.right = H2TEX(12);
	VERIFY(::FillRect(hdc, &rect, black));

	// Draw the hack in blue
	const int hx = H2TEX(12) + cx - M2TEX(3.6576);
	const int hy = M2TEX(0.4572) / 2;
	rect.right = hx;
	rect.left = hx - CURLING_LINE_WIDTH * 2;
	rect.top = H2TEX(3) - hy;
	rect.bottom = rect.top + CURLING_LINE_WIDTH;
	VERIFY(::FillRect(hdc, &rect, blue));
	rect.bottom = H2TEX(3) + hy;
	rect.top = rect.bottom - CURLING_LINE_WIDTH;
	VERIFY(::FillRect(hdc, &rect, blue));
	rect.left = rect.right - CURLING_LINE_WIDTH;
	rect.top = H2TEX(3) - hy;
	VERIFY(::FillRect(hdc, &rect, blue));

	// Add scratch marks
	Random random;
	random.reset();
	int rop2 = ::SetROP2(hdc, R2_XORPEN);
	for (int grey = 3; grey < 32; grey += 4)
	{
		HPEN scratch = ::CreatePen(PS_SOLID, 0, RGB(grey, grey, grey));
		ASSERT(scratch != NULL);
		HGDIOBJ pen = ::SelectObject(hdc, scratch);
		ASSERT(pen != NULL);
		for (int scratches = 0; scratches < 100; scratches++)
		{
			rect.top = random.get(my);
			rect.left = random.get(mx);
			switch (random.get(3))
			{
				case 1:
					// North East quadrant
					rect.left += mx;
					break;

				case 2:
					// South West quadrant
					rect.top += my;
					break;

				default:
					// South East quadrant
					rect.top += my;
					rect.left += mx;
					break;
			}
			rect.right = rect.left + random.get(TextureWidth / 64);
			VERIFY(::MoveToEx(hdc, rect.left, rect.top, NULL));
			VERIFY(::LineTo(hdc, rect.right + 1, rect.top));
		}
		::SelectObject(hdc, pen);
		VERIFY(::DeleteObject(scratch));
	}
	VERIFY(::SetROP2(hdc, rop2));

	// Hoarding text
	static const Hoarding hoarding[6] =
	{
		{ NULL,				"Arial",				RGB(0xFF,0xFF,0xFF), RGB(0xFF,0x00,0x00) },
		{ "Cactus Forever",	"Comic Sans MS",		RGB(0xFF,0xFF,0xFF), RGB(0xFF,0x80,0x00) },
		{ "Curling",		"Courier New",			RGB(0x00,0xFF,0x00), RGB(0x00,0x10,0x00) },
		{ "chilliant.com",	"Lucida Calligraphy",	RGB(0xFF,0x00,0x00), RGB(0x40,0x40,0xFF) },
		{ "Ian Taylor",		"Times Roman",			RGB(0x00,0x00,0xFF), RGB(0xFF,0xFF,0x00) },
		{ "Advertise Here",	"Arial Narrow",			RGB(0x00,0x00,0x00), RGB(0xC0,0xC0,0xFF) },
	};
	rect.left = 1;
	rect.right = H2TEX(6) - 1;
	rect.top = 1;
	rect.bottom = H2TEX(1) - 1;
	LOGFONT font = { rect.bottom - rect.top - 2 };
	font.lfWeight = FW_BOLD;
	font.lfQuality = ANTIALIASED_QUALITY;
	::SetTextAlign(hdc, TA_BOTTOM|TA_CENTER|TA_NOUPDATECP);
	for (int i = 0; i < dimof(hoarding); i++)
		hoarding[i].render(hdc, font, rect, message);

	// Clean up
	::SelectObject(hdc, hppen);
	::SelectObject(hdc, hpbrush);
	VERIFY(::DeleteObject(cyan));
	VERIFY(::DeleteObject(blue));
	VERIFY(::DeleteObject(red));
	VERIFY(::DeleteObject(black));
	VERIFY(::DeleteObject(white));
	return S_OK;
}
