//----------------------------------------------------------------------------
//	SOURCE:	RenderX7.cpp
//	AUTHOR:	Ian Taylor
//	NOTES:	Curling game DirectX 7.0 renderer implementation
//----------------------------------------------------------------------------

#include "Curling.h"
#include "Render.h"
#include "Game.h"
#include "Stone.h"
#include "Rink.h"
#include "TCom.h"

// DirectX
#define D3D_OVERLOADS
#include <d3d.h>
#pragma pack(4)

#define VERIFY_OK(expr)		VERIFY_(expr,== D3D_OK)
#define TRY(function, call)	if (!failed(call, #function)) TEA_NOP(); else return false

TEA_NAMESPACE_USE

// Colours
#define LightAmbient			0.30f
#define LightDiffuse			0.40f
#define LightSpecular			0.30f
#define LightSpecularPower		20.0f
#define ColourBackground		D3DRGB(0.75f, 0.87f, 1.00f)

class RendererX7 : public Renderer
{
	private:

		// Constants
		enum DeviceType
		{
			// In order of superiority
			DeviceTNL = 0,
			DeviceHAL,
			DeviceRGB,
			DeviceREF,
			DeviceNone,
			DeviceTypeCount = DeviceNone
		};
		static const GUID* const DeviceGUID[DeviceTypeCount];

		// Private data
		const char* m_inactive;
		Error m_error;
		DeviceType m_bestDevice;
		IDirectDraw7* m_dd;
		IDirect3D7* m_d3d;
		IDirectDrawSurface7* m_primary;
		IDirectDrawSurface7* m_backBuffer;
		IDirectDrawSurface7* m_texture;
		IDirect3DDevice7* m_device;
		IDirect3DVertexBuffer7* m_vbStone;
		IDirect3DVertexBuffer7* m_vbRink;
		DDSURFACEDESC2 m_backDesc;
		D3DDEVICEDESC7 m_deviceDesc;
		DDPIXELFORMAT m_zbufferDesc;
		RECT m_screen;

		// Callbacks
		static HRESULT CALLBACK enumDevices(char* description, char* device, D3DDEVICEDESC7* desc, void* param);
		static HRESULT CALLBACK enumZBuffers(DDPIXELFORMAT* format, void* param);
		static HRESULT CALLBACK enumTextures(DDPIXELFORMAT* format, void* param);

		// Implementation
		template<typename Interface> void release(Interface*& punk)
		{
			// Release a COM pointer
			MONITOR(RendererX7::release)
			if (punk != NULL)
				punk->Release();
			punk = NULL;
		}
		bool failed(HRESULT hr, const char* function)
		{
			// Test for failure
			MONITOR(RendererX7::failed)
			if (FAILED(hr))
				m_error.set(function, hr);
			return FAILED(hr);
		}

		// DirectX stuff
		static D3DVALUE magnitude(const D3DVECTOR& v) { return Math::sqrt(v.x * v.x + v.y * v.y + v.z * v.z); }
		bool initialiseSurfaces(HWND hwnd);
		bool initialiseTexture(void);
		bool initialiseVertexBuffers(void);
		bool setWorldMatrix(const D3DVALUE xyz[3], D3DVALUE rot, D3DVALUE rz);
		bool setViewMatrix(const GameCamera& camera);
		bool setProjectionMatrix(D3DVALUE cot_fov_by_2, D3DVALUE near_plane);
		bool setLight(D3DVALUE dz);
		bool setRenderState(D3DRENDERSTATETYPE state, DWORD value)
		{
			// Set a render state
			MONITOR(RendererX7::setRenderState)
			ASSERT(m_device != NULL);
			TRY("IDirect3DDevice7::SetRenderState", m_device->SetRenderState(state, value));
			return true;
		}
		bool setTexture(IDirectDrawSurface7* texture)
		{
			// Set the current texture
			MONITOR(RendererX7::setTextureState)
			ASSERT(m_device != NULL);
			TRY("IDirect3DDevice7::SetTexture", m_device->SetTexture(0, texture));
			return true;
		}
		bool setTextureState(D3DTEXTURESTAGESTATETYPE state, DWORD value)
		{
			// Set a texture state
			MONITOR(RendererX7::setTextureState)
			ASSERT(m_device != NULL);
			TRY("IDirect3DDevice7::SetTextureStageState", m_device->SetTextureStageState(0, state, value));
			return true;
		}

		// Scene rendering
		bool scenePrepareOnce(void);
		bool scenePrepare(const GameView& view);
		bool sceneRenderWrap(const GameWorld& world);
		bool sceneRender(const GameWorld& world);
		bool sceneRenderHalf(const GameWorld& world, D3DVALUE mirror);
		bool sceneRenderStone(Stone::What what)
		{
			// Render a single stone model
			MONITOR(RendererX7::sceneRenderStone)
			Stone::Model model;
			Stone::getModel(what, model);
			TRY("IDirect3DDevice7::DrawIndexedPrimitiveVB",
				m_device->DrawIndexedPrimitiveVB(D3DPT_TRIANGLESTRIP, m_vbStone,
												 model.vertexOffset, model.vertexCount,
												 model.indexArray, model.indexCount, 0));
			return true;
		}
		bool sceneRenderRink(Rink::What what)
		{
			// Render a single rink model
			MONITOR(RendererX7::sceneRenderRink)
			Rink::Model model;
			Rink::getModel(what, model);
			TRY("IDirect3DDevice7::DrawPrimitiveVB",
				m_device->DrawPrimitiveVB(D3DPT_TRIANGLELIST, m_vbRink,
										  model.vertexOffset, model.vertexCount, 0));
			return true;
		}

	public:

		// Renderer interface methods
		virtual bool windowReady(void);
		virtual bool windowChanged(Window window, Change what);
		virtual void windowPaint(Window window);
		virtual void sceneUpdate(const GameView& view);
		virtual void destroy(void);

		// Methods
		bool initialise(HWND hwnd);
		void inactive(const char* message);
		void shutdown(bool full);

		// Properties
		bool isActive(void) const { return m_inactive == NULL; }
		bool isHAL(void) const { return m_bestDevice <= DeviceHAL; }
		bool isTNL(void) const { return m_bestDevice <= DeviceTNL; }

		// Debugging
		bool isValid(void) const
		{
			return Memory::isValidRead(this);
		}
};

const GUID* const RendererX7::DeviceGUID[DeviceTypeCount] =
{
	// Same order as enum DeviceType
	&IID_IDirect3DTnLHalDevice,
	&IID_IDirect3DHALDevice,
	&IID_IDirect3DRGBDevice,
	&IID_IDirect3DRefDevice
};

Renderer* Renderer::createDirectX7(HWND hwnd)
{
	MONITOR(Renderer::createDirectX7)

	// Call the appropriate factory
	static BYTE ugly[sizeof(RendererX7)];
	System::memoryZero(ugly, sizeof(ugly));
	RendererX7* renderer = new(ugly) RendererX7;
	ASSERT(renderer != NULL);
	if (!renderer->initialise(hwnd))
	{
		// Failed to initialise the renderer, so go inactive
		if (renderer->isActive())
			renderer->inactive(NULL);
	}
	VERIFY(::InvalidateRect(hwnd, NULL, FALSE));
	return renderer;
}

void RendererX7::inactive(const char* message)
{
	MONITOR(Renderer::inactive)

	// Make this renderer inactive
	if (message == NULL)
	{
		// Format an appropriate error
		static char buffer[1024];
		m_error.format(buffer, dimof(buffer));
		m_inactive = buffer;
	}
	else
		m_inactive = message;
	shutdown(false);
	return;
}

HRESULT CALLBACK RendererX7::enumDevices(char*, char*, D3DDEVICEDESC7* desc, void* param)
{
	MONITOR(RendererX7::enumDevices)

	// Search for the best device
	ASSERT(desc != NULL);
	ASSERT(param != NULL);
	int& best = *static_cast<int*>(param);
	ASSERT((best >= 0) && (best <= DeviceTypeCount));
	for (int i = 0; i < best; i++)
		if (::IsEqualGUID(desc->deviceGUID, *(DeviceGUID[i])))
		{
			// This one's better
			best = i;
			break;
		}
	return D3DENUMRET_OK;
}

HRESULT CALLBACK RendererX7::enumZBuffers(DDPIXELFORMAT* format, void* param)
{
	MONITOR(RendererX7::enumZBuffers)

	// Search for a specific z-buffer size
	ASSERT(format != NULL);
	DDPIXELFORMAT* best = static_cast<DDPIXELFORMAT*>(param);
	ASSERT(best != NULL);
	if (format->dwZBufferBitDepth != best->dwZBufferBitDepth)
		return D3DENUMRET_OK;
	*best = *format;
	return (best->dwStencilBitDepth == 0) ? D3DENUMRET_CANCEL : D3DENUMRET_OK;
}

HRESULT CALLBACK RendererX7::enumTextures(DDPIXELFORMAT* format, void* param)
{
	MONITOR(RendererX7::enumTextures)

	// Search for a good texture format
	ASSERT(format != NULL);
	DDPIXELFORMAT* best = static_cast<DDPIXELFORMAT*>(param);
	ASSERT(best != NULL);
	if (format->dwFlags & (DDPF_LUMINANCE|DDPF_BUMPLUMINANCE|DDPF_BUMPDUDV|DDPF_ALPHAPIXELS))
		return DDENUMRET_OK;
	if (format->dwFourCC != 0)
		return DDENUMRET_OK;
	if (format->dwRGBBitCount != best->dwRGBBitCount)
		return DDENUMRET_OK;
	*best = *format;
	return DDENUMRET_CANCEL;
}

bool RendererX7::initialise(HWND hwnd)
{
	MONITOR(RendererX7::initialise)

	// Load DirectX for windowed mode
	ASSERT(m_dd == NULL);
	TRY("::DirectDrawCreateEx", ::DirectDrawCreateEx(NULL, reinterpret_cast<void**>(&m_dd), IID_IDirectDraw7, NULL));
	ASSERT(m_dd != NULL);
	TRY("IDirectDraw7::SetCooperativeLevel", m_dd->SetCooperativeLevel(hwnd, DDSCL_NORMAL));

	// Query DirectDraw for Direct3D
	ASSERT(m_d3d == NULL);
	TRY("IDirectDraw7::QueryInterface", m_dd->QueryInterface(IID_IDirect3D7, reinterpret_cast<void**>(&m_d3d)));
	ASSERT(m_d3d != NULL);

	// Find the best device
	m_bestDevice = DeviceNone;
	TRY("IDirect3D7::EnumDevices", m_d3d->EnumDevices(enumDevices, &m_bestDevice));
	if (m_bestDevice == DeviceNone)
	{
		// No point continuing
		inactive("\nSorry, unable to find suitable DirectX 3D device");
		return false;
	}
	ASSERT((m_bestDevice >= 0) && (m_bestDevice < DeviceTypeCount));

	// Now initialise the surfaces and perform one-time preparation
	return initialiseSurfaces(hwnd) && scenePrepareOnce();
}

bool RendererX7::initialiseSurfaces(HWND hwnd)
{
	MONITOR(RendererX7::initialiseSurfaces)

	// Check the display mode
	ASSERT(m_dd != NULL);
	ASSERT(m_d3d != NULL);
	DDSURFACEDESC2 surface = { sizeof(surface) };
	TRY("IDirectDraw7::GetDisplayMode", m_dd->GetDisplayMode(&surface));
	if (surface.ddpfPixelFormat.dwRGBBitCount < 16)
	{
		// Cannot use paletted display modes
		inactive("\nPlease switch to a display mode with at least 65536 colours (sixteen bits)");
		return false;
	}
	const GUID& deviceGUID = *(DeviceGUID[m_bestDevice]);

	// Create the primary surface
	ASSERT(m_primary == NULL);
	Memory::zero(surface);
	surface.dwSize = sizeof(surface);
	surface.dwFlags = DDSD_CAPS;
	surface.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
	TRY("IDirectDraw7::CreateSurface", m_dd->CreateSurface(&surface, &m_primary, NULL));
	ASSERT(m_primary != NULL);

	// Create a clipper object
	ComPtr<IDirectDrawClipper> clipper;
	TRY("IDirectDraw7::CreateClipper", m_dd->CreateClipper(0, &clipper, NULL));
	ASSERT(clipper != NULL);
	TRY("IDirectDrawClipper::SetHWnd", clipper->SetHWnd(0, hwnd));
	TRY("IDirectDrawSurface7::SetClipper", m_primary->SetClipper(clipper));

	// Create the back buffer
	ASSERT(m_backBuffer == NULL);
	VERIFY(::GetWindowRect(hwnd, &m_screen));
	surface.dwFlags = DDSD_WIDTH|DDSD_HEIGHT|DDSD_CAPS;
	surface.dwWidth = m_screen.right - m_screen.left;
	surface.dwHeight = m_screen.bottom - m_screen.top;
	surface.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN|DDSCAPS_3DDEVICE;
	TRY("IDirectDraw7::CreateSurface", m_dd->CreateSurface(&surface, &m_backBuffer, NULL));
	ASSERT(m_backBuffer != NULL);
	m_backDesc.dwSize = sizeof(m_backDesc);
	TRY("IDirectDrawSurface7::GetSurfaceDesc", m_backBuffer->GetSurfaceDesc(&m_backDesc));

	// Look for a suitable z-buffer
	m_zbufferDesc.dwSize = 0;
	m_zbufferDesc.dwZBufferBitDepth = m_backDesc.ddpfPixelFormat.dwRGBBitCount;
	TRY("IDirect3D7::EnumZBufferFormats", m_d3d->EnumZBufferFormats(deviceGUID, enumZBuffers, &m_zbufferDesc));
	if (m_zbufferDesc.dwSize == 0)
	{
		// Oh, just look for 16 bit z-buffers
		m_zbufferDesc.dwZBufferBitDepth = 16;
		TRY("IDirect3D7::EnumZBufferFormats", m_d3d->EnumZBufferFormats(deviceGUID, enumZBuffers, &m_zbufferDesc));
		if (m_zbufferDesc.dwSize == 0)
		{
			// Bad luck
			inactive("\nSorry, your display adapter does not support z-buffering");
			return false;
		}
	}

	// Create and attach the z-buffer
	surface.dwFlags = DDSD_WIDTH|DDSD_HEIGHT|DDSD_CAPS|DDSD_PIXELFORMAT;
	surface.ddsCaps.dwCaps = DDSCAPS_ZBUFFER|(isHAL() ? DDSCAPS_VIDEOMEMORY : DDSCAPS_SYSTEMMEMORY);
	surface.ddpfPixelFormat = m_zbufferDesc;
	ComPtr<IDirectDrawSurface7> zbuffer;
	TRY("IDirectDraw7::CreateSurface", m_dd->CreateSurface(&surface, &zbuffer, NULL));
	ASSERT(zbuffer != NULL);
	TRY("IDirectDrawSurface7::AddAttachedSurface", m_backBuffer->AddAttachedSurface(zbuffer));

	// Create a rendering device
	ASSERT(m_device == NULL);
	TRY("IDirect3D7::CreateDevice", m_d3d->CreateDevice(deviceGUID, m_backBuffer, &m_device));
	ASSERT(m_device != NULL);
	TRY("IDirect3DDevice7::GetCaps", m_device->GetCaps(&m_deviceDesc));

	// Set the viewport
	D3DVIEWPORT7 viewport = { 0, 0, m_backDesc.dwWidth, m_backDesc.dwHeight, 0.0f, 1.0f };
	TRY("IDirect3DDevice7::SetViewport", m_device->SetViewport(&viewport));

	// And initialise the texture surface
	return initialiseTexture() && initialiseVertexBuffers();
}

bool RendererX7::initialiseTexture(void)
{
	MONITOR(RendererX7::initialiseTexture)

	// Initialise the texture surface
	ASSERT(m_dd != NULL);
	DDSURFACEDESC2 surface = { sizeof(surface) };
	surface.dwFlags = DDSD_WIDTH|DDSD_HEIGHT|DDSD_CAPS|DDSD_TEXTURESTAGE|DDSD_PIXELFORMAT;
	surface.dwWidth = TextureWidth;
	surface.dwHeight = TextureHeight;
	surface.ddsCaps.dwCaps = DDSCAPS_TEXTURE;
	if (isHAL())
		surface.ddsCaps.dwCaps2 = DDSCAPS2_TEXTUREMANAGE;
	else
		surface.ddsCaps.dwCaps |= DDSCAPS_SYSTEMMEMORY;
	surface.dwTextureStage = 0;

	// Look for a suitable texture format
	surface.ddpfPixelFormat.dwSize = 0;
	surface.ddpfPixelFormat.dwRGBBitCount = m_backDesc.ddpfPixelFormat.dwRGBBitCount;
	TRY("IDirect3DDevice7::EnumTextureFormats", m_device->EnumTextureFormats(enumTextures, &surface.ddpfPixelFormat));
	if (surface.ddpfPixelFormat.dwSize == 0)
	{
		// Oh, just look for 16 bit textures
		surface.ddpfPixelFormat.dwRGBBitCount = 16;
		TRY("IDirect3DDevice7::EnumTextureFormats", m_device->EnumTextureFormats(enumTextures, &surface.ddpfPixelFormat));
		if (surface.ddpfPixelFormat.dwSize == 0)
		{
			// Bad luck
			inactive("\nSorry, your display adapter does not have sufficent texture support");
			return false;
		}
	}

	// Create the texture
	ASSERT(m_texture == NULL);
	TRY("IDirectDraw7::CreateSurface", m_dd->CreateSurface(&surface, &m_texture, NULL));
	ASSERT(m_texture != NULL);

	// Fill the texture with the standard stuff
	HDC hdc = NULL;
	TRY("IDirectDrawSurface7::GetDC", m_texture->GetDC(&hdc));
	HRESULT hr = textureLoad(hdc, "DirectX 7.0");
	TRY("IDirectDrawSurface7::ReleaseDC", m_texture->ReleaseDC(hdc));
	TRY("Renderer::textureLoad", hr);
	return true;
}

bool RendererX7::initialiseVertexBuffers(void)
{
	MONITOR(RendererX7::initialiseVertexBuffers)

	// Initialise the vertex buffer
	ASSERT(m_d3d != NULL);

	// Create the stone VB
	D3DVERTEXBUFFERDESC desc = { sizeof(desc) };
	desc.dwCaps = D3DVBCAPS_WRITEONLY;
	if (!isTNL())
		desc.dwCaps |= D3DVBCAPS_SYSTEMMEMORY;
	desc.dwFVF = D3DFVF_XYZ|D3DFVF_NORMAL|D3DFVF_DIFFUSE;
	desc.dwNumVertices = Stone::vertices();
	ASSERT(m_vbStone == NULL);
	TRY("IDirect3D7::CreateVertexBuffer", m_d3d->CreateVertexBuffer(&desc, &m_vbStone, 0));
	ASSERT(m_vbStone != NULL);
	void* memory = NULL;
	TRY("IDirect3DVertexBuffer7::Lock", m_vbStone->Lock(DDLOCK_WAIT|DDLOCK_WRITEONLY|DDLOCK_DISCARDCONTENTS, &memory, NULL));
	ASSERT(memory != NULL);
	struct StoneVertex
	{
		D3DVALUE xyz[3];
		D3DVALUE normal[3];
		D3DCOLOR rgba;
	};
	Stone::Colour rgba;
	Stone stone;
	int count = 0;
	for (StoneVertex* sv = static_cast<StoneVertex*>(memory); stone.next(sv->xyz, sv->normal, rgba); sv++)
	{
		ASSERT(count < Stone::vertices());
		sv->rgba = RGBA_MAKE(rgba.r, rgba.g, rgba.b, rgba.a);
		count++;
	}
	ASSERT(count == Stone::vertices());
	TRY("IDirect3DVertexBuffer7::Unlock", m_vbStone->Unlock());
	TRY("IDirect3DVertexBuffer7::Optimize", m_vbStone->Optimize(m_device, 0));

	// Create the rink VB
	desc.dwFVF = D3DFVF_XYZ|D3DFVF_TEX1;
	desc.dwNumVertices = Rink::vertices();
	ASSERT(m_vbRink == NULL);
	TRY("IDirect3D7::CreateVertexBuffer", m_d3d->CreateVertexBuffer(&desc, &m_vbRink, 0));
	ASSERT(m_vbRink != NULL);
	TRY("IDirect3DVertexBuffer7::Lock", m_vbRink->Lock(DDLOCK_WAIT|DDLOCK_WRITEONLY|DDLOCK_DISCARDCONTENTS, &memory, NULL));
	ASSERT(memory != NULL);
	struct RinkVertex
	{
		D3DVALUE xyz[3];
		D3DVALUE uv[2];
	};
	Rink rink;
	count = 0;
	for (RinkVertex* rv = static_cast<RinkVertex*>(memory); rink.next(rv->xyz, rv->uv); rv++)
	{
		ASSERT(count < Rink::vertices());
		count++;
	}
	ASSERT(count == Rink::vertices());
	TRY("IDirect3DVertexBuffer7::Unlock", m_vbRink->Unlock());
	TRY("IDirect3DVertexBuffer7::Optimize", m_vbRink->Optimize(m_device, 0));
	return true;
}

void RendererX7::shutdown(bool full)
{
	MONITOR(RendererX7::shutdown)

	// Shutdown the renderer
	release(m_vbStone);
	release(m_vbRink);
	if (m_device != NULL)
		VERIFY(setTexture(NULL));
	release(m_texture);
	release(m_device);
	release(m_backBuffer);
	release(m_primary);
	if (full)
	{
		release(m_d3d);
		release(m_dd);
	}
	return;
}

bool RendererX7::windowReady(void)
{
	MONITOR(RendererX7::windowReady)

	// Determine if the screen is ready to be painted
	return isActive() && (m_primary->GetBltStatus(DDGBS_ISBLTDONE) == S_OK);
}

bool RendererX7::windowChanged(HWND hwnd, Change what)
{
	MONITOR(RendererX7::windowChanged)

	// Shutdown and restart the majority of the context we have (sigh)
	ASSERT(hwnd != NULL);
	if (what == ChangePosition)
	{
		// Only need to keep track of the position
		VERIFY(::GetWindowRect(hwnd, &m_screen));
		return false;
	}
	shutdown(false);
	m_inactive = NULL;
	if (!initialiseSurfaces(hwnd) || !scenePrepareOnce())
	{
		// Failed to reinitialise the renderer, so go inactive
		if (isActive())
			inactive(NULL);
	}
	VERIFY(::InvalidateRect(hwnd, NULL, FALSE));
	return true;
}

void RendererX7::windowPaint(HWND hwnd)
{
	MONITOR(RendererX7::windowPaint)

	// Paint the back buffer
	if (isActive())
	{
		// Blit the image
		VERIFY_OK(m_primary->Blt(&m_screen, m_backBuffer, NULL, DDBLT_DONOTWAIT, NULL));
	}
	else
	{
		// We're inactive, so display the reason
		errorPaint(hwnd, m_inactive);
	}
	return;
}

bool RendererX7::setWorldMatrix(const D3DVALUE xyz[3], D3DVALUE rot, D3DVALUE rz)
{
	MONITOR(RendererX7::setWorldMatrix)

	// Set the world matrix with translation, z-rotation and z-reflection
	D3DVALUE sin, cos;
	Math::sincos(rot, sin, cos);
	D3DMATRIX mat;
	Memory::zero(mat);
	mat.m[0][0] = cos;
	mat.m[0][1] = sin;
	mat.m[1][0] = -sin;
	mat.m[1][1] = cos;
	mat.m[2][2] = rz;
	mat.m[3][0] = xyz[0];
	mat.m[3][1] = xyz[1];
	mat.m[3][2] = rz * xyz[2];
	mat.m[3][3] = 1.0f;
	TRY("IDirect3DDevice7::SetTransform", m_device->SetTransform(D3DTRANSFORMSTATE_WORLD, &mat));
	return true;
}

bool RendererX7::setViewMatrix(const GameCamera& camera)
{
	MONITOR(RendererX7::setViewMatrix)

	// Set the view matrix
	D3DVECTOR from(camera.px, camera.py, camera.pz);
	D3DVECTOR at(camera.lx, camera.ly, camera.lz);
	D3DVECTOR view = at - from;
	D3DVALUE length = magnitude(view);
	ASSERT(length > Math::Epsilon);
	view /= length;
	D3DVECTOR up(-view.z * view.x, -view.z * view.y, 1.0f - view.z * view.z);
	length = magnitude(up);
	if (length < Math::Epsilon)
	{
		// Try another orientation (will succeed for all our cases)
		up.x = -view.y * view.x;
		up.y = 1.0f - view.y * view.y;
		up.z = -view.y * view.z;
		length = magnitude(up);
		ASSERT(length > Math::Epsilon);
	}
	up /= length;
	D3DVECTOR right = CrossProduct(view, up);
	D3DMATRIX mat;
	Memory::zero(mat);
	mat.m[0][0] = right.x;
	mat.m[0][1] = up.x;
	mat.m[0][2] = view.x;
	mat.m[1][0] = right.y;
	mat.m[1][1] = up.y;
	mat.m[1][2] = view.y;
	mat.m[2][0] = right.z;
	mat.m[2][1] = up.z;
	mat.m[2][2] = view.z;
	mat.m[3][0] = -DotProduct(from, right);
	mat.m[3][1] = -DotProduct(from, up);
	mat.m[3][2] = -DotProduct(from, view);
	mat.m[3][3] = 1.0f;
	TRY("IDirect3DDevice7::SetTransform", m_device->SetTransform(D3DTRANSFORMSTATE_VIEW, &mat));
	return true;
}

bool RendererX7::setProjectionMatrix(D3DVALUE cot_fov_by_2, D3DVALUE near_plane)
{
	MONITOR(RendererX7::setProjectionMatrix)

	// Set the projection matrix
	D3DMATRIX mat;
	Memory::zero(mat);
	mat.m[0][0] = (cot_fov_by_2 * m_backDesc.dwHeight) / m_backDesc.dwWidth;
	mat.m[1][1] = cot_fov_by_2;
	mat.m[2][2] = 1.0f;
	mat.m[3][2] = -cot_fov_by_2 * near_plane;
	mat.m[2][3] = 1.0f;
	TRY("IDirect3DDevice7::SetTransform", m_device->SetTransform(D3DTRANSFORMSTATE_PROJECTION, &mat));
	return true;
}

bool RendererX7::setLight(D3DVALUE dz)
{
	MONITOR(RendererX7::setLight)

	// Set the single directional light
	D3DLIGHT7 light;
	Memory::zero(light);
	light.dltType = D3DLIGHT_DIRECTIONAL;
	light.dcvDiffuse.r = light.dcvDiffuse.g = light.dcvDiffuse.b = LightDiffuse;
	light.dcvSpecular.r = light.dcvSpecular.g = light.dcvSpecular.b = LightSpecular;
	light.dvDirection.z = dz;
	TRY("IDirect3DDevice7::SetLight", m_device->SetLight(0, &light));
	return true;
}

void RendererX7::sceneUpdate(const GameView& view)
{
	MONITOR(RendererX7::sceneUpdate)

	// Update the back buffer for the new scene
	if (!isActive())
		return;
	if (!scenePrepare(view) || !sceneRenderWrap(view.world))
		if (isActive())
			inactive(NULL);
	return;
}

bool RendererX7::scenePrepareOnce(void)
{
	MONITOR(RendererX7::scenePrepareOnce)

	// Perform one-time scene preparation
	ASSERT(m_dd != NULL);
	ASSERT(m_d3d != NULL);
	ASSERT(m_primary != NULL);
	ASSERT(m_backBuffer != NULL);
	ASSERT(m_device != NULL);
	ASSERT(m_texture != NULL);
	ASSERT(m_vbStone != NULL);
	ASSERT(m_vbRink != NULL);

	// Set miscellaneous render states
	const DWORD ambient = (DWORD)D3DRGBA(LightAmbient, LightAmbient, LightAmbient, 1.0f);
	if (!setRenderState(D3DRENDERSTATE_DITHERENABLE, TRUE) ||
		!setRenderState(D3DRENDERSTATE_SPECULARENABLE, TRUE) ||
		!setRenderState(D3DRENDERSTATE_AMBIENT, ambient) ||
		!setRenderState(D3DRENDERSTATE_AMBIENTMATERIALSOURCE, D3DMCS_COLOR1) ||
		!setRenderState(D3DRENDERSTATE_SPECULARMATERIALSOURCE, D3DMCS_MATERIAL) ||
		!setRenderState(D3DRENDERSTATE_TEXTUREFACTOR, (DWORD)D3DRGBA(1, 1, 1, 0.8)) ||
		!setTextureState(D3DTSS_ALPHAARG1, D3DTA_TFACTOR) ||
		!setTextureState(D3DTSS_MINFILTER, D3DTFN_LINEAR) ||
		!setTextureState(D3DTSS_MAGFILTER, D3DTFG_LINEAR) ||
		!setTextureState(D3DTSS_COLOROP, D3DTOP_SELECTARG1) ||
		!setTextureState(D3DTSS_ADDRESS, D3DTADDRESS_CLAMP))
		return false;

	// Set up antialiasing only if we're a small window
	if ((m_deviceDesc.dpcTriCaps.dwRasterCaps & D3DPRASTERCAPS_ANTIALIASSORTINDEPENDENT) &&
		(m_backDesc.dwWidth <= 256) &&
		(m_backDesc.dwHeight <= 256) &&
		!setRenderState(D3DRENDERSTATE_ANTIALIAS, D3DANTIALIAS_SORTINDEPENDENT))
		return false;

	// Set up anisotropic filtering for textures
	if (m_deviceDesc.dpcTriCaps.dwRasterCaps & D3DPRASTERCAPS_ANISOTROPY)
		if (!setTextureState(D3DTSS_MAGFILTER, D3DTFG_ANISOTROPIC) ||
			!setTextureState(D3DTSS_MINFILTER, D3DTFN_ANISOTROPIC) ||
			!setTextureState(D3DTSS_MAXANISOTROPY, 2))
		return false;

	// Set up shiny material
	D3DMATERIAL7 material;
	Memory::zero(material);
	material.dcvSpecular.r = material.dcvSpecular.g = material.dcvSpecular.b = 1.0f;
	material.dvPower = LightSpecularPower;
	TRY("IDirect3DDevice7::SetMaterial", m_device->SetMaterial(&material));

	// Set up lighting states
	TRY("IDirect3DDevice7::LightEnable", m_device->LightEnable(0, TRUE));

	// Set the projection matrix
	return setProjectionMatrix(CURLING_COT_FOV_BY_2, 0.01f);
}

bool RendererX7::scenePrepare(const GameView& view)
{
	MONITOR(RendererX7::scenePrepare)

	// Set the view matrix
	if (!setViewMatrix(view.camera))
		return false;
	return true;
}

bool RendererX7::sceneRenderWrap(const GameWorld& world)
{
	MONITOR(RendererX7::sceneRenderWrap)

	// Wrap the scene rendering in a matched 'BeginScene' and 'EndScene' pair
	DWORD clear = D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER;
	if (m_zbufferDesc.dwStencilBitDepth > 0)
		clear |= D3DCLEAR_STENCIL;
	TRY("IDirect3DDevice7::Clear", m_device->Clear(0, NULL, clear, ColourBackground, 1.0f, 0));
	TRY("IDirect3DDevice7::BeginScene", m_device->BeginScene());
	if (sceneRender(world))
	{
		// End the scene
		TRY("IDirect3DDevice7::EndScene", m_device->EndScene());
		return true;
	}
	VERIFY_OK(m_device->EndScene());
	return false;
}

bool RendererX7::sceneRenderHalf(const GameWorld& world, D3DVALUE mirror)
{
	MONITOR(RendererX7::sceneRender)

	// Render everything not in the z=0 plane
	ASSERT(m_device != NULL);
	D3DVALUE point[3] = { 0.0f, 0.0f, 0.0f };
 	if (!setWorldMatrix(point, 0.0f, mirror) ||
		!setTexture(m_texture) ||
		!sceneRenderRink(Rink::Hoardings) ||
		!setTexture(NULL))
		return false;
	for (int i = 0; i < world.stones; i++)
	{
		// Render the stone
		point[0] = world.stone[i].x;
		point[1] = world.stone[i].y;
 		if (!setWorldMatrix(point, world.stone[i].rot, mirror))
			return false;
		switch (world.stone[i].colour)
		{
			case GameStone::Red:
				// Red stone
				if (!sceneRenderStone(Stone::RedStone))
					return false;
				break;

			case GameStone::Yellow:
				// Yellow stone
				if (!sceneRenderStone(Stone::YellowStone))
					return false;
				break;

			default:
				// Oh dear
				UNREACHABLE();
				break;
		}
	}
	return true;
}

bool RendererX7::sceneRender(const GameWorld& world)
{
	MONITOR(RendererX7::sceneRender)

	// Send the primitives to the device
	VALIDATE(world);
	ASSERT(m_device != NULL);
	ASSERT(m_vbStone != NULL);
	ASSERT(m_vbRink != NULL);

	// Render the reflection first
	if (!setRenderState(D3DRENDERSTATE_CULLMODE, D3DCULL_CW) ||
		!setLight(1.0f) ||
		!sceneRenderHalf(world, -1.0f) ||
		!setLight(-1.0f) ||
		!setRenderState(D3DRENDERSTATE_CULLMODE, D3DCULL_CCW))
		return false;

	// Render the z=0 plane
	D3DVALUE point[3] = { 0.0f, 0.0f, 0.0f };
	if (!setTexture(m_texture) ||
		!setRenderState(D3DRENDERSTATE_ZWRITEENABLE, FALSE) ||
		!setRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE) ||
		!setRenderState(D3DRENDERSTATE_SRCBLEND, D3DBLEND_SRCALPHA) ||
		!setRenderState(D3DRENDERSTATE_DESTBLEND, D3DBLEND_INVSRCALPHA) ||
		!setWorldMatrix(point, 0.0f, 1.0f) ||
		!sceneRenderRink(Rink::Ice) ||
		!setTexture(NULL) ||
		!setRenderState(D3DRENDERSTATE_SRCBLEND, D3DBLEND_ZERO) ||
		!setRenderState(D3DRENDERSTATE_DESTBLEND, D3DBLEND_SRCCOLOR))
		return false;
	for (int i = 0; i < world.stones; i++)
	{
		// Render the shadows
		point[0] = world.stone[i].x;
		point[1] = world.stone[i].y;
 		if (!setWorldMatrix(point, 0.0f, 1.0f) || !sceneRenderStone(Stone::Shadow))
			return false;
	}

	// Render the world above the ice
	if (!setRenderState(D3DRENDERSTATE_ZWRITEENABLE, TRUE) ||
		!setRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE) ||
		!sceneRenderHalf(world, 1.0f))
		return false;
	return true;
}

void RendererX7::destroy(void)
{
	MONITOR(RendererX7::destroy)

	// Destroy the rendering session
	shutdown(true);
	return;
}
