//----------------------------------------------------------------------------
//	SOURCE:	Rink.cpp
//	AUTHOR:	Ian Taylor
//	NOTES:	Curling rink model
//----------------------------------------------------------------------------

#include "Curling.h"
#include "Rink.h"

TEA_NAMESPACE_USE

// Squeezing models into a byte stream
#define MODEL_COORD(x, y, z, u, v)	(x)+3+7*((y)+13*(z)), (u)+16*(v),
#define MODEL_HOARD(x, y0, y1, v)	MODEL_COORD(x*2, y0, 0, 0, v+1)	\
									MODEL_COORD(x*3, y0, 1, 0, v)	\
									MODEL_COORD(x*3, y1, 1, 6, v)	\
									MODEL_COORD(x*2, y0, 0, 0, v+1)	\
									MODEL_COORD(x*3, y1, 1, 6, v)	\
									MODEL_COORD(x*2, y1, 0, 6, v+1)
#define MODEL_ICE(y0,y1,u0,u1,v)	MODEL_COORD(+1, y0, 0, u0, v+6)	\
									MODEL_COORD(-1, y0, 0, u0, v)	\
									MODEL_COORD(-1, y1, 0, u1, v)	\
									MODEL_COORD(+1, y0, 0, u0, v+6)	\
									MODEL_COORD(-1, y1, 0, u1, v)	\
									MODEL_COORD(+1, y1, 0, u1, v+6)
// Hand-coded (aarrgghh!)
static const Rink::Data g_hoardings[] =
{
	// Left side
	MODEL_HOARD(-1,  0,  1, 0)
	MODEL_HOARD(-1,  1,  2, 1)
	MODEL_HOARD(-1,  2,  3, 2)
	MODEL_HOARD(-1,  3,  4, 3)
	MODEL_HOARD(-1,  4,  5, 4)
	MODEL_HOARD(-1,  5,  6, 5)
	MODEL_HOARD(-1,  6,  7, 2)
	MODEL_HOARD(-1,  7,  8, 3)
	MODEL_HOARD(-1,  8,  9, 1)
	MODEL_HOARD(-1,  9, 10, 4)
	MODEL_HOARD(-1, 10, 12, 5)

	// Back
	MODEL_COORD(-3, 11, 0, 0, 1)
	MODEL_COORD(-3, 12, 1, 0, 0)
	MODEL_COORD(+3, 12, 1, 6, 0)
	MODEL_COORD(-3, 11, 0, 0, 1)
	MODEL_COORD(+3, 12, 1, 6, 0)
	MODEL_COORD(+3, 11, 0, 6, 1)

	// Right-side
	MODEL_HOARD(+1, 12, 10, 1)
	MODEL_HOARD(+1, 10,  9, 3)
	MODEL_HOARD(+1,  9,  8, 5)
	MODEL_HOARD(+1,  8,  7, 0)
	MODEL_HOARD(+1,  7,  6, 2)
	MODEL_HOARD(+1,  6,  5, 4)
	MODEL_HOARD(+1,  5,  4, 1)
	MODEL_HOARD(+1,  4,  3, 5)
	MODEL_HOARD(+1,  3,  2, 4)
	MODEL_HOARD(+1,  2,  1, 2)
	MODEL_HOARD(+1,  1,  0, 3)
};
static const Rink::Data g_ice[] =
{
	// Left gutter
	MODEL_COORD(-1,  0, 0,  6,  9)
	MODEL_COORD(-2,  0, 0,  6, 10)
	MODEL_COORD(-2, 11, 0, 11, 10)
	MODEL_COORD(-1,  0, 0,  6,  9)
	MODEL_COORD(-2, 11, 0, 11, 10)
	MODEL_COORD(-1, 11, 0, 11,  9)

	// The ice
	MODEL_ICE( 0,  1,  6, 12, 0)
	MODEL_ICE( 1,  3,  0, 12, 6)
	MODEL_ICE( 3,  8,  6, 11, 6)
	MODEL_ICE( 8, 10, 12,  0, 6)
	MODEL_ICE(10, 11, 12,  6, 0)

	// Right gutter
	MODEL_COORD(+1,  0, 0,  6,  9)
	MODEL_COORD(+2, 11, 0, 11, 10)
	MODEL_COORD(+2,  0, 0,  6, 10)
	MODEL_COORD(+1,  0, 0,  6,  9)
	MODEL_COORD(+1, 11, 0, 11,  9)
	MODEL_COORD(+2, 11, 0, 11, 10)
};
static const Rink::Index n_hoardings = dimof(g_hoardings) / 2;
static const Rink::Index n_ice = dimof(g_ice) / 2;
const Rink::Index Rink::m_vertices = n_hoardings + n_ice;

void Rink::reset(void)
{
	MONITOR(Rink::reset)

	// Reset the vertex iterator
	m_phase = 0;
	m_ptr = g_hoardings;
	m_end = g_hoardings + dimof(g_hoardings);
	return;
}

bool Rink::next(real xyz[3], real uv[2])
{
	MONITOR(Rink::next)

	// Fetch the next vertex in the list
	if (m_ptr >= m_end)
	{
		// May need to go to phase two
		if (m_phase > 0)
			return false;
		m_phase++;
		m_ptr = g_ice;
		m_end = g_ice + dimof(g_ice);
	}

	// Extract the packed data
	int z = *(m_ptr++);
	int x = z % 7;
	z /= 7;
	int y = z % 13;
	z /= 13;
	int v = *(m_ptr++);
	int u = v % 16;
	v /= 16;
	ASSERT((x >= 0) && (x <= 6));
	ASSERT((y >= 0) && (y <= 12));
	ASSERT((z >= 0) && (z <= 1));
	ASSERT((u >= 0) && (u <= 12));
	ASSERT((v >= 0) && (v <= 12));

	// Scale the values: World origin is at MODEL(9.5,0,0)
	static const real origin = REAL(9.5);
	static const real half = REAL(CURLING_RINK_WIDTH) * REAL(0.5);
	static const real slant = REAL(CURLING_HOARDINGS_HEIGHT) * REAL(0.5);
	static const real gutter = REAL(CURLING_GUTTER_WIDTH);
	static const real xlookup[7] =
	{
		// Width positions
		-half - gutter - slant,
		-half - gutter,
		-half,
		REAL(0.0),
		half,
		half + gutter,
		half + gutter + slant
	};
	xyz[0] = xlookup[x];
	if (y == 12)
		xyz[1] = (11 - origin) * REAL(CURLING_RINK_LENGTH) / 11 + slant;
	else
		xyz[1] = (y - origin) * REAL(CURLING_RINK_LENGTH) / 11;
	xyz[2] = z * slant * 2;
	uv[0] = u / REAL(12.0);
	uv[1] = v / REAL(12.0);
	return true;
}

void Rink::getModel(What what, Model& model)
{
	MONITOR(Rink::getModel)

	// Fetch a list of indices for a particular model
	switch (what)
	{
		case Hoardings:
			// Hoardings
			model.vertexOffset = 0;
			model.vertexCount = n_hoardings;
			break;

		case Ice:
			// Ice
			model.vertexOffset = n_hoardings;
			model.vertexCount = n_ice;
			break;

		default:
			// Oh dear
			UNREACHABLE();
	}
	return;
}
