//----------------------------------------------------------------------------
//	SOURCE:	Sound.cpp
//	AUTHOR:	Ian Taylor
//	NOTES:	Curling game sound implementation
//----------------------------------------------------------------------------

#include "Curling.h"
#include "Sound.h"
#include "Midi.h"

TEA_NAMESPACE_USE

// Midi channel indices
#define MIDI_SPEED_CHANNEL		1
#define MIDI_BELL_CHANNEL		2
#define MIDI_ORCHESTRA_CHANNEL	3
#define MIDI_KNOCK_CHANNEL		10

bool Sound::initialise(void)
{
	MONITOR(Sound::initialise)

	// Initialise the sound instance
	m_active = false;
	m_speed = 0;
	if (!MidiDevice::initialise())
		return false;
	static const MidiDevice::Msg msg[] =
	{
		// Initialise Channel #1
		MIDI_MSG_PROG(MIDI_SPEED_CHANNEL, 126),			// Applause
		MIDI_MSG_PITCH(MIDI_SPEED_CHANNEL, 0),			// Low pitch bend

		// Initialise Channel #2
		MIDI_MSG_PROG(MIDI_BELL_CHANNEL, 14),			// Tubular bells
		MIDI_MSG_VOLUME(MIDI_BELL_CHANNEL, 64),			// Medium volume

		// Initialise Channel #3
		MIDI_MSG_PROG(MIDI_ORCHESTRA_CHANNEL, 55),		// Orchestra hit
		MIDI_MSG_VOLUME(MIDI_ORCHESTRA_CHANNEL, 127),	// High volume

		// Percussion Channel #10
		MIDI_MSG_PROG(MIDI_KNOCK_CHANNEL, 0),			// General
		MIDI_MSG_VOLUME(MIDI_KNOCK_CHANNEL, 127),		// High volume
		MIDI_MSG_PITCH(MIDI_KNOCK_CHANNEL, 0)			// Low pitch bend
	};
	m_active = MidiDevice::playAsync(msg, dimof(msg));
	return m_active;
}

bool Sound::speed(int speed)
{
	MONITOR(Sound::speed)

	// Set the speed of running stones
	VALIDATE(*this);
	ASSERT((speed >= 0) && (speed <= 127));
	if (!m_active || (speed == m_speed))
		return true;
	if (speed == 0)
	{
		// Turn off the channel
		m_speed = 0;
		return MidiDevice::playAsync(MIDI_MSG_OFF(MIDI_SPEED_CHANNEL, 0, 0));
	}
	if (m_speed == 0)
	{
		// Turn on the channel
		m_speed = speed;
		MidiDevice::Msg msg[] =
		{
			MIDI_MSG_VOLUME(MIDI_SPEED_CHANNEL, speed),
			MIDI_MSG_ON(MIDI_SPEED_CHANNEL, 0, 127)
		};
		return MidiDevice::playAsync(msg, dimof(msg));
	}
	m_speed = speed;
	return MidiDevice::playOutOfBounds(MIDI_MSG_VOLUME(MIDI_SPEED_CHANNEL, speed));
}

bool Sound::bell(int semitones)
{
	MONITOR(Sound::bell)

	// Play a bell around Middle C
	VALIDATE(*this);
	if (!m_active)
		return true;
	semitones += 60;
	ASSERT((semitones >= 0) && (semitones <= 127));
	return MidiDevice::playAsync(MIDI_MSG_ON(MIDI_BELL_CHANNEL, semitones, 64));
}

bool Sound::orchestra(void)
{
	MONITOR(Sound::orchestra)

	// Play an orchestra hit at Middle C
	VALIDATE(*this);
	if (!m_active)
		return true;
	return MidiDevice::playAsync(MIDI_MSG_ON(MIDI_ORCHESTRA_CHANNEL, 60, 64));
}

bool Sound::knock(int speed)
{
	MONITOR(Sound::initialise)

	// Side stick
	VALIDATE(*this);
	if (!m_active)
		return true;
	ASSERT((speed > 0) && (speed <= 127));
	return MidiDevice::playAsync(MIDI_MSG_ON(MIDI_KNOCK_CHANNEL, 37, speed));
}

void Sound::shutdown(void)
{
	MONITOR(Sound::shutdown)

	// Shutdown the sound instance
	VALIDATE(*this);
	if (m_active)
		MidiDevice::shutdown();
	m_active = false;
	m_speed = 0;
	VALIDATE(*this);
	return;
}
