//----------------------------------------------------------------------------
//	HEADER:	TDebug.h
//	AUTHOR:	Ian Taylor
//	NOTES:	Debugging helpers
//----------------------------------------------------------------------------

#ifndef TEA_DEBUG_H
#define TEA_DEBUG_H
#ifdef TEA_DEBUG_INCLUDES
#pragma message("#include " __FILE__)
#endif

// Debug levels (more information the higher the level)
#define TEA_DEBUG_LEVEL_NONE		0
#define TEA_DEBUG_LEVEL_ASSERT		1
#define TEA_DEBUG_LEVEL_VALIDATE	2
#define TEA_DEBUG_LEVEL_FULL		9

// Make sure TEA_DEBUG_LEVEL is set (0-9)
#ifndef TEA_DEBUG_LEVEL
#ifdef _DEBUG
#define TEA_DEBUG_LEVEL				TEA_DEBUG_LEVEL_FULL
#else
#define TEA_DEBUG_LEVEL				TEA_DEBUG_LEVEL_NONE
#endif
#endif

#if TEA_DEBUG_LEVEL > 0

TEA_NAMESPACE_BEGIN

// Debug functions
namespace Debug
{
	inline bool validate(bool value) { return value; }
	template<typename T> inline bool validate(const T& value) { return value.isValid(); }
};

TEA_NAMESPACE_END

// Only in debug target
#define USE(var)				while (&(var) != &(var))
#define BREAKPOINT(why)			if (TEA::System::debugFormat("%s(%d): %s\r\n", __FILE__, __LINE__, why)) TEA_BREAK() else TEA_NOP()
#define BREAKPOINT_(expr, why)	if (expr) TEA_NOP(); else BREAKPOINT(why)

// Debug target
#define TRACE(what)				System::debugFormat##what
#define UNREACHABLE()			BREAKPOINT("Unreachable code reached")
#define ASSERT(expr)			BREAKPOINT_(expr, "Assertion failure: " #expr)
#define VERIFY(expr)			BREAKPOINT_(expr, "Verification failure: " #expr)
#define VERIFY_(expr, cmp)		BREAKPOINT_(expr cmp, "Verification failure: " #expr " " #cmp)

#else

// Release target
#define TRACE(what)				TEA_NOP()
#define UNREACHABLE()			TEA_ASSUME(0)
#define ASSERT(expr)			TEA_ASSUME(expr)
#define VERIFY(expr)			(expr)
#define VERIFY_(expr, cmp)		(expr)

#endif

// Validation only at higher debug levels
#if TEA_DEBUG_LEVEL >= TEA_DEBUG_LEVEL_VALIDATE
#define VALIDATE(expr)			BREAKPOINT_(TEA::Debug::validate(expr), "Validation failure: " #expr)
#else
#define VALIDATE(expr)			TEA_NOP()
#endif

#endif
