//----------------------------------------------------------------------------
//	SOURCE:	TMath.cpp
//	AUTHOR:	Ian Taylor
//	NOTES:	Tea math functions
//----------------------------------------------------------------------------

#include "Tea.h"

TEA_NAMESPACE_USE

// Root accuracy (debug only)
#define ASSERT_ROOT(x, y) //ASSERT(Math::isNear(x, y, 0.001))

// Finer control for surreals
static const surreal g_root_epsilon = 1.e-15;

int Math::solvePoly(const surreal coeff[], int coeffs, surreal root[])
{
	MONITOR(Math::solvePoly)

	// Solve a generic polynomial
	ASSERT((coeffs >= 2) && (coeffs <= 5));
	switch (coeffs)
	{
		case 2:
			return solveLinear(coeff[0], coeff[1], root);

		case 3:
			return solveQuadratic(coeff[0], coeff[1], coeff[2], root);

		case 4:
			return solveCubic(coeff[0], coeff[1], coeff[2], coeff[3], root);

		case 5:
			return solveQuartic(coeff[0], coeff[1], coeff[2], coeff[3], coeff[4], root);
	}
	return -1;
}

int Math::solveLinear(surreal a, surreal b, surreal root[1])
{
	MONITOR(Math::solveLinear)

	// Solve: ax + b = 0
	if (isNearZero(a, 0.000001))
		return 0;
	root[0] = -b / a;
	ASSERT_ROOT(fnLinear(a, 0, root[0]), -b);
	return 1;
}

static surreal Math__quadraticRoot(surreal b, surreal c, surreal x0)
{
	MONITOR(Math__quadraticRoot)

	// Use Newton-Raphson to find the "nearest" x where: x2 + bx + c = 0
	surreal y = Math::fnQuadratic(1, b, c, x0);
	surreal yp = Math::abs(y);
	if (yp < g_root_epsilon)
		return x0;
	surreal x = x0;
    surreal h = 0;
	surreal hp = 0;
	for (int tries = 0; tries < 30; tries++)
	{
		// Evaluate the first derivative
		surreal dydx = Math::fnLinear(2, b, x);
		if (Math::isNearZero(dydx, g_root_epsilon))
		{
			// Derivative is zero; half the latest adjustment and try again
			if (h == 0)
				break;
			h /= 2;
			x -= h;
		}
		else
		{
			// Evaluate the error
			h = -y / dydx;
			surreal ha = Math::abs(h);
			if ((ha < g_root_epsilon) || ((tries > 0) && (ha >= hp)))
				break;
			hp = ha;

			// Work out if this is an improvement
			y = Math::fnQuadratic(1, b, c, x + h);
			surreal ya = Math::abs(y);
			if (ya > yp)
				break;
			yp = ya;
			x += h;
		}
	}
	ASSERT(tries <= 5);
	ASSERT(Math::abs(Math::fnQuadratic(1, b, c, x)) <= Math::abs(Math::fnQuadratic(1, b, c, x0)));
	return x;
}

int Math::solveQuadratic(surreal a, surreal b, surreal c, surreal root[2])
{
	MONITOR(Math::solveQuadratic)

	// Solve: ax2 + bx + c = 0
	if (isNearZero(a, 0.000001))
		return solveLinear(b, c, root);
	surreal p = b / a;
	surreal q = c / a;
	surreal s = sqr(p) - 4 * q;
	if (s <= -Epsilon)
		return 0;
	if (s < Epsilon)
	{
		// One real root
		root[0] = Math__quadraticRoot(p, q, -p / 2);
		ASSERT_ROOT(fnQuadratic(a, b, 0, root[0]), -c);
		return 1;
	}
	s = sqrt(s);
	root[0] = Math__quadraticRoot(p, q, (p + s) / -2);
	root[1] = Math__quadraticRoot(p, q, (p - s) / -2);
	ASSERT_ROOT(fnQuadratic(a, b, 0, root[0]), -c);
	ASSERT_ROOT(fnQuadratic(a, b, 0, root[1]), -c);
	ASSERT(root[0] < root[1]);
	return 2;
}

static surreal Math__cubicRoot(surreal b, surreal c, surreal d, surreal x0)
{
	MONITOR(Math__cubicRoot)

	// Use Newton-Raphson to find the "nearest" x where: x3 + bx2 + cx + d = 0
	surreal y = Math::fnCubic(1, b, c, d, x0);
	surreal yp = Math::abs(y);
	if (yp < g_root_epsilon)
		return x0;
	surreal x = x0;
    surreal h = 0;
	surreal hp = 0;
	for (int tries = 0; tries < 20; tries++)
	{
		// Evaluate the first derivative
		surreal dydx = Math::fnQuadratic(3, 2 * b, c, x);
		if (Math::isNearZero(dydx, g_root_epsilon))
		{
			// Derivative is zero; half the latest adjustment and try again
			if (h == 0)
				break;
			h /= 2;
			x -= h;
		}
		else
		{
			// Evaluate the error
			h = -y / dydx;
			surreal ha = Math::abs(h);
			if ((ha < g_root_epsilon) || ((tries > 0) && (ha >= hp)))
				break;
			hp = ha;

			// Work out if this is an improvement
			y = Math::fnCubic(1, b, c, d, x + h);
			surreal ya = Math::abs(y);
			if (ya > yp)
				break;
			yp = ya;
			x += h;
		}
	}
	ASSERT(tries <= 15);
	ASSERT(Math::abs(Math::fnCubic(1, b, c, d, x)) <= Math::abs(Math::fnCubic(1, b, c, d, x0)));
	return x;
}

int Math::solveCubic(surreal a, surreal b, surreal c, surreal d, surreal root[3])
{
	MONITOR(Math::solveCubic)

	// Solve: x3 + bx2 + cx + d = 0
	if (isNearZero(a, 0.000001))
		return solveQuadratic(b, c, d, root);
	surreal p = b / a;
	surreal q = c / a;
	surreal r = d / a;
	surreal u = q - p * p / 3;
	surreal v = r - p * q / 3 + 2 * p * p * p / 27;
	surreal j = 4 * u * u * u / 27 + v * v;
	if (j > -Epsilon)
	{
		if (j >= Epsilon)
		{
			// One real root, two imaginary
			surreal w = sqrt(j);
			if (v < 0)
				w = nroot((w - v) / 2, 3);
			else
				w = -nroot((w + v) / 2, 3);
			root[0] = Math__cubicRoot(p, q, r, w - u / (3 * w) - p / 3);
			ASSERT_ROOT(fnCubic(a, b, c, 0, root[0]), -d);
			return 1;
		}
		if ((u >= 0) || isNearZero(v))
		{
			// One real, equal root
			root[0] = -nroot(r, 3);
			ASSERT_ROOT(fnCubic(a, b, c, 0, root[0]), -d);
			return 1;
		}

		// Two real roots
		surreal j = nroot(v / 2, 3);
		root[0] = Math__cubicRoot(p, q, r, j - p / 3);
		root[1] = Math__cubicRoot(p, q, r, -2 * j - p / 3);
		order(root[0], root[1]);
		ASSERT_ROOT(fnCubic(a, b, c, 0, root[0]), -d);
		ASSERT_ROOT(fnCubic(a, b, c, 0, root[1]), -d);
		ASSERT(root[0] < root[1]);
		return 2;
	}

	// Three real roots
	ASSERT(u < 0);
	surreal s = sqrt(u / -3);
	surreal t = -v / (2 * s * s * s);
	if (t >= 1)
	{
		// Maybe not
		root[0] = Math__cubicRoot(p, q, r, -s - p / 3);
		root[1] = Math__cubicRoot(p, q, r, 2 * s * 1 - p / 3);
		ASSERT_ROOT(fnCubic(a, b, c, 0, root[0]), -d);
		ASSERT_ROOT(fnCubic(a, b, c, 0, root[1]), -d);
		ASSERT(root[0] < root[1]);
		return 2;
	}
	surreal sin_k, cos_k;
	System::mathSinCos(System::mathAcos(t) / 3, sin_k, cos_k);
	ASSERT((sin_k > 0) && (sin_k <= 1));
	ASSERT((cos_k > 0) && (cos_k <= 1));
	surreal sin_k3 = sin_k * sqrt(3.0);
	root[0] = Math__cubicRoot(p, q, r, s * (-cos_k - sin_k3) - p / 3);
	root[1] = Math__cubicRoot(p, q, r, s * (-cos_k + sin_k3) - p / 3);
	root[2] = Math__cubicRoot(p, q, r, 2 * s * cos_k - p / 3);
	ASSERT_ROOT(fnCubic(a, b, c, 0, root[0]), -d);
	ASSERT_ROOT(fnCubic(a, b, c, 0, root[1]), -d);
	ASSERT_ROOT(fnCubic(a, b, c, 0, root[2]), -d);
	order(root[0], root[2]);
	order(root[0], root[1]);
	order(root[1], root[2]);
	ASSERT(root[0] < root[1]);
	ASSERT(root[1] < root[2]);
	return 3;
}

int Math::solveBiquadratic(surreal a, surreal c, surreal e, surreal root[4])
{
	MONITOR(Math::solveBiquadratic)

	// Solve: ax4 + cx2 + e = 0
	if (isNearZero(a, 0.000001))
		return solveQuadratic(c, 0, e, root);
	int roots = 0;
	surreal z[2];
	switch (solveQuadratic(a, c, e, z))
	{
		case 1:
			// Single root
			if (z[0] < 0)
				return 0;
			root[roots++] = -sqrt(z[0]);
			ASSERT_ROOT(fnQuartic(a, 0, c, 0, 0, root[0]), -e);
			if (root[0] < 0)
				root[roots++] = -root[0];
			break;

		case 2:
			// Double root
			if (z[0] < 0)
			{
				// First root is imaginary
				if (z[1] < 0)
					return 0;
				root[roots++] = -sqrt(z[1]);
				ASSERT_ROOT(fnQuartic(a, 0, c, 0, 0, root[0]), -e);
				if (root[0] < 0)
					root[roots++] = -root[0];
			}
			else
			{
				root[roots++] = -sqrt(z[1]);
				ASSERT_ROOT(fnQuartic(a, 0, c, 0, 0, root[0]), -e);
				root[roots++] = -sqrt(z[0]);
				ASSERT_ROOT(fnQuartic(a, 0, c, 0, 0, root[1]), -e);
				if (root[1] < 0)
					root[roots++] = -root[1];
				root[roots++] = -root[0];
				ASSERT_ROOT(fnQuartic(a, 0, c, 0, 0, root[2]), -e);
			}
	}
	return roots;
}

static surreal Math__quarticBracket(surreal b, surreal c, surreal d, surreal e, surreal x0, surreal x1)
{
	MONITOR(Math__quarticBracket)

	// Use bracketing to find the intermediate x where: x4 + bx3 + cx2 + dx + e = 0
	ASSERT(x0 < x1);
	surreal y0 = Math::fnQuartic(1, b, c, d, e, x0);
	surreal y1 = Math::fnQuartic(1, b, c, d, e, x1);
	while ((y0 * y1) > 0)
	{
		// Need to bracket the solution properly
		if (Math::abs(y0) < Math::abs(y1))
		{
			// Move the left end
			x0 -= (x1 - x0);
			y0 = Math::fnQuartic(1, b, c, d, e, x0);
		}
		else
		{
			// Move the right end
			x1 += (x1 - x0);
			y1 = Math::fnQuartic(1, b, c, d, e, x1);
		}
	}
	surreal x, y;
	ASSERT((y0 * y1) <= 0);
	for (;;)
	{
		x = (x1 + x0) / 2;
		y = Math::fnQuartic(1, b, c, d, e, x);
		if (Math::isNearZero(y, g_root_epsilon))
			break;
		if ((y * y0) > 0)
		{
			// Replace left
			if (x <= x0)
			{
				// use the best end-point
				if (Math::abs(y0) < Math::abs(y1))
				{
					x = x0;
					y = y0;
				}
				else
				{
					x = x1;
					y = y1;
				}
				break;
			}
			x0 = x;
			y0 = y;
		}
		else
		{
			// Replace right
			if (x >= x1)
			{
				// use the best end-point
				if (Math::abs(y0) < Math::abs(y1))
				{
					x = x0;
					y = y0;
				}
				else
				{
					x = x1;
					y = y1;
				}
				break;
			}
			x1 = x;
			y1 = y;
		}
		ASSERT(x0 < x1);
	}
	ASSERT(Math::abs(y) <= Math::abs(y0));
	ASSERT(Math::abs(y) <= Math::abs(y1));
	ASSERT_ROOT(Math::fnQuartic(1, b, c, d, 0, x), -e);
	return x;
}

static int Math__quarticRoot1(surreal b, surreal c, surreal d, surreal e, surreal tpx, surreal tpy, surreal dx, surreal root[2])
{
	MONITOR(Math__quarticRoot1)

	// Find 0, 1 or 2 roots around a single turning point
	if (tpy > -Math::Epsilon)
	{
		if (tpy > Math::Epsilon)
			return 0;

		// Single root
		root[0] = tpx;
		ASSERT_ROOT(Math::fnQuartic(1, b, c, d, 0, root[0]), -e);
		return 1;
	}
	ASSERT(dx > 0);
	root[0] = Math__quarticBracket(b, c, d, e, tpx - dx, tpx);
	root[1] = Math__quarticBracket(b, c, d, e, tpx, tpx + dx);
	ASSERT(root[0] < root[1]);
	return 2;
}

static int Math__quarticRoot2(surreal b, surreal c, surreal d, surreal e, surreal tpx[2], surreal tpy[2], surreal root[2])
{
	MONITOR(Math__quarticRoot2)

	// Find 0, 1 or 2 roots around two turning points
	surreal dx = (tpx[1] - tpx[0]) / 2;
	ASSERT(dx > 0);
	if (tpy[0] >= 0)
		return Math__quarticRoot1(b, c, d, e, tpx[1], tpy[1], dx, root);
	if (tpy[1] >= 0)
		return Math__quarticRoot1(b, c, d, e, tpx[0], tpy[0], dx, root);
	ASSERT(dx > 0);
	root[0] = Math__quarticBracket(b, c, d, e, tpx[0] - dx, tpx[0]);
	root[1] = Math__quarticBracket(b, c, d, e, tpx[1], tpx[1] + dx);
	ASSERT(root[0] < root[1]);
	return 2;
}

int Math::solveQuartic(surreal a, surreal b, surreal c, surreal d, surreal e, surreal root[4])
{
	MONITOR(Math::solveQuartic)

	// Solve: ax4 + bx3 + cx2 + dx + e = 0
	if (isNearZero(a, 0.000001))
		return solveCubic(b, c, d, e, root);

	// Rework to: x4 + bx3 + cx2 + dx + e = 0
	b /= a;
	c /= a;
	d /= a;
	e /= a;

	// Work out turning points
	surreal tpx[3], tpy[3], dx;
	int i, tps = solveCubic(4, 3 * b, 2 * c, d, tpx);
	for (i = 0; i < tps; i++)
		tpy[i] = fnQuartic(1, b, c, d, e, tpx[i]);
	switch (tps)
	{
		case 1:
			// Single turning point; at most two roots
			dx = 1 - tpy[0];
			return Math__quarticRoot1(b, c, d, e, tpx[0], tpy[0], dx, root);

		case 2:
			// Two turning points; at most two roots
			return Math__quarticRoot2(b, c, d, e, tpx, tpy, root);
	}
	ASSERT(tps == 3);

	// Three turning points; up to four roots
	if (tpy[0] > -Epsilon)
	{
		if (tpy[0] > Epsilon)
			return Math__quarticRoot2(b, c, d, e, &tpx[1], &tpy[1], root);

		// Three roots (repeated left)
		root[0] = tpx[0];
		ASSERT_ROOT(fnQuartic(1, b, c, d, 0, root[0]), -e);
		return Math__quarticRoot2(b, c, d, e, &tpx[1], &tpy[1], &root[1]) + 1;
	}
	if (tpy[2] > -Epsilon)
	{
		if (tpy[2] > Epsilon)
			return Math__quarticRoot2(b, c, d, e, &tpx[1], &tpy[1], root);

		// Three roots (repeated right)
		int n = Math__quarticRoot2(b, c, d, e, tpx, tpy, root);
		root[n] = tpx[2];
		ASSERT_ROOT(fnQuartic(1, b, c, d, 0, root[n]), -e);
		ASSERT((n == 0) || (root[n - 1] < root[n]));
		return n + 1;
	}
	if (tpy[1] < Epsilon)
	{
		if (tpy[1] < -Epsilon)
		{
			// Two roots
			dx = (tpx[2] - tpx[0]) / 2;
			ASSERT(dx > 0);
			root[0] = Math__quarticBracket(b, c, d, e, tpx[0] - dx, tpx[0]);
			root[1] = Math__quarticBracket(b, c, d, e, tpx[2], tpx[2] + dx);
			ASSERT(root[0] < root[1]);
			return 2;
		}

		// Three roots (repeated centre)
		dx = min(tpx[2] - tpx[1], tpx[1] - tpx[0]) / 2;
		ASSERT(dx > 0);
		root[0] = Math__quarticBracket(b, c, d, e, tpx[0] - dx, tpx[0]);
		root[1] = tpx[1];
		root[2] = Math__quarticBracket(b, c, d, e, tpx[2], tpx[2] + dx);
		ASSERT_ROOT(fnQuartic(1, b, c, d, 0, root[1]), -e);
		ASSERT(root[0] < root[1]);
		ASSERT(root[1] < root[2]);
		return 3;
	}

	// Four standard roots
	dx = (tpx[1] - tpx[0]) / 2;
	ASSERT(dx > 0);
	root[0] = Math__quarticBracket(b, c, d, e, tpx[0] - dx, tpx[0]);
	root[1] = Math__quarticBracket(b, c, d, e, tpx[0], tpx[1]);
	dx = (tpx[2] - tpx[1]) / 2;
	ASSERT(dx > 0);
	root[2] = Math__quarticBracket(b, c, d, e, tpx[1], tpx[2]);
	root[3] = Math__quarticBracket(b, c, d, e, tpx[2], tpx[2] + dx);
	ASSERT(root[0] < root[1]);
	ASSERT(root[1] < root[2]);
	ASSERT(root[2] < root[3]);
	return 4;
}

int Math::negativeQuadratic(surreal a, surreal b, surreal c, surreal negative[3])
{
	MONITOR(Math::negativeQuadratic)

	// Determine the range of x where ax2 + bx + c < 0
	if (a < 0.000001)
		return 0;
	surreal p = b / a;
	surreal q = c / a;
	surreal s = sqr(p) - 4 * q;
	if (s <= 0)
		return 0;
	s = sqrt(s);
	negative[0] = Math__quadraticRoot(p, q, (p + s) / -2);
	negative[1] = p / -2;
	negative[2] = Math__quadraticRoot(p, q, (p - s) / -2);
	ASSERT(negative[0] < negative[1]);
	ASSERT(negative[1] < negative[2]);
	return 1;
}

int Math__negativeCubicRange(surreal a, surreal b, surreal c, surreal d, surreal x1, surreal x2, surreal x3, surreal negative[3])
{
	MONITOR(Math__negativeCubicRange)

	// Determine if a range between two roots is negative
	ASSERT(x1 < x2);
	ASSERT(x2 < x3);
	if (Math::fnCubic(a, b, c, d, x2) >= 0)
		return 0;
	negative[0] = x1;
	negative[1] = x2;
	negative[2] = x3;
	return 1;
}

int Math::negativeCubic(surreal a, surreal b, surreal c, surreal d, surreal negative[3])
{
	MONITOR(Math::solveQuartic)

	// Determine the range of x where ax3 + bx2 + cx + d < 0
	if (isNearZero(a, 0.000001))
		return negativeQuadratic(b, c, d, negative);

	// Direct from roots
	surreal root[3], qroot[2];
	switch (solveCubic(a, b, c, d, root))
	{
		case 2:
			// Two roots
			VERIFY_(solveQuadratic(3 * a, 2 * b, c, qroot),== 2);
			if (abs(fnCubic(a, b, c, d, qroot[0])) > abs(fnCubic(a, b, c, d, qroot[1])))
				return Math__negativeCubicRange(a, b, c, d, root[0], qroot[0], root[1], negative);
			return Math__negativeCubicRange(a, b, c, d, root[0], qroot[1], root[1], negative);

		case 3:
			// Three roots
			VERIFY_(solveQuadratic(3 * a, 2 * b, c, qroot),== 2);
			return Math__negativeCubicRange(a, b, c, d, root[0], qroot[0], root[1], negative) ||
				   Math__negativeCubicRange(a, b, c, d, root[1], qroot[1], root[2], negative);
	}
	return 0;
}

int Math__negativeQuarticRange(surreal b, surreal c, surreal d, surreal e, surreal tpx, surreal tpy, surreal dx, surreal negative[3])
{
	MONITOR(Math__negativeQuarticRange)

	// Determine a range of x where x4 + bx3 + cx2 + dx + e < 0
	if (tpy >= 0)
		return 0;
	ASSERT(dx > 0);
	negative[0] = Math__quarticBracket(b, c, d, e, tpx - dx, tpx);
	negative[1] = tpx;
	negative[2] = Math__quarticBracket(b, c, d, e, tpx, tpx + dx);
	if ((negative[0] >= negative[1]) || (negative[1] >= negative[2]))
		return 0;
	return 1;
}

int Math::negativeQuartic(surreal b, surreal c, surreal d, surreal e, surreal negative[2][3])
{
	MONITOR(Math::solveQuartic)

	// Determine the ranges of x where x4 + bx3 + cx2 + dx + e < 0
	surreal tpx[3], tpy[3], dx;
	int i, tps = solveCubic(4, 3 * b, 2 * c, d, tpx);
	for (i = 0; i < tps; i++)
		tpy[i] = fnQuartic(1, b, c, d, e, tpx[i]);
	switch (tps)
	{
		case 1:
			// Single turning point; at most one range
			dx = 1 - tpy[0];
			return Math__negativeQuarticRange(b, c, d, e, tpx[0], tpy[0], dx, negative[0]);

		case 2:
			// Two turning points; at most one range
			dx = 1 + (tpx[1] - tpx[0]);
			return Math__negativeQuarticRange(b, c, d, e, tpx[0], tpy[0], dx, negative[0]) ||
				   Math__negativeQuarticRange(b, c, d, e, tpx[1], tpy[1], dx, negative[0]);
	}
	ASSERT(tps == 3);

	// Three turning points; up to two ranges
	if (tpy[1] >= 0)
	{
		// Classic 'W' shape with up to two ranges
		int n = 0;
		dx = tpx[1] - tpx[0];
		n += Math__negativeQuarticRange(b, c, d, e, tpx[0], tpy[0], dx, negative[0]);
		dx = tpx[2] - tpx[1];
		n += Math__negativeQuarticRange(b, c, d, e, tpx[2], tpy[2], dx, negative[n]);
		return n;
	}
	if ((tpy[0] >= tpy[1]) || (tpy[2] >= tpy[1]))
	{
		// Insane but can occur with rounding!
		return 0;
	}
	dx = tpx[2] - tpx[0];
	ASSERT(dx > 0);
	negative[0][0] = Math__quarticBracket(b, c, d, e, tpx[0] - dx, tpx[0]);
	negative[0][1] = tpx[0];
	negative[0][2] = Math__quarticBracket(b, c, d, e, tpx[2], tpx[2] + dx);
	return 1;
}
