//----------------------------------------------------------------------------
//	HEADER:	TMath.h
//	AUTHOR:	Ian Taylor
//	NOTES:	Mathematical functions
//----------------------------------------------------------------------------

#ifndef TEA_MATH_H
#define TEA_MATH_H
#ifdef TEA_DEBUG_INCLUDES
#pragma message("#include " __FILE__)
#endif

// Floating point constants
#define TEA_MATH_PI							3.1415926535897932384626433832795
#define TEA_MATH_EPSILON					REAL_EPSILON

// Integer range calculations
#define _TEA_INTEGER_MAX(bits)				((1 << (bits)) - 1)
#define _TEA_INTEGER_MIN(bits)				(~_TEA_INTEGER_MAX(bits))
#define _TEA_INTEGER_SIGNED(prefix, type)	prefix##Min = _TEA_INTEGER_MIN(sizeof(type) * 8 - 1),	\
											prefix##Max = _TEA_INTEGER_MAX(sizeof(type) * 8 - 1),
#define _TEA_INTEGER_UNSIGNED(prefix, type)	prefix##Min = 0,	\
											prefix##Max = _TEA_INTEGER_MAX(sizeof(type) * 8),

TEA_NAMESPACE_BEGIN

namespace Math
{
	// Integer ranges
	enum IntegerRange
	{
		_TEA_INTEGER_SIGNED(short, short)
		_TEA_INTEGER_UNSIGNED(ushort, unsigned short)
		_TEA_INTEGER_SIGNED(int, int)
		_TEA_INTEGER_UNSIGNED(uint, unsigned int)
		_TEA_INTEGER_SIGNED(long, long)
		_TEA_INTEGER_UNSIGNED(ulong, unsigned long)
	};

	// Floating point constants
	static const real Epsilon = REAL(TEA_MATH_EPSILON);
	static const real Pi = REAL(TEA_MATH_PI);

	// Integer operations
	inline int imod(int x, int y)
	{
		// Integer modulo guaranteed to be positive (unlike Microsoft's)
		MONITOR(Math::imod)
		ASSERT(y > 0);
		int z = x % y;
		if (z < 0)
			z += y;
		ASSERT((z >= 0) && (z < y));
		return z;
	}
	inline int idiv(int& x, int y)
	{
		// Integer division with remainder
		MONITOR(Math::idiv)
		ASSERT(x >= 0);
		ASSERT(y > 0);
		int z = x % y;
		x /= y;
		ASSERT((z >= 0) && (z < y));
		return z;
	}
	inline int iround(int x, int y)
	{
		// Round an integer 'x' up to the nearest multiple of 'y'
		MONITOR(Math::iround)
		ASSERT(x >= 0);
		ASSERT(y > 0);
		int z = ((x - 1) / y + 1) * y;
		ASSERT(z >= x);
		ASSERT((z % y) == 0);
		return z;
	}

	// Epsilon comparisons
	template<typename T>
	inline bool isNearZero(T x, T epsilon = T(TEA_MATH_EPSILON))
	{
		// return x == 0 ~ epsilon;
		MONITOR(Math::isNearZero)
		return (x > -epsilon) && (x < epsilon);
	}
	template<typename T>
	inline bool isNear(T x, T y, T epsilon = T(TEA_MATH_EPSILON))
	{
		// return x ~= y;
		MONITOR(Math::isNear)
		return ((y > 1) || (y < -1)) ? isNearZero((x - y) / y, epsilon) : isNearZero(x - y, epsilon);
	}

	// Conversions
	inline long truncateLong(real value)
	{
		// Truncate towards zero
		MONITOR(Math::truncateLong)
		ASSERT((value >= longMin) && (value <= longMax));
		long result = System::mathTruncate(value);
		ASSERT((value > -1) || ((result >= value) && (result < (value + 1))));
		ASSERT((value < 1) || ((result > (value - 1)) && (result <= value)));
		ASSERT((value < -1) || (value > 1) || (result == 0));
		return result;
	}
	inline int truncate(real value)
	{
		// Truncate towards zero
		MONITOR(Math::truncateLong)
		ASSERT((value >= intMin) && (value <= intMax));
		int result = static_cast<int>(System::mathTruncate(value));
		ASSERT((value > -1) || ((result >= value) && (result < (value + 1))));
		ASSERT((value < 1) || ((result > (value - 1)) && (result <= value)));
		ASSERT((value < -1) || (value > 1) || (result == 0));
		return result;
	}
	inline long roundLong(real value)
	{
		// Round to the nearest long
		MONITOR(Math::roundLong)
		ASSERT((value >= longMin) && (value <= longMax));
		long result = System::mathRound(value);
		ASSERT((value - result) >= REAL(-0.5));
		ASSERT((value - result) <= REAL(0.5));
		return result;
	}
	inline int round(real value)
	{
		// Round to the nearest integer
		MONITOR(Math::roundNearest)
		ASSERT((value >= intMin) && (value <= intMax));
		int result = static_cast<int>(System::mathRound(value));
		ASSERT((value - result) >= REAL(-0.5));
		ASSERT((value - result) <= REAL(0.5));
		return result;
	}
	inline int roundUp(real value)
	{
		// Round up to the next integer
		MONITOR(Math::roundUp)
		ASSERT((value >= intMin) && (value <= intMax));
		int result = static_cast<int>(System::mathRoundUp(value));
		ASSERT(result >= value);
		ASSERT((result - value) < REAL(1.0));
		return result;
	}

	// Powers
	template<typename T>
	inline T sqr(T x)
	{
		// Square
		MONITOR(Math::sqr)
		return x * x;
	}
	template<typename T>
	inline T cube(T x)
	{
		// Cube
		MONITOR(Math::cube)
		return x * x * x;
	}
	template<typename T>
	inline T power(T x, T y)
	{
		// Power
		MONITOR(Math::power)
		ASSERT(y >= 0);
		T z = (T)System::mathPower(x, y);
		ASSERT(z >= 0);
		return z;
	}
	template<typename T>
	inline T sqrt(T x)
	{
		// Square root (assuming x is not negative)
		MONITOR(Math::sqrt)
		ASSERT(x >= 0);
		T y = (T)System::mathSqrt(x);
		ASSERT(isNear(x, y * y));
		return y;
	}
	template<typename T>
	inline T sqrtSafe(T x)
	{
		// Square root (returns zero if x is negative)
		MONITOR(Math::sqrt)
		T y = (T)System::mathSqrt(x);
		ASSERT((x <= 0) ? (y == 0) : (isNearZero(y) || isNear(x, y * y)));
		return y;
	}
	template<typename T>
	inline T nroot(T x, int n)
	{
		// Nth root
		MONITOR(Math::nroot)
		ASSERT(n > 0);
		if (x < 0)
		{
			// Negative roots only for odd powers
			if (n & 1)
				return -(T)System::mathPower(-x, 1.0 / n);
			return 0;
		}
		return (T)System::mathPower(x, 1.0 / n);
	}
	template<typename T>
	inline T hypot(T x, T y)
	{
		// Pythagorean hypot
		MONITOR(Math::sqrt)
		return (T)System::mathSqrt(x * x + y * y);
	}

	// Trigonometry
	template<typename T>
	inline T deg2rad(T degrees)
	{
		// Convert degrees to radians
		MONITOR(Math::deg2rad)
		return degrees * T(TEA_MATH_PI) / 180;
	}
	template<typename T>
	inline T rad2deg(T radians)
	{
		// Convert radians to degrees
		MONITOR(Math::rad2deg)
		return radians / T(TEA_MATH_PI) * 180;
	}
	template<typename T>
	inline void sincos(T radians, T& sin, T& cos)
	{
		// Dual sine and cosine calculations
		MONITOR(Math::sincos)
		System::mathSinCos(radians, sin, cos);
		ASSERT((sin >= -1) && (sin <= 1));
		ASSERT((cos >= -1) && (cos <= 1));
		ASSERT(isNear(sin * sin + cos * cos, T(1)));
	}
	template<typename T>
	inline T sin(T radians)
	{
		// Sine
		MONITOR(Math::sin)
		T z = (T)System::mathSin(radians);
		ASSERT((z >= -1) && (z <= 1));
		return z;
	}
	template<typename T>
	inline T cos(T radians)
	{
		// Cosine
		MONITOR(Math::cos)
		T z = (T)System::mathCos(radians);
		ASSERT((z >= -1) && (z <= 1));
		return z;
	}
	template<typename T>
	inline T tan(T radians)
	{
		// Tangent
		MONITOR(Math::tan)
		return (T)System::mathTan(radians);
	}
	template<typename T>
	inline T asin(T x)
	{
		// Arcsine
		MONITOR(Math::asin)
		ASSERT((x >= -1) && (x <= 1));
		T z = (T)System::mathAsin(x);
		ASSERT((z >= REAL(-0.5 * TEA_MATH_PI)) && (z <= REAL(0.5 * TEA_MATH_PI)));
		ASSERT(isNear(sin(z), x));
		return z;
	}
	template<typename T>
	inline T acos(T x)
	{
		// Arccossine
		MONITOR(Math::acos)
		ASSERT((x >= -1) && (x <= 1));
		T z = (T)System::mathAcos(x);
		ASSERT((z >= REAL(0)) && (z <= REAL(TEA_MATH_PI)));
		ASSERT(isNear(cos(z), x));
		return z;
	}
	template<typename T>
	inline T atan(T x, T y = 1)
	{
		// Arctangent
		MONITOR(Math::atan)
		T z = (T)System::mathAtan(x, y);
		ASSERT((z >= REAL(-TEA_MATH_PI)) && (z <= REAL(TEA_MATH_PI)));
		return z;
	}
	template<typename T>
	inline T sin2cos(T x)
	{
		// Find cos(a) when x = sin(a)
		MONITOR(Math::sin2cos)
		ASSERT((x >= -1) && (x <= 1));
		return Math::sqrt(1 - x * x);
	}
	template<typename T>
	inline T cos2sin(T x)
	{
		// Find sin(a) when x = cos(a)
		MONITOR(Math::cos2sin)
		ASSERT((x >= -1) && (x <= 1));
		return Math::sqrt(1 - x * x);
	}
	template<typename T>
	inline T sin2tan(T x)
	{
		// Find tan(a) when x = sin(a)
		MONITOR(Math::sin2tan)
		ASSERT((x > -1) && (x < 1));
		return x / Math::sqrt(1 - x * x);
	}
	template<typename T>
	inline T cos2tan(T x)
	{
		// Find tan(a) when x = cos(a)
		MONITOR(Math::sin2tan)
		ASSERT((x >= -1) && (x <= 1) && !isNearZero(x));
		return Math::sqrt(1 - x * x) / x;
	}
	template<typename T>
	inline void tan2sincos(T x, T y, T& sin, T& cos)
	{
		// Find sin(a) and cos(a) when tan(a) = x/y
		MONITOR(Math::tan2sincos)
		T h = (T)System::mathSqrt(x * x + y * y);
		ASSERT(h > 0);
		sin = x / h;
		cos = y / h;
		ASSERT((sin >= -1) && (sin <= 1));
		ASSERT((cos >= -1) && (cos <= 1));
		ASSERT(isNear(sin * y, cos * x));
		return;
	}

	// Polynomial calculators
	inline surreal fnLinear(surreal a, surreal b, surreal x) { return a * x + b; }
	inline surreal fnQuadratic(surreal a, surreal b, surreal c, surreal x) { return a * x * x + b * x + c; }
	inline surreal fnCubic(surreal a, surreal b, surreal c, surreal d, surreal x) { return ((a * x + b) * x + c) * x + d; }
	inline surreal fnQuartic(surreal a, surreal b, surreal c, surreal d, surreal e, surreal x) { return (((a * x + b) * x + c) * x + d) * x + e; }

	// Solvers always use higher precision
	int solvePoly(const surreal coeff[], int coeffs, surreal root[]);
	int solveLinear(surreal a, surreal b, surreal root[1]);
	int solveQuadratic(surreal a, surreal b, surreal c, surreal root[2]);
	int solveCubic(surreal a, surreal b, surreal c, surreal d, surreal root[3]);
	int solveQuartic(surreal a, surreal b, surreal c, surreal d, surreal e, surreal root[4]);
	int solveBiquadratic(surreal a, surreal c, surreal e, surreal root[4]);

	// Find out where polynomials are negative
	int negativeQuadratic(surreal a, surreal b, surreal c, surreal negative[3]);
	int negativeCubic(surreal a, surreal b, surreal c, surreal d, surreal negative[3]);
	int negativeQuartic(surreal b, surreal c, surreal d, surreal e, surreal negative[2][3]);

	// Templated functions
	template<typename T> inline T abs(T a) { return (a < 0) ? -a : a; }
	template<typename T> inline int sgn(T a) { return (a < 0) ? -1 : (a > 0) ? 1 : 0; }
	template<typename T> inline T min(T a, T b) { return (a < b) ? a : b; }
	template<typename T> inline T max(T a, T b) { return (a < b) ? b : a; }
	template<typename T> inline T clamp(T a, T min, T max) { return (a < min) ? min : (a > max) ? max : a; }
	template<typename T> inline void order(T& a, T& b) { if (b < a) { T c(a); a = b; b = c; } }
}

TEA_NAMESPACE_END

#endif
