//----------------------------------------------------------------------------
//	SOURCE:	TMonitor.cpp
//	AUTHOR:	Ian Taylor
//	NOTES:	Tea monitor class
//----------------------------------------------------------------------------

#include "Tea.h"
#include "TFormat.h"

#if TEA_MONITOR_LEVEL > 0

// For file logging...
#include <stdio.h>

TEA_NAMESPACE_USE

// Global values
static Monitor::Function g_root = { "ROOT" };
static Monitor::Function* g_first = NULL;
static Monitor g_callRoot(g_root);
static Monitor* g_callStack = &g_callRoot;
static Monitor::Time g_clockDead = 0;
static Monitor::Time* g_clockDestination = &g_clockDead;
static System::Timer g_clockBefore = 0;
static FILE* g_fp = NULL;

Monitor::Monitor(Function& function)
	: m_function(function), m_caller(NULL), m_inFunction(0), m_inChildren(0)
{
	// Add this call to the stack trace
	Monitor::assignClock(g_clockDead);
	if (this == &g_callRoot)
	{
		// Initialise the whole thing
		initialise();
	}
	else
	{
		// This is a subsequent function
		if (function.next == NULL)
		{
			// Add this function to the list of called ones
			ASSERT(g_first != NULL);
			function.next = g_first;
			g_first = &function;
		}
		m_caller = g_callStack;
		g_callStack = this;
	}
	function.calls++;
	assignClock(m_inFunction);
}

Monitor::~Monitor(void)
{
	// Remove this call from the stack
	Monitor::assignClock(g_clockDead);
	m_function.in_function += m_inFunction;
	m_function.in_children += m_inChildren;
	if (m_caller == NULL)
	{
		// The dead time included
		Monitor::shutdown(m_inFunction + m_inChildren);
	}
	else
	{
		// Percolate the information to the caller
		m_caller->addChildren(m_inFunction + m_inChildren);
		assignClock(m_caller->m_inFunction);
		g_callStack = m_caller;
	}
}

void Monitor::initialise(void)
{
	// Initialise the clocks, etc
	TRACE(("Starting monitor at level %d\r\n", TEA_MONITOR_LEVEL));
	ASSERT(g_first == NULL);
	g_clockBefore = System::timerGet();
	g_clockDead = 0;
	g_first = &g_root;
	return;
}

void Monitor::shutdown(Time total)
{
	// Shutdown
	for (const Function* function = g_first; function != NULL; function = function->next)
	{
		// Output statistics for this function
		char buffer[32];
		Monitor::output(function->name);
		Monitor::output("() calls=");
		Monitor::output(Format::fromUnsigned(function->calls, buffer, dimof(buffer)));
		Monitor::outputTime(" total=", function->in_function + function->in_children);
		Monitor::outputTime(" sum=", function->in_function);
		ASSERT(function->calls > 0);
		Monitor::outputTime(" avg=", function->in_function / function->calls);
		if (function->file != NULL)
		{
			// Output the source location for reference
			Monitor::output(" source=");
			Monitor::output(function->file);
			Monitor::output(" line=");
			Monitor::output(Format::fromUnsigned(function->line, buffer, dimof(buffer)));
		}
		Monitor::outputLine();
	}
	Monitor::outputTime("Total execution time = ", total);
	Monitor::outputTime(" ms (", g_clockDead);
	Monitor::output(" dead)");
	Monitor::outputLine();
	if (g_fp != NULL)
	{
		// Close the log file
		fclose(g_fp);
		g_fp = NULL;
	}
}

void Monitor::logToFile(const char* filename, const char* mode)
{
	// Open a log file for the monitor
	ASSERT(g_fp == NULL);
	if (filename == NULL)
		return;
	ASSERT(mode != NULL);
	g_fp = fopen(filename, mode);
	if (g_fp == NULL)
		System::debugOutput("Unable to open monitor log file for writing\r\n");
	return;
}

void Monitor::assignClock(Time& elapsed)
{
	// The monitor clock is always 'ticking' at a particular destination
	System::Timer now = System::timerGet();
	*g_clockDestination += System::timerElapsed(g_clockBefore, now);
	g_clockDestination = &elapsed;
	g_clockBefore = now;
	return;
}

void Monitor::outputLine(void)
{
	// Send the current line out to the device
	if (g_fp == NULL)
		System::debugOutput("\r\n");
	else
	{
		fputc('\n', g_fp);
		fflush(g_fp);
	}
	return;
}

void Monitor::outputTime(const char* before, Time time)
{
	// Display a time
	Monitor::output(before);
	long ms = System::timerMicroseconds(time);
	char buffer[32];
	Monitor::output(Format::fromUnsigned(ms / 1000, buffer, dimof(buffer)));
	Monitor::output(".");
	Monitor::output(Format::fromPadded(ms % 1000, buffer, 4, '0'));
	return;
}

void Monitor::output(const char* text)
{
	// Send more text to the device
	if (g_fp == NULL)
		System::debugOutput(text);
	else
		fputs(text, g_fp);
	return;
}

#endif
