//----------------------------------------------------------------------------
//	SOURCE:	TWin32.cpp
//	AUTHOR:	Ian Taylor
//	NOTES:	Tea Windows 32-bit platform specifics
//----------------------------------------------------------------------------

#include "Tea.h"

#include <windows.h>

// There are no explicit return values in the maths functions
#pragma warning(disable: 4035)

// Disable 'inaccurate on some Pentiums' warning
#pragma warning(disable: 4725)

// Fix the stack format for local assembler routines
#define TEA_MATHAPI_IMPL	__declspec(naked) TEA_MATHAPI

double TEA_MATHAPI_IMPL System__mathSqrt(const double& /*ecx*/)
{
	// return ::sqrt(x);
	__asm
	{
		// Compare against zero first
		fld		qword ptr [ecx]
		ftst
		fnstsw	ax
		test	ah, 0x01
		jne		sqrt_negative
		fsqrt
		ret

sqrt_negative:

		// x is negative
		fstp	st
		fldz
		ret
	}
}

float TEA_MATHAPI_IMPL System__mathSqrt(const float& /*ecx*/)
{
	// return ::sqrt(x);
	__asm
	{
		// Compare against zero first
		fld		dword ptr [ecx]
		ftst
		fnstsw	ax
		test	ah, 0x01
		jne		sqrt_negative
		fsqrt
		ret

sqrt_negative:

		// x is negative
		fstp	st
		fldz
		ret
	}
}

void TEA_MATHAPI_IMPL System__mathSinCos(double& /*ecx*/, double& /*edx*/, const double& /*esp+4*/)
{
	// sin = ::sin(x); cos = ::cos(x);
	__asm
	{
		// Assume x is real and in range
		mov		eax, [esp+4]
		fld		qword ptr [eax]
		fsincos
		fstp	qword ptr [edx]
		fstp	qword ptr [ecx]
		ret		4
	}
}

void TEA_MATHAPI_IMPL System__mathSinCos(float& /*ecx*/, float& /*edx*/, const float& /*esp+4*/)
{
	// sin = ::sin(x); cos = ::cos(x);
	__asm
	{
		// Assume x is real and in range
		mov		eax, [esp+4]
		fld		dword ptr [eax]
		fsincos
		fstp	dword ptr [edx]
		fstp	dword ptr [ecx]
		ret		4
	}
}

TEA::int32 TEA_MATHAPI_IMPL System__mathRound(const double& /*ecx*/, TEA::int32& /*edx*/)
{
	// Conversion to 32 bit integer
	__asm
	{
		// EDX holds the control word rounding flags and is also used for workspace
		fld		qword ptr [ecx]
		push	ecx
		fstcw	[edx+2]
		fwait
		mov		ax, [edx+2]
		mov		ecx, eax
		and		ah, 0xF3
		or		[edx], ax
		fldcw	[edx]
		fistp	dword ptr [edx]
		mov		eax, [edx]
		mov		[edx], cx
		fldcw	[edx]
		pop		ecx
		ret
	}
}

TEA::int32 TEA_MATHAPI_IMPL System__mathRound(const float& /*ecx*/, TEA::int32& /*edx*/)
{
	// Conversion to 32 bit integer
	__asm
	{
		// EDX holds the control word rounding flags and is also used for workspace
		fld		dword ptr [ecx]
		push	ecx
		fstcw	[edx+2]
		fwait
		mov		ax, [edx+2]
		mov		ecx, eax
		and		ah, 0xF3
		or		[edx], ax
		fldcw	[edx]
		fistp	dword ptr [edx]
		mov		eax, [edx]
		mov		[edx], cx
		fldcw	[edx]
		pop		ecx
		ret
	}
}

TEA_NAMESPACE_USE

bool System::debugOutput(const char* buffer)
{
	// Output a debug string verbatim
	::OutputDebugString(buffer);
	return ::IsDebuggerPresent() != FALSE;
}

bool TEA_VARARGS System::debugFormat(const char* fmt, ...)
{
	// Output a formatted debug string
	char buffer[1024];
	va_list varargs;
	va_start(varargs, fmt);
	::wvsprintf(buffer, fmt, varargs);
	va_end(varargs);
	return debugOutput(buffer);
}

void System::debugFlush(void)
{
	// Flush the debug output (cannot really, so just hiccup)
	::Sleep(10);
	return;
}

System::Timer System::timerGet(void)
{
	// Fetch a high-performance timer
	LARGE_INTEGER counter;
	if (::QueryPerformanceCounter(&counter))
		return counter.QuadPart;
	return 0;
}

System::TimerElapsed System::timerElapsed(Timer before, Timer after)
{
	// Difference between two timers
	return after - before;
}

long System::timerMicroseconds(TimerElapsed elapsed)
{
	// Convert a counter to microseconds
	static real divisor = 0;
	if (divisor == 0)
	{
		// Fetch the divisor
		LARGE_INTEGER frequency;
		if (!::QueryPerformanceFrequency(&frequency))
			return long(elapsed);
		divisor = real(1000000.0) / frequency.QuadPart;
	}
	return mathRound(elapsed * divisor);
}

long System::errorGetLast(void)
{
	// Fetch the last Win32 error
	DWORD error = ::GetLastError();
	ASSERT((error >= 0x0001) && (error <= 0xFFFF));
	return MAKE_HRESULT(SEVERITY_ERROR, FACILITY_WIN32, error);
}

bool System::errorFormat(char* buffer, size_t bytes, long errcode)
{
	// Try to fetch the message text from the system
	if (SUCCEEDED(errcode))
		return false;
	if (::FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM|FORMAT_MESSAGE_IGNORE_INSERTS|FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL,
						errcode, LANG_USER_DEFAULT, buffer, bytes - 1, NULL) > 0)
	{
		// Terminated at the first control code
		char* end = buffer;
		while (*end >= ' ')
			end++;
		*end = '\0';
		if (buffer[0] != '\0')
			return true;
	}
	::wsprintf(buffer, "Unknown HRESULT 0x%08lX (%lu)", errcode, HRESULT_CODE(errcode));
	return true;
}

void System::errorReport(Window window, const char* caption, const char* reason)
{
	// Report an error
	DWORD flags = MB_ICONEXCLAMATION|MB_OK;
	if (window == NULL)
		flags |= MB_TASKMODAL;
	VERIFY(::MessageBox(window, reason, caption, flags));
	return;
}

int System::randomSeed(void)
{
	// Use the current time to generate a fairly unique seed
	FILETIME filetime;
	::GetSystemTimeAsFileTime(&filetime);
	return ::GetTickCount() * (filetime.dwLowDateTime ^ filetime.dwHighDateTime);
}
