//----------------------------------------------------------------------------
//	HEADER:	TWin32.h
//	AUTHOR:	Ian Taylor
//	NOTES:	System-specific definitions
//----------------------------------------------------------------------------

#ifdef TEA_DEBUG_INCLUDES
#pragma message("#include " __FILE__)
#endif

#ifndef TEA_WIN32_H
#define TEA_WIN32_H

//----------------------------------------------------------------------------
//	FIRST TIME THROUGH THIS HEADER
//----------------------------------------------------------------------------

// In case we later include windows headers
#define WIN32_LEAN_AND_MEAN
#define STRICT
#define NOMINMAX
#define _WIN32_WINNT	0x400

// Straight out of <windows.h>
#define TEA_WINAPI(x)	extern "C" __declspec(dllimport) x __stdcall 
#define TEA_HANDLE(x)	typedef struct x##__* x;
TEA_HANDLE(HWND)

// Debug definitions
#define TEA_NOP()		0
#define TEA_ASSUME(x)	__assume(x)
#define TEA_BREAK()		__asm { int 3 }

// Variable argument lists
#define TEA_VARARGS		__cdecl

// Don't generate vtables for this class
#define TEA_NOVTABLE	__declspec(novtable)

TEA_NAMESPACE_BEGIN

// Integer types
typedef __int8 int8;
typedef unsigned __int8 uint8;
typedef __int16 int16;
typedef unsigned __int16 uint16;
typedef __int32 int32;
typedef unsigned __int32 uint32;
typedef __int64 int64;
typedef unsigned __int64 uint64;

// Floating point types
#define REAL(x) float(x)
#define REAL_EPSILON 0.000001
typedef float real;
typedef double surreal;

// Fix the stack format for local assembler routines
#define TEA_MATHAPI __fastcall

// Additional definitions
namespace System
{
	// System types
	typedef int64 Timer;
	typedef int64 TimerElapsed;
	typedef HWND Window;

	// Rounding flags
	enum Round8087
	{
		RoundNearest =	0x0000,
		RoundDown =		0x0400,
		RoundUp =		0x0800,
		RoundChop =		0x0C00
	};
}

TEA_NAMESPACE_END

// Global assembler routines (both float and double, just in case the 'real' definition is changed)
TEA::int32 TEA_MATHAPI System__mathRound(const float& x, TEA::int32& how);
TEA::int32 TEA_MATHAPI System__mathRound(const double& x, TEA::int32& how);
void TEA_MATHAPI System__mathSinCos(float& sin, float& cos, const float& x);
void TEA_MATHAPI System__mathSinCos(double& sin, double& cos, const double& x);

#else

//----------------------------------------------------------------------------
//	SECOND TIME THROUGH THIS HEADER
//----------------------------------------------------------------------------

// Intrinsic functions
#include <math.h>
#include <memory.h>
#include <string.h>

// Intrinsic functions
#pragma intrinsic(fabs, sqrt, exp, pow, log, sin, cos, tan, atan, atan2)
#pragma intrinsic(memcpy, memset, memcmp, strlen, strcpy, strcmp)

// Windows stuff
TEA_WINAPI(int) IsBadReadPtr(const void *ptr, size_t bytes);
TEA_WINAPI(int) IsBadWritePtr(void *ptr, size_t bytes);

TEA_NAMESPACE_BEGIN

inline void* System::memoryCopy(void* dst, const void* src, size_t bytes)
{
	// Copy bytes implementation
	return ::memcpy(dst, src, bytes);
}

inline void* System::memoryZero(void* dst, size_t bytes)
{
	// Zero bytes implementation
	return ::memset(dst, 0, bytes);
}

inline bool System::memoryIsRead(const void* ptr, size_t bytes)
{
	// Check the pointer points to a valid read block
	return ::IsBadReadPtr(ptr, bytes) == 0;
}

inline bool System::memoryIsWrite(void* ptr, size_t bytes)
{
	// Check the pointer points to a valid read block
	return ::IsBadWritePtr(ptr, bytes) == 0;
}

inline size_t System::asciizLength(const char* src)
{
	// Fetch the length of ASCIIZ text
	return ::strlen(src);
}

inline void System::asciizCopy(char* dst, const char* src, size_t size)
{
	// Copy ASCIIZ text to a buffer
	if (size == 0)
		return;
	while (--size > 0)
	{
		if (*src == '\0')
			break;
		*(dst++) = *(src++);
	}
	*dst = '\0';
	return;
}

inline void System::asciizConcat(char* dst, const char* src, size_t size)
{
	// Concatenate one string to another
	size_t len1 = ::strlen(dst);
	if (len1 >= size)
		return;
	size_t len2 = ::strlen(src);
	dst += len1;
	if ((len1 + len2) >= size)
		len2 = size - len1 - 1;
	::memcpy(dst, src, len2);
	dst[len2] = '\0';
	return;
}

inline int System::asciizCompare(const char* lhs, const char* rhs)
{
	// Compare two ASCIIZ strings
	return ::strcmp(lhs, rhs);
}

inline double TEA::System::mathSqrt(double value)
{
	// Square root (returns zero for negatives)
	return (value < 0) ? 0 : ::sqrt(value);
}

inline double TEA::System::mathPower(double x, double y)
{
	// Power
	return ::pow(x, y);
}

inline long TEA::System::mathTruncate(double value)
{
	// Truncate a real value towards zero
	int32 how = RoundChop;
	return System__mathRound(value, how);
}

inline long TEA::System::mathRound(double value)
{
	// Round a real value to the nearest integer
	int32 how = RoundNearest;
	return System__mathRound(value, how);
}

inline long TEA::System::mathRoundUp(double value)
{
	// Round a real value towards plus infinity
	int32 how = RoundUp;
	return System__mathRound(value, how);
}

inline long TEA::System::mathRoundDown(double value)
{
	// Round a real value towards minus infinity
	int32 how = RoundDown;
	return System__mathRound(value, how);
}

inline void TEA::System::mathSinCos(float x, float& sin, float& cos)
{
	// Calculate both sine and cosine together
	System__mathSinCos(sin, cos, x);
	return;
}

inline void TEA::System::mathSinCos(double x, double& sin, double& cos)
{
	// Calculate both sine and cosine together
	System__mathSinCos(sin, cos, x);
	return;
}

inline double TEA::System::mathSin(double x)
{
	// Calculate sine
	return ::sin(x);
}

inline double TEA::System::mathCos(double x)
{
	// Calculate cosine
	return ::cos(x);
}

inline double TEA::System::mathTan(double x)
{
	// Calculate tangent
	return ::tan(x);
}

inline double TEA::System::mathAsin(double sin_x)
{
	// Arcsine
	double cos_x = ::sqrt(1 - sin_x * sin_x);
	return::atan2(sin_x, cos_x);
}

inline double TEA::System::mathAcos(double cos_x)
{
	// Arccosine
	double sin_x = ::sqrt(1 - cos_x * cos_x);
	return::atan2(sin_x, cos_x);
}

inline double TEA::System::mathAtan(double x, double y)
{
	// Arctangent
	return ::atan2(x, y);
}

// Ignore inlined functions
#pragma warning(disable: 4711)

TEA_NAMESPACE_END

#endif
