//----------------------------------------------------------------------------
//	SOURCE:	WinMain.cpp
//	AUTHOR:	Ian Taylor
//	NOTES:	Main entry-point of Curling game
//----------------------------------------------------------------------------

#include "Curling.h"
#include <windows.h>
#include <commctrl.h>
#include "App.h"
#include "Dialog.h"
#include "Resource.h"

TEA_NAMESPACE_USE

// Application instances
static HINSTANCE g_hinstance = NULL;
static DialogMain g_main;
static DialogConfig g_config;

static BOOL CALLBACK MainProc(HWND hwnd, UINT umsg, WPARAM wparam, LPARAM lparam)
{
	MONITOR(MainProc)

	// Main message handler for the main dialog box
	if (umsg == WM_INITDIALOG)
	{
		g_main.initialise(hwnd, reinterpret_cast<Application*>(lparam));
		return TRUE;
	}
	if (g_main.handle() != NULL)
		return g_main.onMessage(umsg, wparam, lparam);
	return FALSE;
}

static BOOL CALLBACK ConfigProc(HWND hwnd, UINT umsg, WPARAM wparam, LPARAM lparam)
{
	MONITOR(ConfigProc)

	// Main message handler for the config dialog box
	if (umsg == WM_INITDIALOG)
	{
		g_config.initialise(hwnd, reinterpret_cast<Configuration*>(lparam));
		return TRUE;
	}
	if (g_config.handle() != NULL)
		return g_config.onMessage(umsg, wparam, lparam);
	return FALSE;
}

static BOOL CALLBACK ViolationProc(HWND hwnd, UINT umsg, WPARAM wparam, LPARAM lparam)
{
	MONITOR(ViolationProc)

	// Main message handler for the rule violation dialog box
	switch (umsg)
	{
		case WM_INITDIALOG:
			// Play a sound and set the string
			if (lparam)
				VERIFY(::SendMessage(::GetDlgItem(hwnd, IDC_VIOLATION), WM_SETTEXT, 0, lparam));
			VERIFY(::MessageBeep(MB_ICONHAND));
			return TRUE;

		case WM_COMMAND:
			// Must be a button
			if (HIWORD(wparam) == BN_CLICKED)
				::EndDialog(hwnd, LOWORD(wparam));
			return TRUE;

		case WM_CLOSE:
			// The user clicked on the close button
			::EndDialog(hwnd, IDCLOSE);
			return TRUE;
	}
	return FALSE;
}

static BOOL CALLBACK AboutProc(HWND hwnd, UINT umsg, WPARAM wparam, LPARAM)
{
	MONITOR(AboutProc)

	// Main message handler for the about dialog box
	switch (umsg)
	{
		case WM_COMMAND:
			// Must be a button
			if (HIWORD(wparam) == BN_CLICKED)
				::EndDialog(hwnd, LOWORD(wparam));
			return TRUE;

		case WM_CLOSE:
			// The user clicked on the close button
			::EndDialog(hwnd, IDCLOSE);
			return TRUE;
	}
	return FALSE;
}

static LRESULT CALLBACK CanvasProc(HWND hwnd, UINT umsg, WPARAM wparam, LPARAM lparam)
{
	MONITOR(CanvasProc)

	// Window message handler for the canvas
	switch (umsg)
	{
		case WM_PAINT:
			// General canvas paint
			g_main.onCanvasPaint();
			break;

		case WM_ERASEBKGND:
			// Don't erase the background
			return TRUE;

		case WM_SIZE:
			// The canvas size may have changed
			g_main.onCanvasResize();
			break;

		case WM_LBUTTONDOWN:
			// A mouse click
			g_main.onCanvasClick(LOWORD(lparam), HIWORD(lparam));
			break;

		case WM_LBUTTONDBLCLK:
			// A mouse double click
			g_main.onCanvasThrow();
			break;

		case WM_CREATE:
			// Creation
			g_main.onCanvasCreate(hwnd);
			break;
	}
	return ::DefWindowProc(hwnd, umsg, wparam, lparam);
}

Window Application::hwnd(void) const
{
	MONITOR(Application::hwnd)

	// Fetch the main window handle
	return g_main.handle();
}

Window Application::hcanvas(void) const
{
	MONITOR(Application::hcanvas)

	// Fetch the canvas window handle
	HWND hwnd = g_main.handle();
	return (hwnd == NULL) ? NULL : ::GetDlgItem(hwnd, IDC_CANVAS);
}

void Application::canvasSize(int& sx, int& sy) const
{
	MONITOR(Application::canvasSize)

	// Fetch the canvas size
	HWND hwnd = g_main.handle();
	if (hwnd != NULL)
	{
		hwnd = ::GetDlgItem(hwnd, IDC_CANVAS);
		if (hwnd != NULL)
		{
			RECT rect;
			if (::GetWindowRect(hwnd, &rect))
			{
				// Hurrah
				sx = rect.right - rect.left;
				sy = rect.bottom - rect.top;
				return;
			}
		}
	}
	sx = sy = 0;
	return;
}

bool Application::onDialogAbout(void)
{
	MONITOR(Application::onDialogAbout)

	// Display the dialog box
	HWND hwnd = g_main.handle();
	int retval = ::DialogBox(g_hinstance, MAKEINTRESOURCE(IDD_ABOUT), hwnd, AboutProc);
	return (retval >= 0);
}

int Application::onDialogConfig(void)
{
	MONITOR(Application::onDialogConfig)

	// Display the configuration dialog
	HWND hwnd = g_main.handle();
	return ::DialogBoxParam(g_hinstance, MAKEINTRESOURCE(IDD_CONFIG), hwnd, ConfigProc, reinterpret_cast<LPARAM>(&m_config));
}

bool Application::onDialogViolation(const char* violation)
{
	MONITOR(Application::onDialogFreeGuard)

	// Display the dialog box
	HWND hwnd = g_main.handle();
	int retval = ::DialogBoxParam(g_hinstance, MAKEINTRESOURCE(IDD_VIOLATION), hwnd, ViolationProc, reinterpret_cast<LPARAM>(violation));
	return (retval >= 0);
}

int WINAPI WinMain(HINSTANCE hinstance, HINSTANCE, char*, int cmdshow)
{
	MONITOR(WinMain)

	// Main entry-point
	ASSERT(hinstance != NULL);

	// Initialise the application
	g_hinstance = hinstance;
	g_main.initialise(NULL, NULL);
	g_config.initialise(NULL, NULL);
	Application app;
	if (!app.initialise())
		return 0;

	// Register the slider common control
	INITCOMMONCONTROLSEX icc = { sizeof(icc), ICC_BAR_CLASSES };
	if (!::InitCommonControlsEx(&icc))
		return app.fatalError("::InitCommonControlsEx");

	// Register the 3D canvas class
	WNDCLASS wc = { CS_OWNDC|CS_HREDRAW|CS_VREDRAW|CS_DBLCLKS };
	wc.lpfnWndProc = CanvasProc;
	wc.hInstance = g_hinstance;
	wc.hCursor = ::LoadCursor(NULL, IDC_CROSS);
	wc.lpszClassName = "CurlingCanvas";
	if (!::RegisterClass(&wc))
		return app.fatalError("::RegisterClass");

	// Create the main window
	HWND hwnd = ::CreateDialogParam(g_hinstance, MAKEINTRESOURCE(IDD_MAIN), NULL, MainProc, reinterpret_cast<LPARAM>(&app));
	if (hwnd == NULL)
		return app.fatalError("::CreateDialogParam");
	::ShowWindow(hwnd, cmdshow);

	// Each new game
	MSG msg;
	do
	{
		// Get the options from the user
		if (app.onDialogConfig() != IDOK)
			break;
		app.newRenderer(Renderer::DirectX7);

		// Start a new game
		Game& game = app.newGame();
		game.updateControls(g_main, game.colour());

		// Enable sound if necessary
		if (!app.enableSound(game.config->sound))
			return app.fatalError("Application::enableSound");		
		g_main.setMenuCheck(IDM_OPTIONS_SOUND, game.config->sound);

		// Enable quick camera if necessary
		VERIFY(app.enableQuick(game.config->quick));
		g_main.setMenuCheck(IDM_OPTIONS_QUICK, game.config->quick);

		// Start the first end
		VERIFY(game.nextStone(g_main));

		// Main message loop
		msg.message = WM_NULL;
		do
		{
			// Only perform idle action if we're in asynchronous mode
			if (game.dynamic != NULL)
			{
				// Poll for messages but don't hang around
				if (::PeekMessage(&msg, NULL, 0, 0, PM_REMOVE) == FALSE)
				{
					// Do something during idle time (no messages are waiting)
					app.onIdle(g_main);
					continue;
				}
			}
			else if (::GetMessage(&msg, NULL, 0, 0) != TRUE)
			{
				// Early bath
				ASSERT(msg.message == WM_QUIT);
				break;
			}

			// Translate and dispatch the message
			if (!::IsDialogMessage(hwnd, &msg))
			{
				::TranslateMessage(&msg);
				::DispatchMessage(&msg);
			}
		}
		while (msg.message != WM_QUIT);
		app.silence();
	}
	while (msg.wParam);

	// Ensure we shutdown fully
	app.shutdown();
	return 0;
}
